#include "common.hpp"
#include "Camera.h"
#include <math.h>
#include <algorithm>

//#include <glm\gtx\transform2.hpp>

Camera::Camera(void) : m_position(), m_center(), m_up(), m_old_ProjectionMatrix(), m_old_ViewMatrix()
{
	m_viewMatrix_cpy = new glm::mat4[6];
}


Camera::~Camera(void)
{
}

glm::vec3 Camera::GetPosition()
{
	return m_position;
}

glm::vec3 Camera::GetUpVector()
{
	return m_up;
}

glm::vec3 Camera::GetCenterPositon()
{
	return m_center;
}

int Camera::SetViewport(int width, int height)
{
	m_width = width;
	m_height = height;
	return 1;
}

int Camera::GetWidth()
{
	return m_width;
}

int Camera::GetHeight()
{
	return m_height;
}

glm::mat4 Camera::LookAt(glm::vec3 eye, glm::vec3 center, glm::vec3 up)
{
	m_position = eye;
	m_center = center;
	m_up = up;

	glm::vec3 forward = center - eye;
	forward = glm::normalize(forward);
	glm::vec3 side = glm::cross(forward, up);
	side = glm::normalize(side);
	up = glm::cross(side, forward);

	glm::mat4 m; // Defaults to identity matrix
	m[0][0] = side.x;	m[0][1] = up.x;		m[0][2] = -forward.x;	
	m[1][0] = side.y;	m[1][1] = up.y;		m[1][2] = -forward.y;	
	m[2][0] = side.z;	m[2][1] = up.z;		m[2][2] = -forward.z;	

	m =  glm::translate(m, -eye);

	return m;
}

void Camera::SetProjection(float left, float right, float bottom, float top, float near_plane, float far_plane)
{	
	m_projection = glm::ortho(left, right, bottom, top, near_plane, far_plane);
	nearPlane = near_plane;
	farPlane = far_plane;
}

void Camera::SetProjection(float fovy, float aspect_ratio, float near_plane, float far_plane)
{
	 m_projection = glm::perspective(fovy, aspect_ratio, near_plane, far_plane);
	 nearPlane = near_plane;
	 farPlane = far_plane;
}

void Camera::SetViewMatrix(glm::vec3 eye, glm::vec3 center, glm::vec3 up)
{
	m_viewMatrix = LookAt(eye, center, up);
	//m_viewMatrix = glm::lookAt(eye, center, up);
}

void Camera::SetViewMatrix(int z, glm::vec3 eye, glm::vec3 center, glm::vec3 up)
{
	m_viewMatrix_cpy[z] = LookAt(eye, center, up);
	//m_viewMatrix = glm::lookAt(eye, center, up);
	//CalcPlanes(eye, center, up);
}

void Camera::SetViewMatrix(glm::mat4 matrix)
{
	m_viewMatrix = matrix;
}

void Camera::SetEnvironmentMapping(bool env)
{
	isEnvironmentMapping = env;
}

glm::mat4 Camera::GetProjection()
{
	return m_projection;
}

glm::mat4 Camera::GetViewMatrix()
{
	return m_viewMatrix;
}

glm::mat4 Camera::GetOldProjection()
{
	return m_old_ProjectionMatrix;
}

glm::mat4 Camera::GetOldViewMatrix()
{
	return m_old_ViewMatrix;
}

void Camera::SetActAsOldViewMatrix()
{
	m_old_ViewMatrix = m_viewMatrix;
}

void Camera::SetActAsOldProjectionMatrix()
{
	m_old_ProjectionMatrix = m_projection;
}

glm::mat4 Camera::GetViewMatrix(unsigned int i)
{
	if(isEnvironmentMapping) return m_viewMatrix_cpy[i];
	else					 return	m_viewMatrix;
}