#pragma once
#include <glm\glm.hpp>
#include <glm\gtc\matrix_projection.hpp>
#include <glm\gtc\matrix_transform.hpp>

class Camera
{
public:
	/**
	* Standard Camera Konstruktor
	**/
	Camera(void);
	/**
	* Standard Camera Dekonstruktor
	**/
	~Camera(void);

	/**
	* Sets Projection matrix Orthogonoal
	* @param left linke Camera Lnge
	* @param right rechte Camera Lnge
	* @param bottom untere Camera Lnge
	* @param top obere Camera Lnge
	* @param near_plane near Clippingplane
	* @param far_plane far Clippingplane
	**/
	void SetProjection(float left, float right, float bottom, float top, float near_plane, float far_plane);
	/**
	* Sets Projection matrix Projections
	* @param fovy fovy
	* @param aspect_ratio aspect ratio
	* @param near_plane near Clippingplane
	* @param far_plane far Clippingplane
	**/
	void SetProjection(float fovy, float aspect_ratio, float near_plane, float far_plane);
	/**
	* Return Projection Matrix
	* @return Projection Matrix
	**/
	glm::mat4 GetProjection();
	/**
	* Return ViewMatrix
	* @return View Matrix
	**/
	glm::mat4 GetViewMatrix();	//Diese Methode setzt CalcPlanes nicht mit!!!
	/**
	* Return ViewMatrix
	* @return View Matrix
	**/
	glm::mat4 GetViewMatrix(unsigned int i);	//Diese Methode setzt CalcPlanes nicht mit!!!
	/**
	* Set Viewmatrix
	* @param p ... Eye-point
	* @param p ... Center point
	* @param p ... Up Vektor
	**/
	void SetViewMatrix(glm::vec3 p, glm::vec3 l, glm::vec3 u);	//Diese Methode setzt CalcPlanes immer mit.
	/**
	* Set Viewmatrix
	* @param matrix ... Matrix
	**/
	void SetViewMatrix(glm::mat4 matrix);

	/**
	* Returns Projection Matrix from previous Frame
	* @return matrix ... Matrix
	**/
	glm::mat4 GetOldProjection();
	/**
	* Returns View Matrix from previous Frame
	* @return matrix ... Matrix
	**/
	glm::mat4 GetOldViewMatrix();	

	/**
	* Returns Viewpoing Positions
	* @return Matrix
	**/
	glm::vec3 GetPosition();
	/**
	* Returns UpVector
	* @return Matrix
	**/
	glm::vec3 GetUpVector();
	/**
	* Returns Center Positions
	* @return Matrix
	**/
	glm::vec3 GetCenterPositon();

	/**
	* Sets last frame viewing matrix
	**/
	void SetActAsOldViewMatrix();
	/**
	* Sets last frame projection matrix
	**/
	void SetActAsOldProjectionMatrix();

	/**
	* OpenGL LookAt Method
	**/
	glm::mat4 LookAt(glm::vec3 eye, glm::vec3 center, glm::vec3 up);

	/**
	* OpenGL SetViewport Method
	**/
	int SetViewport(int width, int height);
	/**
	* Returns Viewport width
	**/
	int GetWidth();
	/**
	* Returns Viewport height
	**/
	int GetHeight();

	/**
	* Enables environment mapping
	**/
	void SetEnvironmentMapping(bool env);
	/**
	* Sets Viewing Matrix for each plane of the environment map
	**/
	void SetViewMatrix(int z, glm::vec3 eye, glm::vec3 center, glm::vec3 up);

private:
	glm::mat4 m_projection;
	glm::mat4 m_viewMatrix;
	glm::mat4* m_viewMatrix_cpy;

	glm::mat4 m_old_ViewMatrix;
	glm::mat4 m_old_ProjectionMatrix;

	float nearPlane;
	float farPlane;

	int m_width;
	int m_height;

private:
	glm::vec3 m_position;
	glm::vec3 m_center;
	glm::vec3 m_up;

	bool isEnvironmentMapping;
};

