#include "Clustering.h"
#include "dataanalysis.h"

#define numDivision 8.f
#define PI 3.1415926535f

Clustering::Clustering(void)
{
}


Clustering::~Clustering(void)
{
}

float Dis(glm::vec2 x, glm::vec2 m, glm::mat2 s)
{
	float dim = 2.f;
	glm::vec2 v1 = x-m;
	glm::vec2 tmp = (glm::mat2(v1.x, 0.f, v1.y, 0.f)*glm::inverse(s)) * v1;

	float v = tmp.x;
	return 1.f/(sqrt( (float)(glm::determinant(s))) * powf(2.f*PI, dim/2.f)) * std::exp((float)(-0.5f*v));
}

glm::vec2 Mean(glm::vec2* x, float* g, int N)
{
	glm::vec2 m_new = glm::vec2(0.f, 0.f);
	for(int n = 0; n < N; n++)
	{
		if(!(g[n] != g[n]))	m_new += g[n]*x[n];	
	}

	//std::cout << "LL - M: " << m_new.x << " " << m_new.y << std::endl;
	return m_new;
}

glm::mat2 Covariance(glm::vec2* x, glm::vec2 m_new, float* g, float Nk, int N)
{
	int Dim = 2;
	glm::mat2 c_new;

	for(int n = 0; n < N; n++)
	{
		glm::vec2 v = x[n]-m_new;
		glm::mat2 tmp = glm::mat2(v.x, v.y, 0.f, 0.f);
		c_new += g[n] * (tmp * glm::transpose(tmp));
	}

    //c_new = c_new / Nk;
	//std::cout << "C: " << c_new[0][0] << " "  << c_new[0][1] << " "  << c_new[1][0] << " "  << c_new[1][1] << " " << std::endl;
	return c_new;
}

float LogLikelihood(glm::vec2* x, glm::vec2* m, glm::mat2* c, float* p, glm::vec2 dims)
{
	int N = dims.x;
	int K = dims.y;
	//for(int i = 0; i < K; i++)	std::cout << "m: " << m[i].x << " " << m[i].y << std::endl;
	
    float ll_new = 0;
	for(int n = 0; n < N; n++)
	{
		float helper = 0;
        for(int k = 0; k < K; k++)
		{
			if(p[k] > 0.f && !(p[k] != p[k])) 	helper += p[k]* Dis(x[n], m[k], c[k]);            
		}
    	ll_new += log(helper);
	}

	return ll_new;
}

glm::vec3* parallelKMeans(glm::vec2* x, glm::vec2* m1, int K, int N)
{
	glm::vec2* m = new glm::vec2[K];
	glm::vec3* m_new = new glm::vec3[K];
	for(int k = 0; k < K; k++) 
	{
		m_new[k] = glm::vec3(0.f,0.f,0.f);
		m[k] = m1[k];
	}

	glm::vec2 sum = glm::vec2();
	for(int n = 0; n < N; n++)
	{	
		float tmpL = 9.f;
		int tmpI = -1;
		
		sum += x[n];
		for(int k = 0; k < K; k++)
		{
			float len = glm::distance(x[n],m[k]);
			if(tmpL >= len)
			{
				tmpL = len;
				tmpI = k;
			}
		}
		m_new[tmpI] += glm::vec3(x[n].x, x[n].y, 1.f);
	}

	//for(int n = 0; n < N; n++) std::cout << "X: " << x[n].x << " " << x[n].y << std::endl;
	/*std::cout << "K: " << K << " N: " << N << std::endl;
	std::cout << "SUM: " << sum.x << " " << sum.y << std::endl;
	for(int i = 0; i < K; i++)
	{
		std::cout << "old Mittelwert: " << m[i].x << " " << m[i].y << std::endl;
		std::cout << "new Mittelwert: " << m_new[i].x << " " << m_new[i].y << " i: " << m_new[i].z << " >" << std::endl;
	}
	std::cout << "-----" << std::endl;*/

	return m_new;
}

float** parallelSetCluster(glm::vec2* x, glm::vec2* m, int K, int N)
{
	float** g = new float*[K];
	for(int i = 0; i < K; i++) g[i] = new float[N];

	for(int n = 0; n < N; n++)
	{	
		float tmpL = 999.f;
		int tmpI = -1;
		for(int k = 0; k < K; k++)
		{
			g[k][n] = 0.f;			
			float len = glm::distance(x[n], m[k]);
			if(tmpL >= len)
			{
				tmpL = len;
				tmpI = k;
			}
		}
		g[tmpI][n] = 1.f;
	}
	return g;
}

float parallelCalcJ(glm::vec2* x, glm::vec2* m, float** g, int K, int N)
{
	float J = 0.f;
	for(int n = 0; n < N; n++)
	{
		int it = 0;
		for(int k = 0; k < K; k++) if(g[k][n] != 0.f) it = k;
		J += glm::distance(x[n], m[it]);
	}
	return J;
}

void Clustering::kMeans(glm::vec2** x, glm::vec2* m, float* numComponents, float** g, int N, int K, int Iterations)
{	
	glm::vec2* m_new = new glm::vec2[K];
	float old_J = 9999999999999.f;
	int iNumDiv = numDivision;
	//std::cout << "iNumDiv: " << iNumDiv << " - numDivision: " << numDivision << std::endl;

	for(int k = 0; k < K; k++)	
	{
		do
		{
			m[k] = x[rand()%iNumDiv][rand()%(int)(N/numDivision)];
			m[k].x = 1.f/(float)(k+1.5);
			m[k].y = 0.1f;
		}
		while(m[k].x == 0.f || m[k].y == 0.f);
		//std::cout << "RANDMITTEL: " << m[k].x << " " << m[k].y << std::endl;
	}

	for(int iter = 0; iter < 1; iter++)
	{
		for(int i = 0; i < K; i++) numComponents[i] = 0.f;

		QFuture<glm::vec3*>* futureparallelKMeans = new QFuture<glm::vec3*>[iNumDiv];
		for(int j = 0; j < iNumDiv; j++)	
		{
			/*
			//__CHECK FUNCTIONS__START
			glm::vec2 sum = glm::vec2();
			for(int i = 0; i < N/iNumDiv; i++)	sum += x[j][i];
			std::cout << "sum: " << sum.x << " " << sum.y << std::endl;
			//__CHECK FUNCTIONS__END
			*/
			futureparallelKMeans[j] = QtConcurrent::run(parallelKMeans, x[j], m, K, N/iNumDiv);
		}
		for(int j = 0; j < iNumDiv; j++)	
		{
			glm::vec3* m_tmp = futureparallelKMeans[j].result();
			for(int k = 0; k < K; k++) 
			{
				m_new[k] += glm::vec2(m_tmp[k]);
				numComponents[k] += m_tmp[k].z;
			}
		}

		for(int k = 0; k < K; k++) if(numComponents[k] > 0.f) m[k] = m_new[k]/numComponents[k];
		//for(int k = 0; k < K; k++) std::cout << "Mittelwert: " << m[k].x << " " << m[k].y << std::endl;

		QFuture<float**>* futureparallelSetCluster = new QFuture<float**>[iNumDiv];
		for(int j = 0; j < iNumDiv; j++)	futureparallelSetCluster[j] = QtConcurrent::run(parallelSetCluster, x[j], m, K, N/numDivision);

		for(int j = 0; j < iNumDiv; j++)	
		{
			float** m_tmp = futureparallelSetCluster[j].result();
			for(int i = 0; i < N/iNumDiv; i++)
			{
				int n = i+(N/iNumDiv*j);
				for(int k = 0; k < K; k++) g[k][n] = m_tmp[k][i];
			}
		}

		float J = 0.f;
		QFuture<float>* futureCalcJ = new QFuture<float>[iNumDiv];
		for(int j = 0; j < iNumDiv; j++) futureCalcJ[j] = QtConcurrent::run(parallelCalcJ, x[j], m, g, K, N/iNumDiv);
		for(int j = 0; j < iNumDiv; j++) J += futureCalcJ[j].result();			
		
		if(old_J < J) break;
		old_J = J;
	}

	delete[] m_new;
}

float** EM(glm::vec2* x, glm::vec2* m, glm::mat2* c, float* p, glm::vec2 dims)
{
	int N = dims.x;
	int K = dims.y;
	float** g = new float*[K];
	for(int k = 0; k < K; k++) g[k] = new float[N];

	float* tmp = new float[N];
	for(int n = 0; n < N; n++)	
	{
		tmp[n] = 0.f;
		for(int k = 0; k < K; k++)	tmp[n] += p[k] * Dis(x[n], m[k], c[k]);
	}
	for(int n = 0; n < N; n++)
	{
		for(int k = 0; k < K; k++)
			g[k][n] = p[k] *  Dis(x[n], m[k], c[k]) / tmp[n];
	}
	delete[] tmp;

	return g;
}


void Clustering::DoClustering(glm::vec2* m, glm::mat2* c, glm::vec2* x, int N, int K)
{
	float* mumCompForClass = new float[K];

	float* p = new float[K];
	float** g = new float*[K];
	for(int i = 0; i < K; i++)	g[i] = new float[N];

	
	float ll = 99999999999999.f;
	float* Nk = new float[K];

	//_____________
	int iNumDiv = numDivision;
	glm::vec2**  x1 = new glm::vec2*[iNumDiv];
	for(int j = 0; j < iNumDiv; j++)
	{
		//N...durch 4 teilbar.
		x1[j] = new glm::vec2[N/iNumDiv];
		std::copy(&x[0]+(N/iNumDiv*j), &x[0] + (N/iNumDiv*(j+1)), &x1[j][0]);
	}

	std::cout << "*********k-Means*********" << std::endl;
	kMeans(x1, m, mumCompForClass, g, N, K, 100);			
	for(int i = 0; i < K; i++) std::cout << "m: " << m[i].x << " " << m[i].y << std::endl;
	std::cout << std::endl << std::endl;

	std::cout << "******EM-Algorithmus******" << std::endl;
	for(int k = 0; k < K; k++)
	{
		p[k] = mumCompForClass[k]/(float)(N);
		Nk[k] =  mumCompForClass[k];
		std::cout << "p[k] " << p[k] << " Nk[k] " << Nk[k] << std::endl;;
	}

	QFuture<glm::mat2>* futureCO = new QFuture<glm::mat2>[K*iNumDiv];
	for(int j = 0; j < iNumDiv; j++)
	{
		for(int i = 0; i < K; i++)	
		{
			c[i] = glm::mat2(0.f);
			futureCO[(j*K)+i] = QtConcurrent::run(Covariance, x1[j], m[i], g[i], Nk[i], N/numDivision);
		}
	}				

	glm::vec2* m_new = new glm::vec2[K];
	glm::mat2* c_new = new glm::mat2[K];
	float* p_new = new float[K];
	float ll_new = 0.f;

	for(int j = 0; j < numDivision; j++)	for(int i = 0; i < K; i++)	c[i] += futureCO[(j*K)+i].result();
	for(int i = 0; i < K; i++)	c[i] /= Nk[i];
	for(int it = 0; it < 10; it++)
	{
		//------------------------------
		//E-Step:
		QFuture<float**>* futureEM = new QFuture<float**>[iNumDiv];
		for(int i = 0; i < iNumDiv; i++)	
		{
			futureEM[i] = QtConcurrent::run(EM, x1[i], m, c, p, glm::vec2(N/iNumDiv,K));
		}
		for(int i = 0; i < iNumDiv; i++)
		{
			float** tmpEM = futureEM[i].result();
			for(int j = 0; j < N/iNumDiv; j++)	for(int k = 0; k < K; k++)		g[k][(i*N/iNumDiv)+j] = tmpEM[k][j];
		}

		float*** g1 = new float**[K];
		for(int i = 0; i < K; i++)
		{
			g1[i] = new float*[iNumDiv];
			for(int j = 0; j < iNumDiv; j++)
			{
				g1[i][j] = new float[N/iNumDiv];
				for(int k = 0; k < N/iNumDiv; k++)
				{	
					g1[i][j][k] = g[i][(j*N/iNumDiv)+k];
				}
			}
		}

		//------------------------------
		//M-Step:
		//Calculate Nk
		for(int k = 0; k < K; k++)	
		{
			Nk[k] = 0.f;
			for(int n = 0; n < N; n++)	
				Nk[k] += g[k][n]; 
		}

		//Calculate Mean
		//PRoblem ist, dass G nicht aufteilt ist so wie das X
		QFuture<glm::vec2>* futureM = new QFuture<glm::vec2>[K*iNumDiv];
		for(int k = 0; k < K; k++)	m_new[k] = glm::vec2(0.f, 0.f);
		for(int j = 0; j < iNumDiv; j++) for(int k = 0; k < K; k++) futureM[(j*K)+k] = QtConcurrent::run(Mean, x1[j], g1[k][j], N/iNumDiv);
		for(int j = 0; j < iNumDiv; j++) for(int k = 0; k < K; k++)	m_new[k] += futureM[(j*K)+k].result();
		for(int k = 0; k < K; k++)	if(Nk[k] != 0.f && !(Nk[k] != Nk[k])) m_new[k] /= Nk[k];

		//___________________________________
		QFuture<glm::mat2>* futureC = new QFuture<glm::mat2>[K*iNumDiv];
		for(int j = 0; j < iNumDiv; j++)
		{
			for(int i = 0; i < K; i++)	
			{
				c_new[i] = glm::mat2(0.f);
				futureC[(j*K)+i] = QtConcurrent::run(Covariance, x1[j], m_new[i], g1[i][j], Nk[i], N/numDivision);
			}
		}				
		for(int k = 0; k < K; k++)	p_new[k] = Nk[k] / N;
		for(int j = 0; j < numDivision; j++)	for(int i = 0; i < K; i++)	c_new[i] += futureC[(j*K)+i].result();
		for(int i = 0; i < K; i++)	if(Nk[i] != 0.f) c_new[i] /= Nk[i];

        //Evaluate the log likelihood:
		float ll_new = LogLikelihood(x, m_new, c_new, p_new, glm::vec2(N,K));
        
        //Check for convergences
		if (ll_new != ll_new || ll_new < 0.f || abs(abs((ll)) - abs((ll_new))) < 100.0) break; 
     
        //Save components
		for(int k = 0; k < K; k++)
		{
			m[k] = m_new[k];
			c[k] = c_new[k];
			p[k] = p_new[k];
		}
		ll = ll_new;
	}

	for(int j = 0; j < numDivision; j++)	delete  x1[j];
	delete x1;
	delete[] p_new;
	delete[] m_new;
	delete[] c_new;
}

glm::mat2 Clustering::CalculateCovariance(glm::vec2 radien, float phi)
{
	//the two angles that rotate the coordinate axes to the major and minor axes, respectively. 
	//For a symmetric 2  2 matrix,  sig is equal to phi.

	glm::mat2 tmp1 = glm::mat2(-cos(phi), sin(phi), sin(phi), cos(phi));
	//glm::mat2 tmp2 = glm::mat2(-cos(phi), sin(phi), sin(phi), cos(phi));
	
															//sonst tmp2
	return tmp1 * glm::mat2(radien.x, 0.f, 0.f, radien.y) * glm::transpose(tmp1);
}

float Clustering::getDensity(glm::vec2 pos, glm::vec2 median, glm::mat2 covariance)
{
	glm::vec2 v = pos-median;
	glm::vec2 res = glm::mat2(v.x, 0.f, v.y, 0.f)*glm::inverse(covariance) *v;
	return exp(-0.5f* res.x);
}

glm::vec4* CalculateTexturePart(glm::vec2* means, glm::mat2* covariance, int start, int offset, glm::vec4* color)
{
	int K = 2;

	glm::vec4* tex = new glm::vec4[offset];
	for(int ind = 0; ind < offset; ind++)
	{
		glm::vec2 pos = glm::vec2((float)((start+ind)%1024)/1024.f, (float)((start+ind)/1024)/1024.f);
		for(int k = 0; k < K; k++)
		{
			glm::vec2 v = pos-means[k];
			glm::vec2 res = glm::mat2(v.x, 0.f, v.y, 0.f)*glm::inverse(covariance[k]) *v;
			float density = res.x;
			
			tex[ind] = density*color[k];
		}
	}
	return tex;
}

glm::vec4* Clustering::CalculateTFTexture(glm::vec2* median, glm::mat2* covariance, glm::vec4* color)
{
	int iNumDiv = numDivision;

	glm::vec4* image = new glm::vec4[1024*1024];

	int offset = (1024*1024/iNumDiv);

	QFuture<glm::vec4*>* futureEM = new QFuture<glm::vec4*>[iNumDiv];
	for(int i = 0; i < iNumDiv; i++)	
	{
		int start = i * offset;
		futureEM[i] = QtConcurrent::run(CalculateTexturePart, median, covariance, start, offset, color);
	}
	for(int i = 0; i < iNumDiv; i++)
	{
		glm::vec4* tmpEM = futureEM[i].result();
		for(int j = 0; j < offset; j++)		image[(i*offset)+j] = tmpEM[j];
		delete[] tmpEM;
	}
	return image;
}



void Clustering::SVD(glm::vec2& rw, glm::mat2& ru, glm::mat2& rvt, glm::mat2 covariance)
{
	alglib::real_2d_array C;
	C.setlength(2, 2);
	alglib::real_1d_array w;
	alglib::real_2d_array u;
	alglib::real_2d_array vt;

	C[0][0] = covariance[0][0];
	C[0][1] = covariance[1][0];
	C[1][0] = covariance[0][1];
	C[1][1] = covariance[1][1];

	alglib::rmatrixsvd(C, C.rows(), C.cols(), 1, 1, 2, w, u, vt);

	std::cout << "w : " << w[0] << " " << w[1] << std::endl;
	std::cout << "u : " <<  u[0][0] << " " <<  u[0][1] << " " <<  u[1][0] << " " <<  u[1][1] << std::endl;
	std::cout << "vt: " << vt[0][0] << " " << vt[0][1] << " " << vt[1][0] << " " << vt[1][1] << std::endl;

	rw = glm::vec2(w[0], w[1]);
	ru = glm::mat2(u[0][0], u[0][1], u[1][0], u[1][1]);
	rvt = glm::mat2(vt[0][0], vt[0][1], vt[1][0], vt[1][1]);
}

void Clustering::Eigen(glm::mat2& rvec, glm::vec2& rval, glm::mat2 covariance)
{
	// add this data to ALGLIB's format
	alglib::real_2d_array ptInput;

	double* mat = new double[4];
	mat[0] = covariance[0][0];
	mat[1] = covariance[0][1];
	mat[2] = covariance[1][0];
	mat[3] = covariance[1][1];
	ptInput.setcontent(2, 2, mat);
 
	// this guy gets passed in and is filled in with an integer status code
	alglib::ae_int_t info;
 
	// scalar values that describe variances along each eigenvector
	alglib::real_1d_array eigValues;
 
	// unit eigenvectors which are the orthogonal basis that we want
	alglib::real_2d_array eigVectors;
 
	// perform the analysis
	pcabuildbasis(ptInput, 2, 2, info, eigValues, eigVectors);
 
	// now the vectors can be accessed as follows:
	rvec[0][0] = eigVectors[0][0];
	rvec[1][0] = eigVectors[1][0]; 
	rvec[0][1] = eigVectors[0][1];
	rvec[1][1] = eigVectors[1][1];

	rval.x = eigValues[0];
	rval.y = eigValues[1];

	std::cout << "Eigenwerte: " << eigVectors[0][0] << " " << eigVectors[0][1] << " " << eigVectors[1][0] << " " << eigVectors[1][1] << std::endl;
	std::cout << "EigenValues: " << eigValues[0] << " " << eigValues[1] << std::endl;
}