#pragma once
#include "GLM/gtx.hpp"
#include "GLM/gtc.hpp"
#include "GLM/gtx/type_ptr.hpp"

#include "linalg.h"
#include "solvers.h"

#include <QColor>
#include <QtConcurrentRun>
#include <QFuture>
#include <QFutureWatcher>

// system includes
#include <fstream>
#include <iostream>

class Clustering
{
public:
	/**
	 * Standard Konstruktor Clustering
	 **/
	Clustering(void);
	/**
	 * Standard Dekonstruktor Clustering
	 **/
	~Clustering(void);

	/**
	 * Calculates Eigenvectors and values
	 * @param rvec ... Eigenvector
	 * @param rval ... Eigenvalues
	 * @param covariance ... Covariance Matrix
	 **/
	static void Eigen(glm::mat2& rvec, glm::vec2& rval, glm::mat2 covariance);
	/**
	 * Performs Singluar Value Decomposition
	 * @param rw ... rw values
	 * @param ru ... rv matrix
	 * @param rvt ... rvt matrix
	 * @param covariance ... Covariance Matrix
	 **/
	static void SVD(glm::vec2& rw, glm::mat2& ru, glm::mat2& rvt, glm::mat2 covariance);
	/**
	 * Returns Density of sample ponts
	 * @param pos ... Position to estimate Density
	 * @param mean ... mean value of cluster
	 * @param covariance ... covariance of cluster
	 **/
	static float getDensity(glm::vec2 pos, glm::vec2 mean, glm::mat2 covariance);
	/**
	 * Method calculates Covariance Matrix
	 * @param radii ... radiii
	 * @param phi ... phi
	 * @return Covariance Matrix
	 **/
	static glm::mat2 CalculateCovariance(glm::vec2 radii, float phi);
	/**
	 * Performs kMeans Clustering
	 * @param x ... Data vector
	 * @param x ... Data vector
	 * @param m ... Meanvalues
	 * @param numComponents ... Number of Samples of each Clusters
	 * @param g ... Properbility
	 * @param K ... Number of Clusters
	 * @param N ... Number of Samples
	 * @param Iterations... Number of Iterations
	 **/
	static void kMeans(glm::vec2** x, glm::vec2* m, float* numComponents, float** g, int N, int K, int Iterations);
	/**
	 * Performs Clustering 
	 * @param m ... Meanvalues
	 * @param c ... Covariance Matrix
	 * @param x ... Data vector
	 * @param k ... Number of Clusters
	 **/
	static void DoClustering(glm::vec2* m, glm::mat2* c, glm::vec2* x, int N, int K);

	/**
	 * Draw Cluster on TF
	 * @param m ... Meanvalues
	 * @param c ... Covariance Matrix
	 * @param x ... Color
	 **/
	static glm::vec4* CalculateTFTexture(glm::vec2* m, glm::mat2* covariance, glm::vec4* color);
};