#include "FBO.h"

FBO::FBO() : Texture(4), fbo1(0),  g_iWidth(1), g_iHeight(1), g_iDepth(1)
{

}

FBO::FBO(int width, int height, bool bingbong, unsigned int internal_format, unsigned int format, unsigned int interpolation_min, unsigned int interpolation_max, int type, unsigned int num_texture, bool helper) : Texture(2, bingbong), fbo1(0),  g_iWidth(width), g_iHeight(height), g_iDepth(1), m_internal_format(internal_format), m_format(format), m_interpolation_min(interpolation_min), m_interpolation_max(interpolation_max), m_type(type), m_num_texture(num_texture)
{	
	std::cout << "CREATE FBO" << std::endl;
	CreateFBO(depthbuffer1, texname, fbo1);
	if(bingbong)	CreateFBO(depthbuffer2, texname2, fbo2);	

	buffers.push_back(GL_COLOR_ATTACHMENT0);
	for(int i = 0; i < multi_outputs.size(); i++)
	{
		buffers.push_back(GL_COLOR_ATTACHMENT1+i);
	}
}

FBO::FBO(int width, int height, int depth, bool bingbong, unsigned int internal_format, unsigned int format, unsigned int interpolation_min, unsigned int interpolation_max, int type) : Texture(3, bingbong), fbo1(0), g_iWidth(width), g_iHeight(height), g_iDepth(depth), m_internal_format(internal_format), m_format(format), m_interpolation_min(interpolation_min), m_interpolation_max(interpolation_max), m_type(type), m_num_texture(1)
{	
	Create3DFBO(texname, fbo1);
	if(bingbong)	Create3DFBO(texname2, fbo2);

	buffers.push_back(GL_COLOR_ATTACHMENT0);
}

void FBO::CreateFBO(GLuint& depth, GLuint& texn, GLuint& mfbo)
{
	GLenum FBOstatus;

	float value;
	glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &value);
	std::cout << "Anistropic: " << value << std::endl;

	glGenTextures(1, &depth);
	glBindTexture(GL_TEXTURE_2D, depth);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, m_interpolation_min);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, m_interpolation_max);
	//glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	//glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );	
	if(m_type == TT_ANISOTRPIC)	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, value);
	glTexImage2D( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT32F, g_iWidth, g_iHeight, 0, GL_DEPTH_COMPONENT, GL_FLOAT, 0);
	if(m_type == TT_ANISOTRPIC || m_type == TT_MIPMAP) glGenerateMipmap(GL_TEXTURE_2D); 
	
	glGenTextures(1,&texn);
	glBindTexture(GL_TEXTURE_2D, texn);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, m_interpolation_min);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, m_interpolation_max);
	//glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	//glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	if(m_type == TT_ANISOTRPIC)	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, value);
	glTexImage2D(GL_TEXTURE_2D, 0, m_internal_format, g_iWidth, g_iHeight, 0, m_format, GL_FLOAT, 0);
	if(m_type == TT_ANISOTRPIC || m_type == TT_MIPMAP) glGenerateMipmap(GL_TEXTURE_2D); 
	glBindTexture(GL_TEXTURE_2D, 0);


	multi_outputs.resize(m_num_texture-1);
	for(int i = 0; i < multi_outputs.size(); i++)
	{
		unsigned int tex_tmp = 0;
		glGenTextures(1,&tex_tmp);
		glBindTexture(GL_TEXTURE_2D, tex_tmp);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, m_interpolation_min);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, m_interpolation_max);
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		if(m_type == TT_ANISOTRPIC)	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, value);
		glTexImage2D(GL_TEXTURE_2D, 0, m_internal_format, g_iWidth, g_iHeight, 0, m_format, GL_FLOAT, 0);
		if(m_type == TT_ANISOTRPIC || m_type == TT_MIPMAP) glGenerateMipmap(GL_TEXTURE_2D); 
		glBindTexture(GL_TEXTURE_2D, 0);

		Texture::TexturePtr tptr = Texture::TexturePtr(new Texture(2));
		tptr->SetTexName(tex_tmp);
		multi_outputs[i] = tptr;
	}

	glGenFramebuffers(1, &mfbo);
	glBindFramebuffer(GL_FRAMEBUFFER, mfbo);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT ,GL_TEXTURE_2D, depth, 0);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, texn, 0);
	for(int i = 0; i < multi_outputs.size(); i++)
	{
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT1+i, GL_TEXTURE_2D, multi_outputs[i]->GetTexName(), 0);
	}

//	GLenum FBOstatus;
	FBOstatus = glCheckFramebufferStatus(GL_FRAMEBUFFER);
	if(FBOstatus != GL_FRAMEBUFFER_COMPLETE) CheckFBO(FBOstatus);
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

void FBO::Create3DFBO(GLuint& texn, GLuint& fbo)
{
	glGenTextures(1,&texn);
	glBindTexture(GL_TEXTURE_3D, texn);
			
	// Remove artefact on the edges of the shadowmap
	glTexParameteri( GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri( GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri( GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);

	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glTexImage3D(GL_TEXTURE_3D, 0, m_internal_format, g_iWidth, g_iHeight, g_iDepth, 0, m_format, GL_FLOAT, 0);
	glBindTexture(GL_TEXTURE_3D, 0);

	GLenum FBOstatus;

	glGenFramebuffers(1, &fbo);		
	glBindFramebuffer(GL_FRAMEBUFFER, fbo);

	for(int z = 0; z < g_iDepth; z++)
	{
		glFramebufferTexture3D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_3D, texn, 0,z);
	}

	FBOstatus = glCheckFramebufferStatus(GL_FRAMEBUFFER);
	if(FBOstatus != GL_FRAMEBUFFER_COMPLETE) CheckFBO(FBOstatus);
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

FBO::~FBO(void)
{
	std::cout << "!!DELETE FBO" << std::endl;

	glDeleteTextures(1, &texname);
	glDeleteFramebuffers(1, &fbo1);
}

void FBO::Bind()
{
	//std::cout << "###" << std::endl;
	//std::cout << "Bind FBO" << std::endl;
	if(m_pingbong) 
	{
		if(current_texture == 0) 
		{
			glBindFramebuffer(GL_FRAMEBUFFER, fbo1);	//FRAMEBUFFEROBJECT START!!!
		}
		else 
		{
			glBindFramebuffer(GL_FRAMEBUFFER, fbo2);	//FRAMEBUFFEROBJECT START!!!
		}	
	}
	else	
	{
		glBindFramebuffer(GL_FRAMEBUFFER, fbo1);	//FRAMEBUFFEROBJECT START!!!
	}
	//std::cout << "###" << std::endl;
}

void FBO::Bind(int z)
{
	if(m_pingbong) 
	{
		if(current_texture == 0) 
		{
			glBindFramebuffer(GL_FRAMEBUFFER, fbo1);
			glFramebufferTexture3D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_3D, texname, 0, z);
		}
		else
		{
			glBindFramebuffer(GL_FRAMEBUFFER, fbo2);
			glFramebufferTexture3D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_3D, texname2, 0, z);
		}
	}
	else	glFramebufferTexture3D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_3D, texname, 0, z);
}

void FBO::BindTexture(int z)
{
	if(m_pingbong) 
	{
		if(current_texture == 0)	glFramebufferTexture3D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_3D, texname, 0, z);
		else						glFramebufferTexture3D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_3D, texname2, 0, z);
	}
	else	glFramebufferTexture3D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_3D, texname, 0, z);
}

void FBO::Unbind()
{
	if(m_pingbong) current_texture = (current_texture+1)%2;
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

void FBO::BindColorAttachementAtTextureAndEnable(unsigned int i)
{
	glActiveTexture(GL_TEXTURE0+i);
	glBindTexture(GL_TEXTURE_3D, texname);
}

int FBO::GetWidth()
{
	return g_iWidth;
}

int FBO::GetHeight()
{
	return g_iHeight;
}

int FBO::GetDepth()
{
	return g_iDepth;
}

void FBO::DrawBuffers()
{
	glDrawBuffers(buffers.size(), &buffers[0]);
}