#pragma once
#include "common.hpp"
#include "Texture.h"
#include "GL/glew.h"
#include <vector>

class FBO : public Texture
{
public:
	typedef boost::shared_ptr<FBO> FBOPtr;

	/**
	 * Standardconstructor FBO
	 */
	FBO();
	/**
	 * Constructor
	 * @param width ..Breite
	 * @param height ..Hhe
	 * @param bingbong ..Create Pingbong FBV
	 * @param internal_format ..internal format
	 * @param interpolation_min .. interpolation min
	 * @param interpolation_max .. interpolation max
	 * @param type ... Interpolation Type
	 * @param num_texture Helper Variable
	 * @param helper  Helper Variable
	 */
	FBO(int width, int height, bool bingbong=false, unsigned int internal_format=GL_RGBA32F, unsigned int format=GL_RGBA, unsigned int interpolation_min=GL_LINEAR, unsigned int interpolation_max=GL_LINEAR, int type=0, unsigned int num_texture = 1, bool helper = false);
	/**
	 * Constructor
	 * @param width ..Breite
	 * @param height ..Hhe
	 * @param bingbong ..Create Pingbong FBV
	 * @param internal_format ..internal format
	 * @param interpolation_min .. interpolation min
	 * @param interpolation_max .. interpolation max
	 * @param helper  Helper Variable
	 */
	FBO(int width, int height, int depth, bool bingbong=false, unsigned int internal_format=GL_RGBA32F, unsigned int format=GL_RGBA, unsigned int interpolation_min=GL_LINEAR, unsigned int interpolation_max=GL_LINEAR, int type=0);
	/**
	 * Standarddeconstructor
	 */
	~FBO(void);

	/**
	 * Check FBO
	 */
	void CheckFBO(GLenum FBOstatus)
	{
		if(FBOstatus == GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT)			printf("Incomplete attachment\n");
		if(FBOstatus == GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT)	printf("Missing attachment\n");
		if(FBOstatus == GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT)	printf("Incomplete dimensions\n");
		if(FBOstatus == GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER)			printf("Incomplete read buffer\n");
		if(FBOstatus == GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER)			printf("GL_FRAMEBUFFER_COMPLETE fr Blur Effekt konnte ned benutzt werden\n");
		if(FBOstatus == GL_FRAMEBUFFER_UNSUPPORTED)						printf("Framebufferobjects unsupported\n");
	};

private:
	void CreateFBO(GLuint& depth, GLuint& texn, GLuint& fbo);
	void Create3DFBO(GLuint& texn, GLuint& fbo);

public:
	/**
	 * Bind FBO
	 */
	void Bind();
	/**
	 * Bind FBO
	 * @param z z variable
	 */
	void Bind(int z);
	/**
	 * Bind Texture
	 * @param z z variable
	 */
	void BindTexture(int z);

	/**  
	Unbind: Entbindet den FBO.
   */
	void Unbind();
	/**  
	 Bind Color Attachement
   */
	void BindColorAttachementAtTextureAndEnable(unsigned int i);
	/*
	 * Draw Buffers
	 */
	void DrawBuffers();
	/**
	 * Returns width
	 * @param width
	 */
	int GetWidth();
	/**
	 * Returns height
	 * @param height
	 */
	int GetHeight();
	/**
	 * Returns Depth
	 * @param depth
	 */
	int GetDepth();

	GLuint fbo1;
	GLuint fbo2;
	GLuint depthbuffer2;
	GLuint depthbuffer1;
	std::vector<GLenum> buffers;
	int g_iWidth;
	int g_iHeight;
	int g_iDepth;

	unsigned int m_format;
	unsigned int m_internal_format;
	unsigned int m_interpolation_min;
	unsigned int m_interpolation_max;
	int m_type;


	unsigned int m_num_texture;
	std::vector<Texture::TexturePtr> multi_outputs;
};

