#include "Geometry.h"


Geometry::Geometry(void) : m_rendertype(RT_TRIANGLE), renderTo3DTexture(1), World(), Local(), bones(0), oldBones(0), m_bones_count(0), start_time(0.f)
{
}

Geometry::Geometry(unsigned int type) : m_rendertype(RT_TRIANGLE), m_type(type), renderTo3DTexture(1), World(), Local(), bones(0), m_bones_count(0), start_time(0.f)
{
}

Geometry::~Geometry(void)
{
}

void Geometry::SetShader(unsigned int id, MyShader* shader)
{
	if(m_shader.size() <= id) m_shader.resize(id+1);
	m_shader[id] = shader;
}

MyShader* Geometry::GetShader(unsigned int id)
{
	return m_shader[id];
}

void Geometry::CreateDictShaderToVBO(unsigned int id, std::map<std::string, int> dict)
{	
	if(m_dict_shader_vbo.size() <= id) m_dict_shader_vbo.resize(id+1);

	m_shader[id]->bind();
	for(std::map<std::string, int>::iterator it = dict.begin(); it != dict.end(); it++)
	{
		GLint attrib =  m_shader[id]->get_attrib_location(it->first);
		m_dict_shader_vbo[id].insert(std::pair<int, int>(attrib, it->second));
	}
	m_shader[id]->unbind();
}

void Geometry::SetDictShaderToVBO(unsigned int id, std::map<int, int> dict)
{
	if(m_dict_shader_vbo.size() <= id) m_dict_shader_vbo.resize(id+1);
	
	m_dict_shader_vbo[id] = dict;
}


void Geometry::CreateDictShaderUniforms(unsigned int id, std::map<std::string, int> dict)
{
	if(m_dict_shader_uniform.size() <= id) m_dict_shader_uniform.resize(id+1);

	m_shader[id]->bind();
	for(std::map<std::string, int>::iterator it = dict.begin(); it != dict.end(); it++)
	{
		GLint attrib =  m_shader[id]->get_uniform_location(it->first);
		m_dict_shader_uniform[id].insert(std::pair<int, int>(attrib, it->second));
	}
	m_shader[id]->unbind();
}

void Geometry::CreateDictShaderSpecificUniforms(unsigned int id)
{
	if(m_shader_uniforms.size() <= id) return;

	m_shader[id]->bind();
	for(int i = 0; i < m_shader_uniforms[id].size(); i++)
	{
		m_shader_uniforms[id][i]->m_id = m_shader[id]->get_uniform_location(m_shader_uniforms[id][i]->m_var);
	}
	m_shader[id]->unbind();
}

void Geometry::SetDictShaderUniforms(unsigned int id, std::map<int, int> dict)
{
	if(m_dict_shader_uniform.size() <= id) m_dict_shader_uniform.resize(id+1);
	m_dict_shader_uniform[id] = dict;
}

std::map<int, int> Geometry::GetDictShaderToVBO(unsigned int id)
{
	return m_dict_shader_vbo[id];
}

std::map<int, int> Geometry::GetDictShaderUniform(unsigned int id)
{
	return m_dict_shader_uniform[id];
}