#pragma once
#include <glm/glm.hpp>
#include "Camera.h"


class Light
{
public:
	/**
	 * Standardconstructor
	 **/
	Light(void);
	/**
	 * Konstruktor
	 * @param pos Lightpos
	 * @param ambient ambienter Lichtanteil
	 * @param diffuse diffuse Lichtanteil
	 * @param specular specular Lichtanteil
	 */
	Light(glm::vec3 pos, glm::vec4 ambient, glm::vec4 diffuse, glm::vec4 specular);
	/**
	 * Konstruktor
	 * @param pos Lightpos
	 * @param ambient ambienter Lichtanteil
	 * @param diffuse diffuse Lichtanteil
	 * @param specular specular Lichtanteil
	 */
	Light(glm::vec3 pos, glm::vec4 ambient, glm::vec4 diffuse, glm::vec4 specular, float shineness);
	/**
	 * Konstruktor
	 * @param pos Lightpos
	 * @param ambient ambienter Lichtanteil
	 * @param diffuse diffuse Lichtanteil
	 * @param specular specular Lichtanteil
	 */
	Light(glm::vec3 pos, glm::vec4 ambient, glm::vec4 diffuse, glm::vec4 specular, float shineness, float constant_attenuation, float linear_attenuation, float quadratic_attenuation);
	/**
	 * Dekonstruktor
	 */
	~Light(void);

	/**
	 * This method sets the light position
	 * @param pos lightpos
	 */
	void SetPosition(glm::vec3 pos);
	/**
	 * This method sets the ambient value
	 * @param ambient ambient value
	 */
	void SetAmbient(glm::vec4 ambient);
	/**
	 * This method sets the diffuse value
	 * @param diffuse diffuse value
	 */
	void SetDiffuse(glm::vec4 diffuse);
	/**
	 * This method sets the specular value
	 * @param specular specular value
	 */
	void SetSpecular(glm::vec4 specular);
	/**
	 * This method sets the constant attenuation value
	 * @param constant attenuation constant attenuation value
	 */
	void SetConstantAttenuation(float attenuation);
	/**
	 * This method sets the specular value
	 * @param linearAttenuation linear attenuation value
	 */
	void SetLinearAttenuation(float attenuation);
	/**
	 * This method sets the attentuation value
	 * @param attentuation attentuation value
	 */
	void SetQuadraticAttenuation(float attenuation);
	/**
	 * This method sets the shinesess value
	 * @param shinesess shinesess value
	 */
	void SetShineness(float shineness);

	/** 
	 * This method returns the Camera for Shadow mapping
	 * @return Camera camera for shadow mapping
	 */
	Camera* GetCamera();
	/** 
	 * This method sets the camera for Shadow mapping
	 * @camera Camera camera for shadow mapping
	 */
	void SetCamera(Camera* camera);

	/** 
	 * This method returns the position 
	 * @return position position value
	 */
	glm::vec3 GetPosition();
	/** 
	 * This method returns the ambient value 
	 * @return ambient ambient value
	 */
	glm::vec4 GetAmbient();
	/** 
	 * This method returns the diffuse value 
	 * @return diffuse diffuse value
	 */
	glm::vec4 GetDiffuse();
	/** 
	 * This method returns the specular value 
	 * @return specular specular value
	 */
	glm::vec4 GetSpecular();
	/** 
	 * This method returns the constant  
	 * @return position position value
	 */
	float GetConstantAttenuation();
	/** 
	 * This method returns the linear attenuation 
	 * @return linear_attenuation linear attenuation value
	 */
	float GetLinearAttenuation();
	/** 
	 * This method returns the quadratic attenuation 
	 * @return quadratic_attenuation quadratic attenuation value
	 */
	float GetQuadraticAttenuation();
	/** 
	 * This method returns the shineness value 
	 * @return shineness shineness value
	 */
	float GetShineness();

private:
	glm::vec3 m_position;
	glm::vec4 m_ambient;
	glm::vec4 m_diffuse;
	glm::vec4 m_specular;
	float m_constant_attenuation;
	float m_linear_attenuation;
	float m_quadratic_attenuation;
	float m_shineness;
	Camera* m_camera;
};

