#include <QtGui>
#include "MainWindow.h"
#include "TransferPoints.h"
#include "SceneManager.h"
#include "Volume.h"
#include <QFileDialog>
#include <QImage>
#include <QMessageBox>
#include <QProgressBar>
#include <QProgressDialog>

#include <iostream>
using std::cout;
using std::endl;
using std::cerr;



MainWindow::MainWindow(QWidget* parent) : QMainWindow(parent), m_num_cluster(2)
{
	setupUi(this);

	//transferWidget->hide();
	transferFunctionWidget->initTransferPoints("default", 2);
	transferFunctionWidget->initTransferPoints("default", 0);

	//MainWindow
	connect(actionOpen, SIGNAL(triggered()), this, SLOT(openFileDialog()));
	connect(actionSave, SIGNAL(triggered()), this, SLOT(saveFileDialog()));
	//connect(checkBox_showTransfer, SIGNAL(stateChanged(int)), this, SLOT(showTransferFunctionClicked(int)));
	//connect(checkBox_applyTransfer, SIGNAL(stateChanged(int)), this, SLOT(applyTranferFunctionClicked(int)));

	connect(radioButton_fixStep, SIGNAL(toggled(bool)), spinBox_Step, SLOT(setEnabled(bool)));
	connect(radioButton_adaptiveStep, SIGNAL(toggled(bool)), spinBox_Step, SLOT(setDisabled(bool)));
	connect(checkBox_activateLight, SIGNAL(toggled(bool)), checkBox_Phong, SLOT(setEnabled(bool)));
	
	//transferfunctionWidget
	connect(colortriangle, SIGNAL(colorChanged(QColor)), transferFunctionWidget, SIGNAL(changeColor(QColor)));
	//connect(saveButton, SIGNAL(clicked(bool)), this, SLOT(saveTransferFunction()));
	//connect(loadButton, SIGNAL(clicked(bool)), this, SLOT(loadTransferFunction()));
	connect(resetButton, SIGNAL(clicked(bool)), this, SLOT(resetTransferFunctionButton()));

	connect(actionLoad_Transferfunction, SIGNAL(triggered()), this, SLOT(loadTransferFunction()));
	connect(actionSave_Transferfunction, SIGNAL(triggered()), this, SLOT(saveTransferFunction()));
	connect(actionReset, SIGNAL(triggered()), this, SLOT(resetTransferFunctionButton()));

	//openGLWidget
	connect(tabWidget, SIGNAL(currentChanged(int)), openGLWidget, SLOT(setPerspective(int)));
	connect(this, SIGNAL(volumeLoaded()), openGLWidget, SLOT(initSliceTexture()));
	connect(sliceSlider_x, SIGNAL(valueChanged(int)), openGLWidget, SLOT(setXSlicing(int)));
	connect(sliceSlider_y, SIGNAL(valueChanged(int)), openGLWidget, SLOT(setYSlicing(int)));
	connect(sliceSlider_z, SIGNAL(valueChanged(int)), openGLWidget, SLOT(setZSlicing(int)));
	connect(this, SIGNAL(applyTransferFunction(float*, int, int)), openGLWidget, SLOT(generateTransferTexture(float*, int, int)));
	connect(this, SIGNAL(resetTransferFunction()), openGLWidget, SLOT(resetTransferTexture()));
	connect(transferFunctionWidget, SIGNAL(transferFunctionChanged(float*, int, int)), openGLWidget, SLOT(generateTransferTexture(float*, int, int)));
	connect(checkBox_activateLight, SIGNAL(toggled(bool)), openGLWidget, SLOT(setLight(bool)));
	connect(radioButton_Composite, SIGNAL(toggled(bool)), openGLWidget, SLOT(setCompositing(bool)));
	connect(radioButton_MIP, SIGNAL(toggled(bool)), openGLWidget, SLOT(setMip(bool)));
	connect(radioButton_XRay, SIGNAL(toggled(bool)), openGLWidget, SLOT(setAverage(bool)));
	connect(checkBox_Phong, SIGNAL(toggled(bool)), openGLWidget, SLOT(setPhong(bool)));
	connect(radioButton_adaptiveStep, SIGNAL(toggled(bool)), openGLWidget, SLOT(setAdaptive(bool)));
	connect(spinBox_Step, SIGNAL(valueChanged(double)), openGLWidget, SLOT(setStepSize(double)));
	connect(horizontalSlider_x, SIGNAL(valueChanged(int)), openGLWidget, SLOT(setXClipping(int)));
	connect(horizontalSlider_y, SIGNAL(valueChanged(int)), openGLWidget, SLOT(setYClipping(int)));
	connect(horizontalSlider_z, SIGNAL(valueChanged(int)), openGLWidget, SLOT(setZClipping(int)));

	//OpenGLColorgradientWidget
	connect(this, SIGNAL(resetColorgradient()), openGLColorgradientWidget, SLOT(resetTransferTexture()));
	connect(this, SIGNAL(applyColorgradient(float*, int, int)), openGLColorgradientWidget, SLOT(generateTransferTexture(float*, int, int)));
	connect(transferFunctionWidget, SIGNAL(ColorgradientChanged(float*, int, int)), openGLColorgradientWidget, SLOT(generateTransferTexture(float*, int, int)));

	connect(openGLWidget, SIGNAL(valuesUpdated(glm::vec2*, int)), transferFunctionWidget->m_transferPoints, SLOT(updateValues(glm::vec2*, int)));	
	connect(this, SIGNAL(valuesMode(int)), transferFunctionWidget->m_transferPoints, SLOT(updateMode(int)));	
	connect(this, SIGNAL(valuesMode(int)), openGLColorgradientWidget, SLOT(setMode(int)));	

	connect(this, SIGNAL(setActionObjectType(int)), transferFunctionWidget->m_transferPoints, SLOT(updateActionObjectType(int)));	

	connect(bt_apply, SIGNAL(clicked(bool)), this, SLOT(applyTranferFunctionClicked(bool)));

	connect(this, SIGNAL(DoingClustering(int)), transferFunctionWidget->m_transferPoints, SLOT(DoClustering(int)));	

	connect(this, SIGNAL(SetIsoSurfaceChanged(float)), openGLWidget, SLOT(changeIsoSurface(float)));
}

MainWindow::~MainWindow()
{	
	
}

void MainWindow::openFileDialog()
{
	QString fileName = QFileDialog::getOpenFileName(this, tr("Open Data File"), "", tr("DataFiles (*.dat)"));

	if (fileName != QString::null) {
		
		//QMessageBox::information(this, "Load Volume ", "from: " + fileName);
		QProgressDialog pd("Loading File", QString() , 0, 100, this);
		
		QProgressBar* bar = new QProgressBar(&pd);
		bar->setRange(0, 100);
		bar->setValue(0);
		pd.setBar(bar);
		pd.setMinimumWidth(350);
		pd.setMinimumDuration(1000);
		pd.setWindowModality(Qt::WindowModal);
		pd.setValue(0);
		pd.setAutoClose(true);
		pd.show();
	
		pd.setValue(1);
		SceneManager::get().loadVolume(fileName.toLocal8Bit().data());
		pd.setValue(50);	
		pd.setValue(100);

		emit volumeLoaded();

		sliceSlider_x->setMaximum(SceneManager::get().getVolume()->GetWidth());
		sliceSlider_y->setMaximum(SceneManager::get().getVolume()->GetHeight());
		sliceSlider_z->setMaximum(SceneManager::get().getVolume()->GetDepth());

		sliceSlider_x->setSliderPosition(SceneManager::get().getVolume()->GetWidth() / 2);
		sliceSlider_y->setSliderPosition(SceneManager::get().getVolume()->GetHeight() / 2);
		sliceSlider_z->setSliderPosition(SceneManager::get().getVolume()->GetDepth() / 2);

		double* histogram = SceneManager::get().getVolume()->getHistogram(8);
		int size = 4096/8;
		transferFunctionWidget->setHistogram(histogram, size);

		spinBox_Step->setValue(0.005);


		emit applyColorgradient(transferFunctionWidget->getTransferTextureData(), COLORS_IN_TRANSFER_ARRAY, 0);
		emit applyColorgradient(transferFunctionWidget->getTransferTextureData(), COLORS_IN_TRANSFER_ARRAY, 1);
		emit applyColorgradient(transferFunctionWidget->getTransferTextureData(), COLORS_IN_TRANSFER_ARRAY, 2);
	} 
}

void MainWindow::saveFileDialog()
{
	QString saveFileName = QFileDialog::getSaveFileName(this, tr("Save Image File"), "", tr("ImageFiles (*.png)"));
	glReadBuffer(GL_FRONT);
	QImage image = openGLWidget->grabFrameBuffer();
	bool saved = image.save(saveFileName);
	if(saved == false) QMessageBox::critical(this,"Error Message","Could not save Image");
	else QMessageBox::information(this, "Information Message", "Screenshot saved " + saveFileName);
}

void MainWindow::saveTransferFunction()
{
	QString saveFileName = QFileDialog::getSaveFileName(this, tr("Save Transferfunction"), "", tr("TextFile (*.txt)"));

	if (saveFileName != QString::null) 
	{
		transferFunctionWidget->saveTransferPoints(saveFileName);
		QMessageBox::information(this, "Information Message", "Transferfunction saved " + saveFileName);
	}
}

void MainWindow::loadTransferFunction()
{
	QString fileName = QFileDialog::getOpenFileName(this, tr("Open Text File"), "", tr("TextFile (*.txt)"));

	if (fileName != QString::null) 
	{
		int mode = 0;
		transferFunctionWidget->initTransferPoints(fileName, mode);
	}
}

void MainWindow::resetTransferFunctionButton()
{
	transferFunctionWidget->initTransferPoints("reset", m_mode);
	emit resetColorgradient();
}

void MainWindow::showTransferFunctionClicked(int state)
{
	/*
	//Obsolet
	if(state) 
	{
		// checked
		emit applyColorgradient(transferFunctionWidget->getTransferTextureData(), COLORS_IN_TRANSFER_ARRAY, 0);
		emit applyColorgradient(transferFunctionWidget->getTransferTextureData(), COLORS_IN_TRANSFER_ARRAY, 1);
		emit applyColorgradient(transferFunctionWidget->getTransferTextureData(), COLORS_IN_TRANSFER_ARRAY, 2);
		transferWidget->show();
	} 
	else 
	{
		// not checked
		emit resetColorgradient();
		transferWidget->hide();
	}
	*/
}

void MainWindow::applyTranferFunctionClicked(bool state)
{
	applyTranferFunctionClicked(Qt::Checked);
}

void MainWindow::applyTranferFunctionClicked(int state)
{
	if(state == 1)	state == Qt::Checked;

	if ((state == Qt::Checked) && (SceneManager::get().getVolume() != 0)) 
	{
		transferFunctionWidget->SetMode(m_mode);
		transferFunctionWidget->setTranferFunktionApplied(true);
		emit applyTransferFunction(transferFunctionWidget->getTransferTextureData(), COLORS_IN_TRANSFER_ARRAY, m_mode);
	} 
	else 
	{
		transferFunctionWidget->setTranferFunktionApplied(false);
		emit resetTransferFunction();
	}
}

void MainWindow::TabIndexChanged(int value)
{
	RenderStates::ModusChanged = value;	
}

void MainWindow::gradientChanged(int value)
{
	RenderStates::ParticleGradientValue = (float)(value)/100.f;	
}

void MainWindow::curvatureChanged(int value)
{
	RenderStates::ParticleCurvatureValue = (float)(value)/100.f;
}

void MainWindow::flockingChanged(int value)
{
	RenderStates::ParticleFlockingValue = (float)(value)/100.f;
}

void MainWindow::predefinedChanged(int value)
{
	RenderStates::ParticlePredefinedValue = (float)(value)/100.f;
}

void MainWindow::densityChanged(int value)
{
	RenderStates::ParticleDensityValue = (float)(value)/100.f;
}

void MainWindow::gradientEnabled(bool val)
{
	RenderStates::ParticleGradientEnable = val; 
}

void MainWindow::flockingEnabled(bool val)
{
	RenderStates::ParticleFlockingEnable = val;
}

void MainWindow::predefinedEnabled(bool val)
{
	RenderStates::ParticlePredefinedEnable = val;
}

void MainWindow::densityEnabled(bool val)
{
	RenderStates::ParticleDensityEnable = val;
}

void MainWindow::fluidDenstiyEnabled(bool val)
{

}

void MainWindow::fluidViscosityEnabled(bool val)
{
}

void MainWindow::fluidPressureEnabled(bool val)
{
}

void MainWindow::fluidCollisionEnabled(bool val)
{
	RenderStates::FluidCollisionEnable = val;
}

void MainWindow::fluidCollisionCTEnabled(bool val)
{
	RenderStates::FluidCollisionCTEnable = val;
}

void MainWindow::fluidSpringChanged(double val)
{
	RenderStates::FluidSpringValue = val;
}

void MainWindow::fluidDampingChanged(double val)
{
	RenderStates::FluidDampingValue = val;
}

void MainWindow::fluidShearChanged(double val)
{
	RenderStates::FluidShearValue = val;
}

void MainWindow::fluidSpringCTChanged(double val)
{
	RenderStates::FluidSpringCTValue = val;
}

void MainWindow::fluidDampingCTChanged(double val)
{
	RenderStates::FluidDampingCTValue = val;
}

void MainWindow::fluidShearCTChanged(double val)
{
	RenderStates::FluidShearCTValue = val;
}

void MainWindow::fluidParticleMass(double v)
{
	RenderStates::FluidParticleMass = v;
}

void MainWindow::fluidViscosity(double v)
{
	RenderStates::FluidViscosity = v;
}

void MainWindow::fluidSiftness(double v)
{
	RenderStates::FluidSiftness = v;
}

void MainWindow::fluidRestdensity(double v)
{
	RenderStates::FluidRestdensity = v;
}

void MainWindow::particleGravityChanged(double v)
{
	RenderStates::FluidGravity = v;
}

void MainWindow::TransferFunctionIndexChanged(int val)
{
	m_mode = val;
	emit valuesMode(val);
}

void MainWindow::SelectObjects(bool v)
{
	emit setActionObjectType(TransferPoints::Selection);
}

void MainWindow::DrawPoints(bool v)
{
	emit setActionObjectType(TransferPoints::DrawPoint);
}

void MainWindow::DrawRectangles(bool v)
{
	emit setActionObjectType(TransferPoints::DrawRectangle);
}

void MainWindow::DrawEllipse(bool v)
{
	emit setActionObjectType(TransferPoints::DrawEllipse);
}

void MainWindow::numberClusterChanged(int val)
{
	m_num_cluster = val;
}

void MainWindow::doClustering(bool val)
{
	std::cout << "doClustering" << std::endl;
	std::cout << "doClustering" << std::endl;
	DoingClustering(m_num_cluster);
}

void MainWindow::particleSpeedSlider(int val)
{
	RenderStates::ParticleSpeed = (float)(val)/100.f;
}

void MainWindow::particleSpeed(double val)
{
	RenderStates::ParticleSpeed = val;
}

void MainWindow::particleSize(int val)
{
	RenderStates::ParticleSize = (float)(val)/5000.f;
}

void MainWindow::SetNumberOfParticles(int val)
{
	//std::cout << "SET NUMBER OF PARTICLES: " << val << std::endl;
	//RenderStates::NumberOfParticles = val;
}

void MainWindow::Reset(bool v)
{

}

void MainWindow::Play(bool v)
{
	RenderStates::Play = v;
}

void MainWindow::HdrEnable(bool val)
{
	if(val) RenderStates::HDREnable = 1;
	else	RenderStates::HDREnable = 0;
}

void MainWindow::HDRLinear(bool val)
{
	if(val) RenderStates::HDRMode = 0;
}

void MainWindow::HDRReinhard(bool val)
{
	if(val) RenderStates::HDRMode = 1;
}

void MainWindow::HDRReinhardModified(bool val)
{
	if(val) RenderStates::HDRMode = 2;
}

void MainWindow::HDRAdaptivLog(bool val)
{
	if(val) RenderStates::HDRMode = 3;
}

void MainWindow::StandardModus(bool val)
{
	if(val) RenderStates::RenderMode = 0;
}

void MainWindow::ContourModus(bool val)
{
	if(val) RenderStates::RenderMode = 1;
}

void MainWindow::ExposureModus(bool val)
{
	if(val) RenderStates::RenderMode = 2;
}

void MainWindow::curvatureEnabled(bool val)
{
	RenderStates::ParticleCurvatureEnable = val;
	if(val == false){checkBox_curvatur->setEnabled(false);}
	else {checkBox_curvatur->setEnabled(true);}//disable radio button 8 otherwise enable
}
void MainWindow::CurvaturEnabled2(bool val)
{
	if(val == true){
		RenderStates::ParticleCurvatureEnable = false;
		RenderStates::ParticleCurvature2Enable = true;
	}else{
		RenderStates::ParticleCurvatureEnable = true;
		RenderStates::ParticleCurvature2Enable = false;
	}
}

void MainWindow::IsoSurfaceChanged(int val)
{
	RenderStates::IsoSurface = (float)(val)/100.f;
	emit SetIsoSurfaceChanged(RenderStates::IsoSurface);
}