#pragma once
#include <glm\glm.hpp>
#include <boost/shared_ptr.hpp>

class Material
{
public:
	typedef boost::shared_ptr<Material> MaterialPtr;

public:
	/**
	 * Standardconstructor
	 **/
	Material();
	/**
	 * Constructor
	 * @param emissive emissive value
	 * @param ambient ambient value
	 * @param diffuse diffuse value
	 * @param specular specular value
	 * @param shininess shininess value
	 **/
	Material(glm::vec4 emissive, glm::vec4 ambient, glm::vec4 diffuse, glm::vec4 specular, float shininess);
	~Material();

	/**
	 * This method sets the emissive value
	 * @param emissive emissive value
	 */
	void SetEmissive(glm::vec4 emissive);
	/**
	 * This method sets the ambinet value
	 * @param ambinet ambinet value
	 */
	void SetAmbient(glm::vec4 ambient);
	/**
	 * This method sets the diffuse value
	 * @param diffuse diffuse value
	 */
	void SetDiffuse(glm::vec4 diffuse);
	/**
	 * This method sets the specular value
	 * @param specular specular value
	 */
	void SetSpecular(glm::vec4 specular);
	/**
	 * This method sets the shininess value
	 * @param shininess shininess value
	 */
	void SetShininess(float shininess);

	/**
	 * This method returns the emissive value of the material
	 * @return emissive value
	 */
	glm::vec4 GetEmissive();
	/**
	 * This method returns the ambient value of the material
	 * @return ambient value
	 */
	glm::vec4 GetAmbient();
	/**
	 * This method returns the diffuse value of the material
	 * @return diffuse value
	 */
	glm::vec4 GetDiffuse();
	/**
	 * This method returns the specular value of the material
	 * @return specular value
	 */
	glm::vec4 GetSpecular();
	/**
	 * This method returns the shininess value of the material
	 * @return shininess value
	 */
	float GetShininess();

private:
	glm::vec4 m_emissive; // default: ColorRGBA(0,0,0,1)
	glm::vec4 m_ambient; // default: ColorRGBA(0.2,0.2,0.2,1)
	glm::vec4 m_diffuse; // default: ColorRGBA(0.8,0.8,0.8,1)
	glm::vec4 m_specular; // default: ColorRGBA(0,0,0,1)
	float m_shininess; // default: 1
};

