#include "MyRenderer.h"


MyRenderer::MyRenderer(void) : m_clear_depth_buffer(true)
{
}


MyRenderer::~MyRenderer(void)
{
}

void MyRenderer::DrawView(View* node)
{
	this->m_view = node;
	glViewport(0,0,m_view->GetCamera()->GetWidth(), m_view->GetCamera()->GetHeight());

	if(m_view->GetFBO() != 0 && m_view->GetFBO()->GetTexName() > 0)  
	{	
		m_view->GetFBO()->Bind();	//ACHTUNG

		if(m_view->GetFBO()->m_dimension != 3)
		{
			//if(m_view->GetFBO()->m_pingbong == false || m_view->GetFBO()->m_dimension != 2)
			{
				if(m_clear_depth_buffer)
				{
					ClearColorBuffer();
					ClearDepthBuffer();
				}
			}
		}

		m_view->GetFBO()->DrawBuffers();
	}
	else
	{
		if(m_clear_depth_buffer)
		{
			ClearColorBuffer();
			ClearDepthBuffer();
		}
	}

	TraverseSceneGraph(m_view->GetRoot(), glm::mat4());

	if(m_view->GetFBO() != 0 && m_view->GetFBO()->GetTexName() > 0)
	{
		if(m_view->GetFBO()->m_dimension == 2)
		{
			m_view->GetFBO()->Unbind();
			
			if(m_view->GetFBO()->m_pingbong) 
			{
				if(m_view->GetFBO()->current_texture == 1)	glBindTexture(GL_TEXTURE_2D, m_view->GetFBO()->GetTexName());
				else										glBindTexture(GL_TEXTURE_2D, m_view->GetFBO()->GetTexName2());
			}
			else	glBindTexture(GL_TEXTURE_2D, m_view->GetFBO()->GetTexName());

			glGenerateMipmap(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, 0);
		}

		if(m_view->GetFBO()->m_dimension == 3)
		{
			m_view->GetFBO()->Unbind();
			glBindTexture(GL_TEXTURE_3D, 0);
		}

		if(m_view->GetFBO()->m_dimension == 4)
		{
			m_view->GetFBO()->Unbind();
			glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
		}
	}
}

void MyRenderer::TraverseSceneGraph(INode* inode, glm::mat4 m)
{
	if(inode->GetGeometry() != 0)
	{
		m_pkGeometry = inode->GetGeometry();
		//m = m * m_pkGeometry->World;
		m = m_pkGeometry->World;

		m_modelMatrix = m*m_pkGeometry->Local;
		//if(Culling(m_modelMatrix))
		{
			BindShader();
			SetUniformsObject();
			EnableTextures();
			DrawGeometry();
			DisableTextures();
			m_pkGeometry->GetShader(m_view->GetPass())->unbind();

			for(int i = 0; i < inode->GetUsed(); i++)
			{
				TraverseSceneGraph(inode->GetChild(i), m);
			}
		}
	}
}


void MyRenderer::EnableTextures()
{
	for(std::map<unsigned int, Texture::TexturePtr>::iterator it = m_pkGeometry->m_TextureObjs.begin(); it != m_pkGeometry->m_TextureObjs.end(); it++)
	{
		/*if(m_view->GetFBO() == 0 || m_view->GetFBO()->GetTexName() != it->second->GetTexName() )*/ EnableTexture(it->second, it->first);
	}
}

void MyRenderer::DisableTextures()
{
	for(std::map<unsigned int, Texture::TexturePtr>::iterator it = m_pkGeometry->m_TextureObjs.begin(); it != m_pkGeometry->m_TextureObjs.end(); it++)
	{
		/*if(m_view->GetFBO() == 0 || m_view->GetFBO()->GetTexName() != it->second->GetTexName())*/ DisableTexture(it->second, it->first);
	}
}

void MyRenderer::BindShader()
{
	m_pkGeometry->GetShader(m_view->GetPass())->bind();
	SetUniformsGlobal();
}

void MyRenderer::SetUniformsGlobal()
{
	std::map<int, int> dict = m_pkGeometry->GetDictShaderUniform(m_view->GetPass());
	for(std::map<int, int>::iterator it = dict.begin(); it != dict.end(); it++)
	{
		switch(it->second)
		{
		case View::UT_IDENTITY: glUniformMatrix4fv( it->first, 1, GL_FALSE, glm::value_ptr(glm::mat4()));
			break;
		case View::UT_TEXTURE0:	glUniform1i(it->first, 0);
			break;
		case View::UT_TEXTURE1:	glUniform1i(it->first, 1);
			break;
		case View::UT_TEXTURE2:	glUniform1i(it->first, 2);
			break;
		case View::UT_TEXTURE3:	glUniform1i(it->first, 3);
			break;
		case View::UT_TEXTURE4:	glUniform1i(it->first, 4);
			break;
		case View::UT_TEXTURE5:	glUniform1i(it->first, 5);
			break;
		case View::UT_TEXTURE6:	glUniform1i(it->first, 6);
			break;
		case View::UT_TEXTURE7:	glUniform1i(it->first, 7);
			break;
		case View::UT_PROJECTIONMATRIX:	glUniformMatrix4fv( it->first, 1, GL_FALSE, glm::value_ptr(m_view->GetCamera()->GetProjection()));
			break;
		case View::UT_VIEWMATRIX:	glUniformMatrix4fv( it->first, 1, GL_FALSE, glm::value_ptr(m_view->GetCamera()->GetViewMatrix()));
			break;
		case View::UT_OLDPROJECTIONMATRIX:	glUniformMatrix4fv( it->first, 1, GL_FALSE, glm::value_ptr(m_view->GetCamera()->GetOldProjection()));
			break;
		case View::UT_OLDVIEWMATRIX:	glUniformMatrix4fv( it->first, 1, GL_FALSE, glm::value_ptr(m_view->GetCamera()->GetOldViewMatrix()));
			break;
		case View::UT_LIGHT_POSITION:	glUniform3f( it->first, m_view->GetLight()->GetPosition().x, m_view->GetLight()->GetPosition().y, m_view->GetLight()->GetPosition().z);
			break;
		case View::UT_LIGHT_AMBIENT:	glUniform4f( it->first, m_view->GetLight()->GetAmbient().x, m_view->GetLight()->GetAmbient().y, m_view->GetLight()->GetAmbient().z, m_view->GetLight()->GetAmbient().w);
			break;
		case View::UT_LIGHT_DIFFUSE:	glUniform4f( it->first, m_view->GetLight()->GetDiffuse().x, m_view->GetLight()->GetDiffuse().y, m_view->GetLight()->GetDiffuse().z, m_view->GetLight()->GetDiffuse().w);
			break;
		case View::UT_LIGHT_SPECULAR:	glUniform4f( it->first, m_view->GetLight()->GetSpecular().x, m_view->GetLight()->GetSpecular().y, m_view->GetLight()->GetSpecular().z, m_view->GetLight()->GetSpecular().w);
			break;
		case View::UT_LIGHT_CONSTANTATTENUATION:	glUniform1f(it->first, m_view->GetLight()->GetConstantAttenuation());
			break;
		case View::UT_LIGHT_LINEARATTENUATION:	glUniform1f(it->first, m_view->GetLight()->GetLinearAttenuation());
			break;
		case View::UT_LIGHT_QUADRATICATTENUATION:	glUniform1f(it->first, m_view->GetLight()->GetQuadraticAttenuation());
			break;
		case View::UT_LIGHT_SHINENESS:	glUniform1f(it->first, m_view->GetLight()->GetShineness());
			break;
		case View::UT_VIEWPORT_WIDTH:	glUniform1i(it->first, m_view->GetCamera()->GetWidth());
			break;
		case View::UT_VIEWPORT_HEIGHT:	glUniform1i(it->first, m_view->GetCamera()->GetHeight());
			break;
		case View::UT_VIEWPORT: glUniform2i(it->first, m_view->GetCamera()->GetWidth(), m_view->GetCamera()->GetHeight());
			break;
		case View::UT_FBO_SIZE_3D: 	glUniform3f(it->first, 64.f, 64.f,64.f);
			break;
		case View::UT_TIME_DURATION: glUniform1f(it->first, (float)this->m_duration);
			break;
		case View::UT_EYE_POSITION: glUniform3f(it->first, m_view->GetCamera()->GetPosition().x, m_view->GetCamera()->GetPosition().y, m_view->GetCamera()->GetPosition().z);
			break;
		case View::UT_VOLUMESIZE: glUniform3f(it->first, m_view->GetGrid().x, m_view->GetGrid().y, m_view->GetGrid().z);
			break;
		case View::UT_RENDERMODE: glUniform1i(it->first, RenderStates::RenderMode);
			break;
		case View::UT_ADDAPTIVE: glUniform1i(it->first, RenderStates::Addaptive);
			break;
		case View::UT_STEPSIZE: glUniform1f(it->first, RenderStates::Stepsize);
			break;
		case View::UT_COMPOSITING_MODE: glUniform1i(it->first, RenderStates::Compositing);
			break;
		case View::UT_MIP: glUniform1i(it->first, RenderStates::MIP);
			break;
		case View::UT_TRANSFER: glUniform1i(it->first, RenderStates::Transfer);
			break;
		case View::UT_AVERAGE: glUniform1i(it->first, RenderStates::Average);
			break;
		case View::UT_LIGHT: glUniform1i(it->first, RenderStates::Light);
			break;
		case View::UT_PHONG: glUniform1i(it->first, RenderStates::Phong);
			break;
		case View::UT_PARTICLE_SIZE: glUniform1f(it->first, RenderStates::ParticleSize);
			break;
		case View::UT_HDR_ENABLE: glUniform1i(it->first, RenderStates::HDREnable);
			break;
		case View::UT_HDR_MODE: glUniform1i(it->first, RenderStates::HDRMode);
			break;
		
		}
	}
}

void MyRenderer::SetUniformsObject()
{
	if(m_pkGeometry == 0) return;
	
	std::map<int, int> dict = m_pkGeometry->GetDictShaderUniform(m_view->GetPass());
	for(std::map<int, int>::iterator it = dict.begin(); it != dict.end(); it++)
	{
		switch(it->second)
		{
		case View::UT_MODELMATRIX:	glUniformMatrix4fv( it->first, 1, GL_FALSE, glm::value_ptr(m_modelMatrix));
			break;
		case View::UT_MODELVIEWPROJECTIONMATRIX:	glUniformMatrix4fv( it->first, 1, GL_FALSE, glm::value_ptr(m_view->GetCamera()->GetProjection()*m_view->GetCamera()->GetViewMatrix()*m_modelMatrix));
			break;
		case View::UT_MODELVIEWPROJECTIONLIGHTMATRIX:	
			glUniformMatrix4fv( it->first, 1, GL_FALSE, glm::value_ptr(m_view->GetLight()->GetCamera()->GetProjection() * m_view->GetLight()->GetCamera()->GetViewMatrix() * m_modelMatrix));
			break;
		case View::UT_WORLDINVERSETRANSPOSE3x3:
			glUniformMatrix3fv( it->first, 1, GL_FALSE, glm::value_ptr(glm::mat3((glm::transpose(glm::inverse(m_modelMatrix))) )));
			break;
		case View::UT_MATERIAL_EMISSIVE:	
			glUniform4f( it->first, m_pkGeometry->m_MatrialObjs->GetEmissive().x, m_pkGeometry->m_MatrialObjs->GetEmissive().y, m_pkGeometry->m_MatrialObjs->GetEmissive().z, m_pkGeometry->m_MatrialObjs->GetEmissive().w);
			break;
		case View::UT_MATERIAL_AMBIENT:		
			if(m_pkGeometry->m_MatrialObjs != 0)	
			{
				glUniform4f( it->first, m_pkGeometry->m_MatrialObjs->GetAmbient().x, m_pkGeometry->m_MatrialObjs->GetAmbient().y, m_pkGeometry->m_MatrialObjs->GetAmbient().z, m_pkGeometry->m_MatrialObjs->GetAmbient().w);
			}
			else									glUniform4f( it->first, 0.1f, 0.1f, 0.1f, 1.f);
			break;
		case View::UT_MATERIAL_DIFFUSE:	
			if(m_pkGeometry->m_MatrialObjs != 0)	
			{
				glUniform4f( it->first, m_pkGeometry->m_MatrialObjs->GetDiffuse().x, m_pkGeometry->m_MatrialObjs->GetDiffuse().y, m_pkGeometry->m_MatrialObjs->GetDiffuse().z, m_pkGeometry->m_MatrialObjs->GetDiffuse().w);				
			}
			else									glUniform4f( it->first, 0.9f, 0.9f, 0.9f, 1.f);
			break;
		case View::UT_MATERIAL_SPECULAR:	
			if(m_pkGeometry->m_MatrialObjs != 0)	glUniform4f( it->first, m_pkGeometry->m_MatrialObjs->GetSpecular().x, m_pkGeometry->m_MatrialObjs->GetSpecular().y, m_pkGeometry->m_MatrialObjs->GetSpecular().z, m_pkGeometry->m_MatrialObjs->GetSpecular().w);
			else									glUniform4f( it->first, 0.9f, 0.9f, 0.9f, 1.f);
			break;
		case View::UT_MATERIAL_SHININESS:	
			glUniform1f( it->first, m_pkGeometry->m_MatrialObjs->GetShininess());
			break;
		case View::UT_BONES:	
			if(m_pkGeometry->m_bones_count > 0 && m_pkGeometry->bones != 0)
			{
				glUniformMatrix4fv(it->first, m_pkGeometry->m_bones_count, GL_FALSE, glm::value_ptr(m_pkGeometry->bones[0]));
			}
			break;
		case View::UT_OLDBONES:	
			if(m_pkGeometry->m_bones_count > 0 && m_pkGeometry->oldBones != 0)
			{
				glUniformMatrix4fv(it->first, m_pkGeometry->m_bones_count, GL_FALSE, glm::value_ptr(m_pkGeometry->oldBones[0]));
			}
			break;

		}
	}

	this->GetErrors();
}
