#pragma once
#include "utility.hpp"
#include "FBO.h"
#include "VertexGeometry.h"
#include "View.h"
#include "RenderStates.h"

#include <glm/gtx/transform.hpp> 
#include <glm/gtc/type_ptr.hpp> 
#include <glm\gtc\matrix_projection.hpp>
#include <glm\gtc\matrix_transform.hpp>


class MyRenderer
{
public:
	/**
	 * Standardconstructor
	 **/
	MyRenderer(void);
	/**
	 * Standarddeconstructor
	 **/
	~MyRenderer(void);

public:
	//Bei Inline Operationen wird einfach der Code im Aufrufteil kopiert, sollte etwas bei der Performance bringen, dafr ist das Programm grer...
	/**
	 * This method clears the color buffer
	 **/
	inline virtual void ClearColorBuffer() = 0;
	/**
	 * This method clears the depth buffer
	 **/
	inline virtual void ClearDepthBuffer() = 0;
	/**
	 * This method clears the stencil buffer
	 **/
	inline virtual void ClearStencilBuffer() = 0;
	/**
	 * This method clears the accumulation buffer
	 **/
	inline virtual void ClearAccumBuffer() = 0;
	/**
	 * This method sets the clear color
	 * @param color clear color
	 **/
	inline virtual void SetClearColor(glm::vec4 color) = 0;

	/**
	 * This method swaps the buffer
	 **/
	inline virtual void SwapBuffer() = 0;
	/**
	 * This method returns the error
	 **/
	inline virtual void GetErrors() = 0;

	/**
	 * This method sets the culling mode to FRONT
	 **/
	inline virtual void SetFrontCulling()=0;
	/**
	 * This method sets the culling mode to BACK
	 **/
	inline virtual void SetBackCulling()=0;
	/**
	 * This method enables Backfaceculling
	 **/
	inline virtual void EnableBackFaceCulling()=0;
	/**
	 * This method disables Backfaceculling
	 **/
	inline virtual void DisableBackFaceCulling()=0;

	/**
	 * This method sets the uniforms of the shader
	 */
	void SetUniformsObject();
	/**
	 * This method binds a shader
	 */
	void BindShader();
	/**
	 * This method sets the global uniforms of the shader
	 */
	void SetUniformsGlobal();

	/**
	* This methods enables all textures in the view
	*/
	void EnableTextures();
	/**
	* This methods disables all textures in the view
	*/
	void DisableTextures();

	/**
	 * This method draws the view
	 */
	void DrawView(View* node);

	bool m_clear_depth_buffer;

protected:
	/**
	 * This method enables a texture
	 * @param texture Texture to enable
	 * @param i position in texturestack
	 **/
	inline virtual void EnableTexture(Texture::TexturePtr texture, int i) = 0;
	/**
	 * This method disable a texture
	 * @param texture Texture to enable
	 * @param i position in texturestack
	 **/
	inline virtual void DisableTexture(Texture::TexturePtr texture, int i) = 0;

	/**
	 * This Method draw the geometry object
	 */
	virtual void DrawVertexGeometry() = 0;

	/**
	 * This Method draws the object
	 */
	virtual void DrawGeometry() = 0;
	
	/**
	 * This method traverese the scene graph
	 * @param inode actual node in the rendering graph
	 * @param m actual matrix
	**/
	void MyRenderer::TraverseSceneGraph(INode* inode, glm::mat4 m);

	glm::mat4 m_modelMatrix;

	float m_duration;

	View* m_view;
	Geometry* m_pkGeometry;
};

