#include <fstream>

#include "MyShader.hpp"

MyShader::MyShader(const string &path) :
_success(false), shadername(path)
{
	// Load the shader files
	std::cout << "Create: " << path << std::endl;

	string vertex_shader_source;
	if (file_exists(path+".vert")) {
		vertex_shader_source = read_file(path+".vert");
		std::cout << "Vertex MyShader founded" << std::endl;
	} else {
		cerr << "Vertex shader file "
			<< path <<".vert does not exist." << endl;
		return;
	}

	string geometry_shader_source;
	if (file_exists(path+".geom")) {
		geometry_shader_source = read_file(path+".geom");
		std::cout << "Geometry MyShader founded" << std::endl;
		m_geometry = true;
	} else {
		m_geometry = false;
		cerr << "Geometry shader file "
			<< path <<".geom does not exist." << endl;
	}

	string fragment_shader_source;
	if (file_exists(path+".frag")) {
		fragment_shader_source = read_file(path+".frag");
		std::cout << "Fragment MyShader founded" << std::endl;
	} else {
		cerr << "Fragment shader file "
			<< path <<".frag does not exist." << endl;
		return;
	}

	// Compile the shaders

	_vertex_shader = compile(GL_VERTEX_SHADER, vertex_shader_source);
	if (_vertex_shader == 0)
		return;

	get_errors();
	std::cout << "Vertex MyShader compiled" << std::endl;

	if(m_geometry)
	{
		_geometry_shader = compile(GL_GEOMETRY_SHADER, geometry_shader_source);
		if (_geometry_shader == 0)
			return;

		get_errors();
		std::cout << "Geometry MyShader compiled" << std::endl;
	}

	_fragment_shader = compile(GL_FRAGMENT_SHADER, fragment_shader_source);
	if (_fragment_shader == 0)
		return;

	get_errors();

	std::cout << "Fragment MyShader compiled" << std::endl;

	// Link the shaders into a program

	link();
	std::cout << "MyShader linked" << std::endl;

	if (_program == 0)
		return;

	_success = true;
	get_errors();
}

MyShader::~MyShader()
{
	glDeleteProgram(_program);
	glDeleteShader(_vertex_shader);
	glDeleteShader(_fragment_shader);
}

GLuint MyShader::compile (GLenum type, const string &source)
{
	// Create shader object

	GLuint shader = glCreateShader(type);

	if (shader == 0) {
		cerr << "Could not create shader object." << endl;
		return 0;
	}

	// Define shader source and compile

	const char* src = source.data();
	int len = source.size();

	glShaderSource(shader, 1, &src, &len);

	glCompileShader(shader);

	// Check for errors

	int status;

	glGetShaderiv(shader, GL_COMPILE_STATUS, &status);

	if (status != GL_TRUE) {
		cout << "MyShader compilation failed." << endl;
		shader_log(shader);
	}

	get_errors();

	return shader;    
}

void MyShader::link(void)
{
	// Create program handle
	_program = glCreateProgram();

	// Attach shaders and link

	glAttachShader(_program, _vertex_shader);
	if(m_geometry)
	{
		glAttachShader(_program, _geometry_shader);
	}
	glAttachShader(_program, _fragment_shader);


	if(m_geometry)
	{
		std::cout << "glProgramParameteri start" << std::endl;
		glProgramParameteri(_program, GL_GEOMETRY_INPUT_TYPE, GL_POINTS);
        glProgramParameteri(_program, GL_GEOMETRY_OUTPUT_TYPE, GL_TRIANGLE_STRIP); 
        glProgramParameteri(_program, GL_GEOMETRY_VERTICES_OUT, 4); 
		std::cout << "glProgramParameteri end" << std::endl;
	}

	glLinkProgram(_program);

	// Check for problems

	int status;

	glGetProgramiv(_program, GL_LINK_STATUS, &status);

	if (status != GL_TRUE) {
		cout << "MyShader linking failed." << endl;
		program_log(_program);

		glDeleteProgram(_program);
		_program = 0;
	}

	get_errors();
}

#define LOG_BUFFER_SIZE 8096

void MyShader::program_log(GLuint program)
{
	char logBuffer[LOG_BUFFER_SIZE];
	GLsizei length;

	logBuffer[0] = '\0';
	glGetProgramInfoLog(program, LOG_BUFFER_SIZE, &length,logBuffer);

	if (length > 0) {
		cout << logBuffer << endl;
	}
};

void MyShader::shader_log(GLuint shader)
{
	char logBuffer[LOG_BUFFER_SIZE];
	GLsizei length;

	logBuffer[0] = '\0';
	glGetShaderInfoLog(shader, LOG_BUFFER_SIZE, &length,logBuffer);

	if (length > 0) {
		cout << logBuffer << endl;
	}
};
