#pragma once

#include <GL/glew.h>
#include <QWidget>
#include <QGLWidget>
//#include <QtOpenGL>
#include <QGLContext>
#include <QGLFormat>

#include "GLM/glm.hpp"
#include "Shader.h"
#include "TransferFunctionWidget.h"

class OpenGLColorgradientWidget : public QGLWidget
{
	Q_OBJECT

public:
	/** The constructor. */
	OpenGLColorgradientWidget(QWidget* parent = 0);
	/** The destructor. */
	~OpenGLColorgradientWidget();

public slots:
	/**
	 * Generates a 1D texture for the gives RGBA values.
	 * @param rgbaValues The RGBA values for the texture.
	 * @param textureSize The size of the texture.
	 */
	void generateTransferTexture(float* rgbaValues, int textureSize, int type);
	/** Deletes the current transfer texture. */
	void resetTransferTexture();

	void setMode(int mode);

protected:
	/** Initialize GLEW and sets up an OpenGL 3.2 context. */
	void initializeGL();
	/** 
	 * Sets the new viewport and calculates the vertices from the quad. 
	 * @param width The new width of the window, measured in pixel.
	 * @param height The new height of the window, measured in pixel.
	 */
	void resizeGL(int width,int height);
	/** 
	 * Renders the transfer texture. 
	 */
	void paintGL();

private:
	/** The texturehandle vor the transfer function texture*/
	GLuint textureHandle;
	/** The shader program. */
	Shader* shader;
	/** The projection matrix. */
	glm::mat4 projectionMatrix;
	/** The width of the screen. */
	int screenWidth;
	/** The height of the screen. */
	int screenHeight;
	/** The vertices. */
	glm::vec4* vertices;
	/** True if the transfer texture is loaded.*/
	bool textureLoaded;
	/** Calculates the vertices. */
	void calculateVertexCoord();
	/** Basic OpenGL Error-Handling function. If encountered an error it prints an error-message to the console.*/
	void checkGLError();

	GLuint textureParticleHandle;
	int m_mode;
};