// local includes
#include "OpenGLWidget.h"

// system includes
#include <fstream>
#include <iostream>

#include "jama_eig.h"

#include "GLM/gtx.hpp"
#include "GLM/gtc.hpp"
#include "GLM/gtx/type_ptr.hpp"

using std::cout;
using std::endl;
using std::cerr;

// global variables
//ggf anpassen!!!
float nearplane = 0.1f;
float farplane = 100.0f;
//field of view angle, in degrees, in the y direction
float fovAngle = 20.0f;

GLuint slicesTextureHandle;
GLuint pcdTextureHandle;

OpenGLWidget::OpenGLWidget(QWidget* parent) : QGLWidget(parent), perspective(false), moveVolume(false), zoomVolume(false), isStarted(false)
{
	this->slice = new Slicing();
	this->rayCasting = new RayCasting();
}

OpenGLWidget::~OpenGLWidget()
{
	glDeleteTextures(1,&transferFunctionTextureHandle);
	delete slice;
	delete rayCasting;
}

void OpenGLWidget::setPerspective(int tabIndex)
{
	makeCurrent();
	if (tabIndex == 0) {
		this->perspective = false;
	} else {
		this->perspective = true;
	}
	this->setCameraProjection();
	updateGL();
}

glm::vec4 GetCurvature(float* vol, int x0, int y0, int z0, 
									unsigned int iTextureWidth, unsigned int iTextureHeight, unsigned int iTextureDepth ){

	//CENTRAL DIFFERENCING

	// 6 points central + offsets in xyz
	// central = x0,y0,z0

	int point_index = 4 * (x0 + y0 * iTextureWidth + z0 *iTextureWidth * iTextureHeight);
	int offsetX = 4;
	int offsetY = iTextureWidth * 4;
	int offsetZ = iTextureWidth * iTextureHeight * 4;

	int offsetX_left = offsetX;
	int offsetX_right = offsetX;

	int offsetY_left = offsetY;
	int offsetY_right = offsetY;

	int offsetZ_left = offsetZ;
	int offsetZ_right = offsetZ;

	float x_scale = 0.5f;
	float y_scale = 0.5f;
	float z_scale = 0.5f;
	
	float treshhold = 0.01f;
	
	if(vol[point_index + 3] < treshhold)
	{
		//skip the calculatio
		return glm::vec4(0.0f, 0.0f, 0.0f, 0.0f);

	} 
	else 
	{

		//check if point is at a Border 
		if(x0 == 0) {offsetX_left = 0; x_scale = 1.0f;}
		else if(x0 == iTextureWidth-1) {offsetX_right = 0; x_scale = 1.0f;}

		if(y0 == 0) {offsetY_left = 0; y_scale = 1.0f;}
		else if(y0 == iTextureHeight-1) {offsetY_right = 0; y_scale = 1.0f;}

		if(z0 == 0) {offsetZ_left = 0; z_scale = 1.0f;}
		else if(z0 == iTextureDepth-1) {offsetZ_right = 0; z_scale = 1.0f;}
	

		//(pointx1 - pointx2 )*0.5f first row hessian
		glm::vec3 pointx1 = glm::vec3(
			vol[point_index + offsetX_right + 0],
			vol[point_index + offsetX_right + 1],
			vol[point_index + offsetX_right + 2]
		);
		glm::vec3 pointx2 = glm::vec3(
			vol[point_index - offsetX_left + 0],
			vol[point_index - offsetX_left + 1],
			vol[point_index - offsetX_left + 2]
		);
		//(pointy1 - pointy2)*0.5f second row hessian
		glm::vec3 pointy1 = glm::vec3(
			vol[point_index + offsetY_right + 0],
			vol[point_index + offsetY_right + 1],
			vol[point_index + offsetY_right + 2]
		);
		glm::vec3 pointy2 = glm::vec3(
			vol[point_index - offsetY_left + 0],
			vol[point_index - offsetY_left + 1],
			vol[point_index - offsetY_left + 2]
		);
		//(pointz1 - pointz2)*0.5f third row hessian
		glm::vec3 pointz1 = glm::vec3(
			vol[point_index + offsetZ_right + 0],
			vol[point_index + offsetZ_right + 1],
			vol[point_index + offsetZ_right + 2]
		);
		glm::vec3 pointz2 = glm::vec3(
			vol[point_index - offsetZ_left + 0],
			vol[point_index - offsetZ_left + 1],
			vol[point_index - offsetZ_left + 2]
		);


		pointx1 = (pointx1 - pointx2 ) * x_scale;
		pointy1 = (pointy1 - pointy2 ) * y_scale;
		pointz1 = (pointz1 - pointz2 ) * z_scale;

		//MatrixSolver Eigenvalue + eigenvector calculation with Jama

		const int N = 3;


		TNT::Array2D< double > A( N, N, 0.0 );   /* create an NxN array; all zeros */
	
		//Zeile/Spalte
		A[0][0] = pointx1.x;
		A[0][1] = pointx1.y;
		A[0][2] = pointx1.z;
		A[1][0] = pointy1.x;
		A[1][1] = pointy1.y;
		A[1][2] = pointy1.z;
		A[2][0] = pointz1.x;
		A[2][1] = pointz1.y;
		A[2][2] = pointz1.z;

		JAMA::Eigenvalue<double> eigenvalues = JAMA::Eigenvalue<double>( A );

		TNT::Array1D<double> eigs_real;
		eigenvalues.getRealEigenvalues( eigs_real );

		int index = 0;
		for (int i = 1; i < eigs_real.dim(); i++)
		{
			if(eigs_real[i] > eigs_real[i-1]) index = i;
		}

		// eigenvecs in der spalte
		TNT::Array2D<double> eig_vecs( N, N, 0.0 );

		eigenvalues.getV( eig_vecs );

		return glm::vec4(eig_vecs[0][index], eig_vecs[1][index], eig_vecs[2][index], eigs_real[index]);
		
	}
}

glm::vec4 GetGradient(Volume* vol,	int x0, int y0, int z0, 
									int x1, int y1, int z1, 
									int x2, int y2, int z2, 
									unsigned int iTextureWidth, unsigned int iTextureHeight, unsigned int iTextureDepth)
{
	double sum_wkxk2 = 0.0f;
	double sum_wkxkyk = 0.0f;
	double sum_wkxkzk = 0.0f;

	double sum_wkyk2 = 0.0f;
	double sum_wkykxk = 0.0f;
	double sum_wkykzk = 0.0f;
	
	double sum_wkzk2 = 0.0f;
	double sum_wkzkxk = 0.0f;
	double sum_wkzkyk = 0.0f;
	
	double sum_wk = 0.0f;
	double sum_wkxk = 0.0f;
	double sum_wkyk = 0.0f;
	double sum_wkzk = 0.0f;

	double sumA = 0.0f;
	double sumB = 0.0f;
	double sumC = 0.0f;
	double sumD = 0.0f;

	double A = 0.0f;
	double B = 0.0f;
	double C = 0.0f;
	double D = 0.0f;

	glm::vec3 pos(x0, y0, z0);
	
	int neigboorhood = 1; //STEUERT DIE GR?E DES FILTERS
	for(int x = -1*neigboorhood; x <= neigboorhood; x++)
	{
		for(int y = -1*neigboorhood; y <= neigboorhood; y++)
		{
			for(int z = -1*neigboorhood; z <= neigboorhood; z++)
			{
				if(x != 0 || y != 0 || z != 0) 
				{
					//GAUSS FILTERUNG oder so!!!
					//IRGENDEIN PAPER VON DER TU...
					int X = (int)(pos.x+x);
					int Y = (int)(pos.y+y);
					int Z = (int)(pos.z+z);

					if(X < 0) X = 0;
					if(Y < 0) Y = 0;
					if(Z < 0) Z = 0;
					if(X >= vol->GetWidth())	X = vol->GetWidth()-1;
					if(Y >= vol->GetHeight())	Y = vol->GetHeight()-1;
					if(Z >= vol->GetDepth())	Z = vol->GetDepth()-1;

					double fk = vol->Get(X,Y,Z).GetValue();
					//WICHTIG!!!!!!!!!!!!!!!!
					double xk = x;
					double yk = y;
					double zk = z;
					glm::vec3 diff = glm::vec3(0,0,0)-glm::vec3(xk,yk,zk);

					double wk = 0.0;
					wk=1.0/glm::length(diff);				

					sum_wkxk2  += wk*xk*xk;
					sum_wkxkyk += wk*xk*yk;
					sum_wkxkzk += wk*xk*zk;
					sum_wkxk   += wk*xk;

					sum_wkykxk += wk*yk*xk;
					sum_wkyk2  += wk*yk*yk;
					sum_wkykzk += wk*yk*zk;
					sum_wkyk   += wk*yk;

					sum_wkzkxk += wk*zk*xk;
					sum_wkzkyk += wk*zk*yk;
					sum_wkzk2  += wk*zk*zk;
					sum_wkzk   += wk*zk;

					sum_wk += wk;
				
					sumA += (wk*fk*xk);
					sumB += (wk*fk*yk);
					sumC += (wk*fk*zk);
					sumD += (wk*fk);
				}
			}
		}
	}

	A = sumA / sum_wkxk2;
	B = sumB / sum_wkyk2;
	C = sumC / sum_wkzk2;
	D = sumD / sum_wk;
	
	return glm::vec4(A,B,C,D);
	/*/
	//CENTRAL DIFFERENCING
	if(x2 < 0) x2 = 0;
	if(y2 < 0) y2 = 0;
	if(z2 < 0) z2 = 0;
	
	if(x1 >= iTextureWidth)		x1 = iTextureWidth-1;
	if(y1 >= iTextureHeight)	y1 = iTextureHeight-1;
	if(z1 >= iTextureDepth)		z1 = iTextureDepth-1;

	//http://www.cg.tuwien.ac.at/~theussl/DA/node40.html
	glm::vec3 vol1 = glm::vec3(
		vol->Get(x1, y0, z0).GetValue(),
		vol->Get(x0, y1, z0).GetValue(),
		vol->Get(x0, y0, z1).GetValue()
	);

	glm::vec3 vol2 = glm::vec3(
		vol->Get(x2, y0, z0).GetValue(),
		vol->Get(x0, y2, z0).GetValue(),
		vol->Get(x0, y0, z2).GetValue()
	);

	glm::vec3 grad = ((vol1-vol2)/2.f);
	return glm::vec4(grad.x, grad.y, grad.z, vol->Get(x0, y0, z0).GetValue());
	//*/
}

void OpenGLWidget::initSliceTexture()
{
	checkGLError();
	std::cout << "initSliceTexture" << std::endl;

	if (SceneManager::get().getVolume() != 0) 
	{
		makeCurrent();

		unsigned int tf_size = 1024*4;
		float4* tf_tmp = new float4[tf_size];
		for(unsigned int j = 0; j < tf_size; j++)	tf_tmp[j] = make_float4( 1.f, 1.f, 1.f, 1.f);

		m_transferfunction = new CudaTexture<float4,2>(tf_size/1,1,0);	
		initTransferFunctionTexture(m_transferfunction->GetCudaTexture());
		cutilSafeCall(cudaMemcpyToArray(m_transferfunction->GetCudaTexture(), 0, 0, tf_tmp, sizeof(float4)*tf_size, cudaMemcpyHostToDevice));
		rayCasting->m_texture = m_transferfunction;


	/*unsigned int weights_size = 256;
	float *data   = new float[4 * weights_size];
	float *ptr    = data;
	float  phi[4], x3, x2, x = 0.f;
	float  h      = 1.f / (float)(weights_size-1);
	float  c      = 1.f / 6.f;

	for(unsigned int i = 0; i < weights_size; ++i, x += h)
	{
		x3 = (x2 = x * x) * x;
		// cubic spline basis functions
		// non-negative partition of one; in [0, 2/3]
		phi[0] = c * (-x3 + 3*x2 - 3*x + 1.f);
		phi[1] = c * (3.f*x3 - 6.f*x2 + 4.f);
		phi[2] = c * (-3.f*x3 + 3.f*x2 + 3.f*x + 1.f);
		phi[3] = c * (x3);
		// offsets; first with changed sign; in [0, 1]
		*ptr++ = 1.0f + x - phi[1] / (phi[0] + phi[1]);
		*ptr++ = 1.0f - x + phi[3] / (phi[2] + phi[3]);
		// weight for left sample (lerp: right weight = 1-left weight)
		*ptr++ = phi[0] + phi[1];
		*ptr++ = phi[0] + phi[1];
	}

	std::cout << "Create HG Texture: < " << weights_size << std::endl;
	rayCasting->m_hg = new CudaTexture<float4,2>(weights_size,1,0);	
	
	std::cout << "INIT HG & HG_DEV Texture" << std::endl;
	initHGTexture(rayCasting->m_hg->GetCudaTexture());
	BindHGCuda(rayCasting->m_hg->GetCudaTexture());
	cutilSafeCall(cudaMemcpyToArray(rayCasting->m_hg->GetCudaTexture(), 0, 0, data, sizeof(float4)*weights_size, cudaMemcpyHostToDevice));

	delete[] data;
	
	data   = new float[4 * weights_size];
	ptr    = data;

	ptr = data; x = 0.f;
	for(int i = 0; i < weights_size; ++i, x += h)
	{
		x2 = x * x;
		// cubic spline basis functions derivatives
		// 0,1 negative; 2,3 positive; in [-2/3, 2/3]
		phi[0] = c * (-3.f*x2 + 6.f*x - 3.f);
		phi[1] = c * (9.f*x2 - 12.f*x);
		phi[2] = c * (-9.f*x2 + 6.f*x + 3.f);
		phi[3] = c * (3.f*x2);
		// offsets; first with changed sign; in [0, 1]
		*ptr++ = 1.f + x - phi[1] / (phi[0] + phi[1]);
		*ptr++ = 1.f - x + phi[3] / (phi[2] + phi[3]);
		// weight for right sample (central differences: left weight = -right weight)
		*ptr++ = phi[2] + phi[3];
		*ptr++ = phi[2] + phi[3];
	}

	std::cout << "Create HG_DEV Texture" << std::endl;
	rayCasting->m_hg_dev = new CudaTexture<float4,2>(weights_size,1,0);	

	initHGDevTexture(rayCasting->m_hg_dev->GetCudaTexture());
	BindHGDevCuda(rayCasting->m_hg_dev->GetCudaTexture());
	cutilSafeCall(cudaMemcpyToArray(rayCasting->m_hg_dev->GetCudaTexture(), 0, 0, data, sizeof(float4)*weights_size, cudaMemcpyHostToDevice));
	
	std::cout << "Initialized" << std::endl;
	*/


		std::cout << "Create Cuda TF" << std::endl;

		Volume* volume = SceneManager::get().getVolume();

		glGenTextures(1, &slicesTextureHandle);
		checkGLError(); //Error generating texture
		std::cout << "   glGenTextures setted" << std::endl;

		glGenTextures(1, &pcdTextureHandle);
		checkGLError(); //Error generating texture
		std::cout << "   glGenTextures setted" << std::endl;

		if (slicesTextureHandle) 
		{
			//bind texture
			glBindTexture(GL_TEXTURE_3D, slicesTextureHandle);
			checkGLError(); //Could not bind texture
			std::cout << "   BindTextures setted" << std::endl;

			int iTextureWidth = volume->GetWidth();
			int iTextureHeight = volume->GetHeight();
			int iTextureDepth = volume->GetDepth();

			float* vol = new float[volume->GetDepth()*volume->GetHeight()*volume->GetWidth()*4];
			float* pcd = new float[volume->GetDepth()*volume->GetHeight()*volume->GetWidth()*4];

			for(int z = 0; z < iTextureDepth; z++)
			{
				for(int y = 0; y < iTextureHeight; y++)
				{
					for(int x = 0; x < iTextureWidth; x++)
					{
						glm::vec4 gradient = GetGradient(volume, x, y, z, x+1, y+1, z+1, x-1, y-1, z-1, iTextureWidth, iTextureHeight, iTextureDepth);

						vol[(z*iTextureWidth*iTextureHeight*4)+(y*iTextureWidth*4)+(x*4)+0] = gradient.x;
						vol[(z*iTextureWidth*iTextureHeight*4)+(y*iTextureWidth*4)+(x*4)+1] = gradient.y;
						vol[(z*iTextureWidth*iTextureHeight*4)+(y*iTextureWidth*4)+(x*4)+2] = gradient.z;
						vol[(z*iTextureWidth*iTextureHeight*4)+(y*iTextureWidth*4)+(x*4)+3] = gradient.w;

						pcd[(z*iTextureWidth*iTextureHeight*4)+(y*iTextureWidth*4)+(x*4)+0] = gradient.x;
						pcd[(z*iTextureWidth*iTextureHeight*4)+(y*iTextureWidth*4)+(x*4)+1] = gradient.y;
						pcd[(z*iTextureWidth*iTextureHeight*4)+(y*iTextureWidth*4)+(x*4)+2] = gradient.z;
						pcd[(z*iTextureWidth*iTextureHeight*4)+(y*iTextureWidth*4)+(x*4)+3] = gradient.w;
					}
				}
			}
			/*/
			float* vol = new float[volume->GetDepth()*volume->GetHeight()*volume->GetWidth()];

			for(int z = 0; z < iTextureDepth; z++)
			{
				for(int y = 0; y < iTextureHeight; y++)
				{
					for(int x = 0; x < iTextureWidth; x++)
					{
						glm::vec4 gradient = GetGradient(volume, x, y, z, x+1, y+1, z+1, x-1, y-1, z-1, iTextureWidth, iTextureHeight, iTextureDepth);
						vol[(z*iTextureWidth*iTextureHeight)+(y*iTextureWidth)+(x)] = gradient.w;
					}
				}
			}
			//*/

			for(int z = 0; z < iTextureDepth; z++)
			{
				for(int y = 0; y < iTextureHeight; y++)
				{
					for(int x = 0; x < iTextureWidth; x++)
					{
						glm::vec4 curvatur = GetCurvature(vol, x, y, z, iTextureWidth, iTextureHeight, iTextureDepth);
						pcd[(z*iTextureWidth*iTextureHeight*4)+(y*iTextureWidth*4)+(x*4)+0] = curvatur.x;
						pcd[(z*iTextureWidth*iTextureHeight*4)+(y*iTextureWidth*4)+(x*4)+1] = curvatur.y;
						pcd[(z*iTextureWidth*iTextureHeight*4)+(y*iTextureWidth*4)+(x*4)+2] = curvatur.z;
						pcd[(z*iTextureWidth*iTextureHeight*4)+(y*iTextureWidth*4)+(x*4)+3] = curvatur.w;
					}
				}
			}
			
			int N = volume->GetDepth()*volume->GetHeight()*volume->GetWidth();
			int K = 5;
			glm::vec2* x = new glm::vec2[N];

			//_____________
			glm::vec2 minValue = glm::vec2(99.f, 99.f);
			glm::vec2 maxValue = glm::vec2(0.f, 0.f);
			for(int i = 0; i < N; i++) 
			{	
				x[i].x = vol[(i*4)+3];
				x[i].y = sqrt(powf(vol[(i*4)],2.0) + powf(vol[(i*4)+1],2.0) + powf(vol[(i*4)+2],2.0));

				minValue.x = min(minValue.x, x[i].x);
				minValue.y = min(minValue.y, x[i].y);
				maxValue.x = max(maxValue.x, x[i].x);
				maxValue.y = max(maxValue.y, x[i].y);
			}
			
			/*
			//_________________________
			//Create Histogramm
			unsigned char* tf_histogramm;
			if(maxValue.y > 1.f)	tf_histogramm = new unsigned char[1024*(int)(1024*maxValue.y)*3];
			else					tf_histogramm = new unsigned char[1024*1024*3];
			for(int i = 0; i < 1024*1024*3; i++) tf_histogramm[i] = 0;
			for(int j = 0; j < iNumDiv; j++)
			{
				for(int i = 0; i < N/iNumDiv; i++) 
				{	
					int tmp_x = (int)(x1[j][i].x*1024.f);
					int tmp_y = (int)(x1[j][i].y*1024.f);
					tf_histogramm[(1024*(1024-tmp_y)*3)+(tmp_x*3)+0] = min(255, tf_histogramm[(1024*(1024-tmp_y)*3)+(tmp_x*3)+0]+5);
					tf_histogramm[(1024*(1024-tmp_y)*3)+(tmp_x*3)+1] = min(255, tf_histogramm[(1024*(1024-tmp_y)*3)+(tmp_x*3)+1]+5);
					tf_histogramm[(1024*(1024-tmp_y)*3)+(tmp_x*3)+2] = min(255, tf_histogramm[(1024*(1024-tmp_y)*3)+(tmp_x*3)+2]+5);
				}
			}

			std::cout << std::endl;
			std::cout << "HISTOGRAMM CREATED " << std::endl;
			std::cout << std::endl;
			std::cout << std::endl;
			//_________________________
			*/

			setValuesUpdated(x, N);

			//upload texture
			glTexImage3D(GL_TEXTURE_3D, 0, GL_RGBA32F, iTextureWidth, iTextureHeight, iTextureDepth, 0, GL_RGBA, GL_FLOAT, vol);
			glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
			glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
			glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
			glBindTexture(GL_TEXTURE_3D, 0);
			delete[] vol;

			Texture::TexturePtr tr = Texture::TexturePtr(new Texture(3));
			tr->SetTexName(slicesTextureHandle);
			tr->m_width = volume->GetWidth();
			tr->m_height = volume->GetHeight();
			tr->m_depth = volume->GetDepth();
			std::cout << "Texture created" << std::endl;

			m_volumen = new CudaTexture<float4, 3>(tr);				
			m_volumen->Map();
			initVolumenTexture(m_volumen->GetCudaTexture(), m_volumen->GetCudaExtend());
			std::cout << "m_volumen done" << std::endl;

		//--------------
			//upload texture
			glBindTexture(GL_TEXTURE_3D, pcdTextureHandle);
			checkGLError(); //Could not bind texture

			glTexImage3D(GL_TEXTURE_3D, 0, GL_RGBA32F, iTextureWidth, iTextureHeight, iTextureDepth, 0, GL_RGBA, GL_FLOAT, pcd);
			glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
			glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
			glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
			glBindTexture(GL_TEXTURE_3D, 0);
			delete[] pcd;


			Texture::TexturePtr tr_pcd = Texture::TexturePtr(new Texture(3));
			tr_pcd->SetTexName(pcdTextureHandle);
			tr_pcd->m_width = volume->GetWidth();
			tr_pcd->m_height = volume->GetHeight();
			tr_pcd->m_depth = volume->GetDepth();

			m_pcd = new CudaTexture<float4, 3>(tr_pcd);				
			m_pcd->Map();
			initPCDTexture(m_pcd->GetCudaTexture(), m_pcd->GetCudaExtend());
			std::cout << "m_volumen done" << std::endl;
		//--------------------

			rayCasting->m_volumeBricks_back->SetGrid(glm::vec3(volume->GetWidth(), volume->GetHeight(), volume->GetDepth()));
			rayCasting->m_volumeBricks_front->SetGrid(glm::vec3(volume->GetWidth(), volume->GetHeight(), volume->GetDepth()));
			cudaExtent volumeSize = make_cudaExtent(volume->GetWidth(), volume->GetHeight(), volume->GetDepth());
			cudaExtent blocks_size = make_cudaExtent(16, 16, 16);
			cudaExtent blocks_num = make_cudaExtent(volumeSize.width/blocks_size.width, volumeSize.height/blocks_size.height, volumeSize.depth/blocks_size.depth);
			
			int number_volumeBlocks = blocks_num.width * blocks_num.height * blocks_num.depth;
			m_volumeBlock = new CudaArray<float4>();
			m_volumeBlock->CreateBuffer(number_volumeBlocks, true); 
				
			m_volumeBlock->BindVBO();
			CalcVolumeBlockCuda(	m_volumeBlock->GetDevRead(), volumeSize, blocks_size, blocks_num);
			m_volumeBlock->UnbindVBO();
			std::cout << "CalcVolumeBlockCuda done" << std::endl;

			//m_volumeBlock->Debug(0, 100);

			rayCasting->m_plane_tmp->m_vertex_count = number_volumeBlocks;
			rayCasting->m_plane_tmp->m_vbovert = m_volumeBlock->GetDevReadVBO();

			BindTransferfunctionCuda(m_transferfunction->GetCudaTexture());
			rayCasting->AfterLoadingOfVolume(iTextureWidth, iTextureHeight, iTextureDepth);
			//UnbindTransferfunctionCuda();
			std::cout << "AfterLoadingOfVolume done" << std::endl;

			rayCasting->initPixelBuffer();
			rayCasting->display();

			UnbindVolumeCuda();
			m_volumen->Unmap();	
			m_pcd->Unmap();

			rayCasting->m_volumen = m_volumen;
			rayCasting->m_pcd = m_pcd;
		
			checkGLError(); //Could not upload texture
			std::cout << "   3D Texture created" << std::endl;

			this->slice->setSlicingTextureHandle(slicesTextureHandle);
			this->slice->setVolumeSize(volume->GetWidth(), volume->GetHeight(), volume->GetDepth());
			this->slice->setTextureLoaded(true);
			this->rayCasting->setVolumeTextureHandle(slicesTextureHandle);
			this->rayCasting->setPCDTextureHandle(pcdTextureHandle);
			this->rayCasting->setVolumeSize(volume->GetWidth(), volume->GetHeight(), volume->GetDepth());
			this->rayCasting->setTextureLoaded(true);

			checkGLError(); //Error generating texture
			std::cout << "   Values created" << std::endl;
			updateGL();
		}
	}

	checkGLError();
	std::cout << "initSliceTextured" << std::endl;
}

void OpenGLWidget::setXSlicing(int x)
{
	if (SceneManager::get().getVolume() != 0 ) 
	{
		makeCurrent();
		this->slice->setXSlice((float)x/(float)SceneManager::get().getVolume()->GetWidth());
		updateGL();
	}
}

void OpenGLWidget::setYSlicing(int y)
{
	if (SceneManager::get().getVolume() != 0 ) 
	{
		makeCurrent();
		this->slice->setYSlice((float)y/(float)SceneManager::get().getVolume()->GetHeight());
		updateGL();
	}
}

void OpenGLWidget::setZSlicing(int z)
{
	if (SceneManager::get().getVolume() != 0 ) 
	{
		makeCurrent();
		this->slice->setZSlice((float)z/(float)SceneManager::get().getVolume()->GetDepth());
		updateGL();
	}
}

void OpenGLWidget::setXClipping(int x)
{
	if (SceneManager::get().getVolume() != 0 && this->perspective) 
	{
		makeCurrent();
		this->rayCasting->setXClipping(1.0f - (x/100.0f));
		updateGL();
	}
}

void OpenGLWidget::setYClipping(int y)
{
	if (SceneManager::get().getVolume() != 0 && this->perspective) 
	{
		makeCurrent();
		this->rayCasting->setYClipping(1.0f - (y/100.0f));
		updateGL();
	}
}

void OpenGLWidget::setZClipping(int z)
{
	if (SceneManager::get().getVolume() != 0 && this->perspective) 
	{
		makeCurrent();
		this->rayCasting->setZClipping(1.0f - (z/100.0f));
		updateGL();
	}
}

void OpenGLWidget::generateTransferTexture(float* rgbaValues, int textureSize, int type)
{
	makeCurrent();
	m_mode = type;

	std::cout << "generateTransferTexture(float* rgbaValues, int textureSize, int type)" << std::endl;

	if(type == 0 || type == 2)
	{
		std::cout << "if(type == 0 || type == 2)" << std::endl;
		float* tmp = new float[4096*4];
		for(int y = 0; y < 1; y++)
		{
			for(int x = 0; x < 4096; x++)
			{
				tmp[(y*4096*4)+(x*4)+0] = rgbaValues[(x*4)+0];
				tmp[(y*4096*4)+(x*4)+1] = rgbaValues[(x*4)+1];
				tmp[(y*4096*4)+(x*4)+2] = rgbaValues[(x*4)+2];
				tmp[(y*4096*4)+(x*4)+3] = rgbaValues[(x*4)+3];
			}
		}

		std::cout << "type: " << type << std::endl;
		glActiveTexture(GL_TEXTURE0); 
		if(type == 0)
		{
			glBindTexture(GL_TEXTURE_1D, transferFunctionTextureHandle);
			glTexSubImage1D(GL_TEXTURE_1D, 0, 0, textureSize, GL_RGBA, GL_FLOAT, tmp);
		
			cutilSafeCall(cudaMemcpyToArray(m_transferfunction->GetCudaTexture(), 0, 0, tmp, sizeof(m_transferfunction->GetCudaArray()), cudaMemcpyHostToDevice));
				
			this->slice->setTransferTextureHandle(transferFunctionTextureHandle);
			this->slice->setTransferFunction(true);
	
			this->rayCasting->setTransferTextureHandle(transferFunctionTextureHandle);
			this->rayCasting->setTransferFunction(true);

		}
		else
		{
			glBindTexture(GL_TEXTURE_1D, transferFunctionTextureParticleHandle);
			glTexSubImage1D(GL_TEXTURE_1D, 0, 0, textureSize, GL_RGBA, GL_FLOAT, tmp);
	
			rayCasting->setTransferTextureParticleHandle(transferFunctionTextureParticleHandle);
		}

		std::cout << "ENDE: if(type == 0 || type == 2)" << std::endl;
	}	
	else
	{
		std::cout << "START 1: generateTransferTexture(float* rgbaValues, int textureSize, int type)" << std::endl;
		float* tmp = new float[1024*1024*4];
		for(int x = 0; x < (1024*1024*4); x++)	tmp[x] = max(0.f, min(1.f, rgbaValues[x]));
		
		/*unsigned int* tmp2 = new unsigned int[1024*1024*4];
		for(int x = 0; x < 1024*1024*4; x++)	tmp2[x] = (unsigned int)(min(255.f, 255.f*rgbaValues[x]));
		QImage image = QImage((unsigned char *) tmp2, 1024, 1024, QImage::Format_ARGB32);
		bool saved = image.save("D:\\2dtransfertexture.jpg");*/
		
		glActiveTexture(GL_TEXTURE0); 
		glBindTexture(GL_TEXTURE_2D, transferFunctionTextureHandle2D);
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 1024, 1024, GL_RGBA, GL_FLOAT, tmp);
		
		//cutilSafeCall(cudaMemcpyToArray(m_transferfunction->GetCudaTexture(), 0, 0, tmp, sizeof(m_transferfunction->GetCudaArray()), cudaMemcpyHostToDevice));
				
		this->rayCasting->setTransferTexture2DHandle(transferFunctionTextureHandle2D);
		this->rayCasting->setTransferFunction(true);

		std::cout << "START4: generateTransferTexture(float* rgbaValues, int textureSize, int type)" << std::endl;
	}

	rayCasting->m_type = m_mode;
	std::cout << "ENDE: generateTransferTexture(float* rgbaValues, int textureSize, int type)" << std::endl;
}

void OpenGLWidget::resetTransferTexture()
{
	makeCurrent();
	//glDeleteTextures(1,&transferFunctionTextureHandle);
	this->slice->setTransferFunction(false);
	this->rayCasting->setTransferFunction(false);
	rayCasting->m_type = m_mode;
	updateGL();
}

void OpenGLWidget::setLight(bool active)
{
	makeCurrent();
	if(active)	RenderStates::Light = 1;
	else		RenderStates::Light = 0;
	this->rayCasting->setLight(active);
	updateGL();
}

void OpenGLWidget::setPhong(bool active)
{
	makeCurrent();
	if(active)	RenderStates::Phong = 1;
	else		RenderStates::Phong = 0;
	this->rayCasting->setPhong(active);
	updateGL();
}

void OpenGLWidget::setMip(bool active)
{
	makeCurrent();
	if(active)
	{
		RenderStates::Average = 0;
		RenderStates::MIP = 1;
		RenderStates::Compositing = 0;
	}
	else		RenderStates::Average = 0;
	this->rayCasting->setMip(active);
	updateGL();
}

void OpenGLWidget::setAverage(bool active)
{
	makeCurrent();
	if(active)	
	{
		RenderStates::Average = 1;
		RenderStates::MIP = 0;
		RenderStates::Compositing = 0;
	}
	else		RenderStates::Average = 0;
	this->rayCasting->setAverage(active);
	updateGL();
}

void OpenGLWidget::setCompositing(bool active)
{
	makeCurrent();
	if(active)	
	{
		RenderStates::Average = 0;
		RenderStates::MIP = 0;
		RenderStates::Compositing = 1;
	}
	else		RenderStates::Compositing = 0;
	this->rayCasting->setCompositing(active);
	updateGL();
}

void OpenGLWidget::setAdaptive(bool active)
{
	makeCurrent();
	if(active)	RenderStates::Addaptive = 1;
	else		RenderStates::Addaptive = 0;
	this->rayCasting->setAdaptive(active);
	updateGL();

}

void OpenGLWidget::setStepSize(double step)
{
	makeCurrent();
	RenderStates::Stepsize = (float)(step);
	this->rayCasting->setStepSize(step);
	updateGL();
}

void OpenGLWidget::initializeGL()
{
	// --- Init GLEW ---

	glewExperimental= GL_TRUE;

	cout << "- Initializing GLEW..." << endl;
	GLenum err = glewInit(); // initialize GLEW

	if (err != GLEW_OK) {// check for error
		cerr << "- GLEW Error: " << glewGetErrorString(err) << endl;
		exit(1);
	} else {
		cout << "- GLEW initialized!" << endl;
	}

	this->checkGLError();
	std::cout << "   GLEW setted" << std::endl;

	//Test OpenGL 3.3
	cout << "- Testing OpenGL Version..." << endl;
	if (glewIsSupported("GL_VERSION_3_3")) {
		// retrieve our OpenGL-version
		cout << "- OpenGL Version: " << glGetString(GL_VERSION) << endl;
	} else {
		cout << "- OpenGL 3.3 ist not supported!" << endl;
		cout << "- OpenGL Version: " << glGetString(GL_VERSION) << endl;
		//exit(1);
	}

	this->checkGLError();
	std::cout << "   GL 3.3 setted" << std::endl;

	// --- Do some setup for OpenGL ---
	// enable back-face culling
	glEnable(GL_CULL_FACE);
	glEnable(GL_BLEND);
	glDepthMask(GL_TRUE);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glCullFace(GL_BACK);
	glShadeModel(GL_SMOOTH);
	this->checkGLError();
	std::cout << "   smoothed" << std::endl;

	// enable depth-buffer
    glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);

	this->checkGLError();
	std::cout << "   depth function seted" << std::endl;
	
	//set clear color
	qglClearColor(Qt::black);

	//set Viewport is done in resizeGL

	// --- init slicing view ---
	this->slice->init();

	this->checkGLError();
	std::cout << "   slice->initized" << std::endl;

	// --- init 3D view ---
	this->rayCasting->init();

	this->checkGLError();
	std::cout << "   rayCasting->initized" << std::endl;

	// --- init matrices ---
	// move camera 4 units backward, i.e. move the scene 4 units forward
	SceneManager::get().getCamera()->SetViewMatrix(glm::vec3(0.f,0.f,1.f), glm::vec3(0.f, 0.f, 0.f), glm::vec3(0.f,1.f,0.f));

	this->checkGLError();
	std::cout << "   SceneManager->set view matrix" << std::endl;

	// --- done with initializing stuff ---
	this->checkGLError();

	unsigned int tf_size = 1024*4;
	float4* tf_tmp = new float4[tf_size];
	for(unsigned int j = 0; j < tf_size; j++)	tf_tmp[j] = make_float4( 1.0f, 1.f, 1.f, 1.f);

	glGenTextures(1,&transferFunctionTextureHandle);
	glBindTexture(GL_TEXTURE_1D, transferFunctionTextureHandle);
	glTexImage1D(GL_TEXTURE_1D, 0, GL_RGBA32F, tf_size, 0, GL_RGBA, GL_FLOAT, tf_tmp);
	glTexParameteri(GL_TEXTURE_1D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_1D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glBindTexture(GL_TEXTURE_1D, 0);
	delete[] tf_tmp;


	tf_size = 1024*1024;
	tf_tmp = new float4[tf_size];
	for(unsigned int j = 0; j < tf_size; j++)	tf_tmp[j] = make_float4( 1.0f, 1.f, 1.f, 1.f);

	glGenTextures(1,&transferFunctionTextureHandle2D);
	glBindTexture(GL_TEXTURE_2D, transferFunctionTextureHandle2D);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, 1024, 1024, 0, GL_RGBA, GL_FLOAT, tf_tmp);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glBindTexture(GL_TEXTURE_2D, 0);
	delete[] tf_tmp;

	tf_size = 1024*4;
	tf_tmp = new float4[tf_size];
	for(unsigned int j = 0; j < tf_size; j++)	tf_tmp[j] = make_float4( 1.0f, 1.f, 1.f, 1.f);
	
	glGenTextures(1,&transferFunctionTextureParticleHandle);
	glBindTexture(GL_TEXTURE_1D, transferFunctionTextureParticleHandle);
	glTexImage1D(GL_TEXTURE_1D, 0, GL_RGBA32F, tf_size, 0, GL_RGBA, GL_FLOAT, tf_tmp);
	glTexParameteri(GL_TEXTURE_1D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_1D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glBindTexture(GL_TEXTURE_1D, 0);
	delete[] tf_tmp;

	std::cout << "initializeGL" << std::endl;
	checkGLError();
	std::cout << "!!!" << std::endl;
}

void OpenGLWidget::resizeGL(int width, int height)
{
	if (height == 0)
	{
		height = 1;
	}

	glViewport(0, 0, width, height);

	this->screenWidth = width;
	this->screenHeight = height;

	this->setCameraProjection();

	this->slice->resize(width, height);
	this->rayCasting->resize(width, height);

	updateGL();
}

void OpenGLWidget::paintGL()
{
	if(!isStarted)	
	{
		timer.start();
		isStarted = true;
	}

	int elapsed = timer.elapsed();
	timer.restart();

	float elapsedTime = (float)(elapsed)/1000.f;
	QString t =QString::number(max(0.0, min(10000.0, 1.0/(double)(elapsedTime))));
	window()->setWindowTitle("Volume Rendering Application: " + t + " fps"); 

	// Main method for rendering
	if (RenderStates::ModusChanged == 0)	this->slice->render();
	else									this->rayCasting->render(elapsedTime);
	update();
}

void OpenGLWidget::mousePressEvent(QMouseEvent *event)
{
	if (event->button() == Qt::LeftButton) 
	{
		this->lastMovePos = glm::vec2(event->pos().x(), event->pos().y());
		this->moveVolume = true;
	} 
	else if (event->button() == Qt::RightButton) 
	{
		this->lastZoomYPos = event->pos().y();
		this->zoomVolume = true;
	}
}

void OpenGLWidget::mouseMoveEvent(QMouseEvent *event)
{
	if ((event->buttons() & Qt::LeftButton) && this->moveVolume) 
	{
		//volume wird bewegt!
		if (this->perspective) 
		{
			glm::vec2 pos = glm::vec2(event->pos().x(), event->pos().y());
			glm::vec2 diff = pos - this->lastMovePos;

			if (glm::abs(diff.x) > 0.000001f || glm::abs(diff.y) > 0.000001f) 
			{
				glm::mat4 view = SceneManager::get().getCamera()->GetViewMatrix();

				glm::vec4 rotationAxis = glm::vec4(diff.y, diff.x, 0.0f, 1.0) * view;
				float rotationAngle = glm::length(diff) / PI * 1.0f;
				glm::mat4 rotation = glm::rotate(glm::mat4(1.0f), rotationAngle, glm::vec3(rotationAxis.x, rotationAxis.y, rotationAxis.z));

				SceneManager::get().getCamera()->SetViewMatrix(view * rotation);
			}
			this->lastMovePos = pos;
		}
	} 
	else if ((event->buttons() & Qt::RightButton) && this->zoomVolume && this->perspective) 
	{
		int pos = event->pos().y();
		float diff = pos - this->lastZoomYPos;
		diff = diff * 0.005f;

		glm::vec3 distance = SceneManager::get().getCamera()->GetPosition();
		glm::vec3 delta = glm::normalize(distance) * diff;
		SceneManager::get().getCamera()->SetViewMatrix(glm::translate(SceneManager::get().getCamera()->GetViewMatrix(), delta));

		this->lastZoomYPos = pos;
	}
}

void OpenGLWidget::mouseReleaseEvent(QMouseEvent *event)
{
	if ((event->button() == Qt::LeftButton) && this->moveVolume) this->moveVolume = false;
}

void OpenGLWidget::wheelEvent(QWheelEvent *event)
{
	if (event->orientation() == Qt::Vertical && this->perspective) 
	{
		int numDegrees = event->delta() / 8;
		float numSteps = numDegrees / 15;
		numSteps = numSteps * 0.5f;

		glm::vec3 distance = SceneManager::get().getCamera()->GetPosition();
		glm::vec3 delta = glm::normalize(distance) * numSteps;
		SceneManager::get().getCamera()->SetViewMatrix(glm::translate(SceneManager::get().getCamera()->GetViewMatrix(), delta));
		event->accept();
	} 
	else event->ignore();
}

void OpenGLWidget::setCameraProjection()
{
	Camera* cam = SceneManager::get().getCamera();
	if(perspective) cam->SetProjection(fovAngle, (float(screenWidth)/float(screenHeight)), nearplane, farplane);
	else			cam->SetProjection((float)-(0.5f), (float)(0.5f), (float)-(0.5f), (float)(0.5f));
}

void OpenGLWidget::checkGLError()
{
	GLenum error = glGetError();
	if (error != GL_NO_ERROR)
	{
		std::cout << "#########################" << std::endl;
		std::cout << "OpenGLWidget " << std::endl;
		switch (error)
		{
			case GL_INVALID_ENUM:
				cout << "GL: enum argument out of range." << endl;
				break;
			case GL_INVALID_VALUE:
				cout << "GL: Numeric argument out of range." << endl;
				break;
			case GL_INVALID_OPERATION:
				cout << "GL: Operation illegal in current state." << endl;
				break;
			case GL_INVALID_FRAMEBUFFER_OPERATION:
				cout << "GL: Framebuffer object is not complete." << endl;
				break;
			case GL_OUT_OF_MEMORY:
				cout << "GL: Not enough memory left to execute command." << endl;
				break;
			default:
				cout << "GL: Unknown error." << endl;
		}
		std::cout << "#########################" << std::endl;
	}
}

void OpenGLWidget::setValuesUpdated(glm::vec2* values, int size)
{
	emit valuesUpdated(values, size);
}

void OpenGLWidget::changeIsoSurface(float val)
{
	std::cout << "OpenGLWidget::changeIsoSurface(float val)" << std::endl;
	rayCasting->SetIsoValue(val);
}