#pragma once
#include "VolumeCuda.cu"

__constant__ Particle particle_parameter;

uint Floor(uint a, uint b){
    return (a % b != 0) ? (a / b + 1) : (a / b);
}

// calculate position in uniform grid
__device__ int3 calcGridPos(float3 p)
{
    int3 gridPos;
    gridPos.x = floor(p.x * 64.0);
    gridPos.y = floor(p.y * 64.0);
    gridPos.z = floor(p.z * 64.0);

	if(gridPos.x  < 0.0) gridPos.x = 0;
	if(gridPos.y  < 0.0) gridPos.y = 0;
	if(gridPos.z  < 0.0) gridPos.z = 0;
	if(gridPos.x  >= 64.0) gridPos.x = 63;
	if(gridPos.y  >= 64.0) gridPos.y = 63;
	if(gridPos.z  >= 64.0) gridPos.z = 63;

    return gridPos;
}

__device__ uint calcGridIndex(int3 gridPos)
{
    gridPos.x = gridPos.x & (64-1); 
    gridPos.y = gridPos.y & (64-1);
    gridPos.z = gridPos.z & (64-1);        
    return __umul24(__umul24(gridPos.z, 64), 64) + __umul24(64, 64) + gridPos.x;
}

__device__ 
float4 lerp2(float4 a, float4 b, float w)
{
	return lerp(a, b, w);
	//return (a*(1.0-w)) + (w*b);
}

__device__ 
float3 lerp2(float3 a, float3 b, float w)
{
	return lerp(a, b, w);
	//return (a*(1.0-w)) + (w*b);
}

__device__ 
float2 lerp2(float2 a, float2 b, float w)
{
	return lerp(a, b, w);
	//return (a*(1.0-w)) + (w*b);
}

__device__ 
float lerp2(float a, float b, float w)
{
	return lerp(a, b, w);
	//return (a*(1.0-w)) + (w*b);
}


__device__
float calcGrad(float4 hg_x, float4 hg_y, float4 hg_z, int type, float3 coord, float3 size_source)
{
	float3 coord_source100 = coord - make_float3(hg_x.x/size_source.x, 0.0, 0.0);
	float3 coord_source000 = coord + make_float3(hg_x.y/size_source.x, 0.0, 0.0);

	float3 coord_source110 = coord_source100 - make_float3(0.0, hg_y.x/size_source.y, 0.0);
	float3 coord_source010 = coord_source000 - make_float3(0.0, hg_y.x/size_source.y, 0.0);
	       coord_source100 = coord_source100 + make_float3(0.0, hg_y.y/size_source.y, 0.0);
	       coord_source000 = coord_source000 + make_float3(0.0, hg_y.y/size_source.y, 0.0);

	float3 coord_source111 = coord_source110 - make_float3(0.0, 0.0, hg_z.x/size_source.z);
	float3 coord_source011 = coord_source010 - make_float3(0.0, 0.0, hg_z.x/size_source.z);
	float3 coord_source101 = coord_source100 - make_float3(0.0, 0.0, hg_z.x/size_source.z);
	float3 coord_source001 = coord_source000 - make_float3(0.0, 0.0, hg_z.x/size_source.z);
	       coord_source110 = coord_source110 + make_float3(0.0, 0.0, hg_z.y/size_source.z);
	       coord_source010 = coord_source010 + make_float3(0.0, 0.0, hg_z.y/size_source.z);
  	       coord_source100 = coord_source100 + make_float3(0.0, 0.0, hg_z.y/size_source.z);
	       coord_source000 = coord_source000 + make_float3(0.0, 0.0, hg_z.y/size_source.z);


	//ab da sollte es wieder stimmen
	// fetch four linearly interpolated inputs
	float4 n_val;
	n_val.z = tex3D( volume_tex, coord_source001.x, coord_source001.y, coord_source001.z).w;
	n_val.w = tex3D( volume_tex, coord_source101.x, coord_source101.y, coord_source101.z).w;
	n_val.x = tex3D( volume_tex, coord_source011.x, coord_source011.y, coord_source011.z).w;
	n_val.y = tex3D( volume_tex, coord_source111.x, coord_source111.y, coord_source111.z).w;

	float4 f_val;
	f_val.z = tex3D( volume_tex, coord_source000.x, coord_source000.y, coord_source000.z).w;
	f_val.w = tex3D( volume_tex, coord_source100.x, coord_source100.y, coord_source100.z).w;
	f_val.x = tex3D( volume_tex, coord_source010.x, coord_source010.y, coord_source010.z).w;
	f_val.y = tex3D( volume_tex, coord_source110.x, coord_source110.y, coord_source110.z).w;

	float gradient = 0.0;

	// weight along z direction	
	switch(type)
	{
		case 0: 
		{
			f_val = lerp2(n_val, f_val, hg_z.z);
			float2 tmp = lerp2(make_float2(f_val.x, f_val.y), make_float2(f_val.z, f_val.w), hg_y.z);
			gradient = (tmp.y - tmp.x)*hg_x.z;
			break;
		}
		case 1:	
		{
			f_val = lerp2(n_val, f_val, hg_z.z);	
			float2 tmp = (make_float2(f_val.z, f_val.w)-make_float2(f_val.x, f_val.y)) * hg_y.z;
			gradient = lerp2(tmp.x, tmp.y, hg_x.z);
			break;
		}
		case 2:	
		{	
			f_val = (f_val-n_val)*hg_z.z;
			float2 tmp = lerp2(make_float2(f_val.x, f_val.y), make_float2(f_val.z, f_val.w), hg_y.z);
			gradient = lerp2(tmp.x, tmp.y, hg_x.z);	
			break;
		}
	}

	return gradient;
}

__device__
float3 calculateGradient(float3 coord, float3 size_source )
{
	//kann dann 1zu1 fr cuda genommen werden (beides bentzt normalisierte Texturekoordinaten)
	// calc filter texture coordinates where [0,1] is a single texel
	// (can be done in vertex program instead)
	//Es ist ja nur der Nachkommabereich interessant von den Texturkoordinaten [0,128]
	//Nachdem ich GL_REPEAT bei der Textur habe, sollte es passen.

//vec3 coord_hg = (coord*size_source) - vec3(0.5f, 0.5f, 0.5f);
float3 coord_hg = (coord) + (make_float3(0.5f, 0.5f, 0.5f)/size_source);

	// fetch offsets and weights from filter texture
	float4 hg_x = tex2D( bspline_hg_tex, coord_hg.x, 0.0);
	float4 hg_y = tex2D( bspline_hg_tex, coord_hg.y, 0.0);
	float4 hg_z = tex2D( bspline_hg_tex, coord_hg.z, 0.0);

	float4 hg_x_dev = tex2D( bspline_hg_dev_tex, coord_hg.x, 0.0);
	float4 hg_y_dev = tex2D( bspline_hg_dev_tex, coord_hg.y, 0.0);
	float4 hg_z_dev = tex2D( bspline_hg_dev_tex, coord_hg.z, 0.0);

	float3 gradient;
	gradient.x = calcGrad(hg_x_dev, hg_y, hg_z, 0, coord, size_source);
	gradient.y = calcGrad(hg_x, hg_y_dev, hg_z, 1, coord, size_source);
	gradient.z = calcGrad(hg_x, hg_y, hg_z_dev, 2, coord, size_source);

	//gradient = (view * vec4(gradient, 0.0)).xyz;

	return (gradient)*2.0;
}

__device__
float3x3 calcHesse(float3 coord, float3 size_source)
{
	float3 coord_hg = (coord) + (make_float3(0.5f, 0.5f, 0.5f)/size_source);

	// fetch offsets and weights from filter texture
	float4 hg_x = tex2D( bspline_hg_tex, coord_hg.x,0.5);
	float4 hg_y = tex2D( bspline_hg_tex, coord_hg.y,0.5);
	float4 hg_z = tex2D( bspline_hg_tex, coord_hg.z,0.5);

	float4 hg_x_dev = tex2D( bspline_hg_dev_tex, coord_hg.x,0.5);
	float4 hg_y_dev = tex2D( bspline_hg_dev_tex, coord_hg.y,0.5);
	float4 hg_z_dev = tex2D( bspline_hg_dev_tex, coord_hg.z,0.5);
	
	float xx = 0.0;
	float yy = 0.0;
	float zz = 0.0;
	float xy = 0.0;
	float xz = 0.0;
	float yz = 0.0;
	
	//__________________________
	//Hesse xx
	float3 mlf_tc = coord - make_float3(0.0, hg_y.x/size_source.y, 0.0);
	float3 muf_tc = coord + make_float3(0.0, hg_y.y/size_source.y, 0.0);
              
	float3	mln_tc = mlf_tc - make_float3(0.0, 0.0, hg_z.x/size_source.z);
	float3	mun_tc = muf_tc - make_float3(0.0, 0.0, hg_z.x/size_source.z);
			mlf_tc = mlf_tc + make_float3(0.0, 0.0, hg_z.y/size_source.z);
			muf_tc = muf_tc + make_float3(0.0, 0.0, hg_z.y/size_source.z);
	
	float3 lln_tc = mln_tc - make_float3(1.0/size_source.x, 0.0, 0.0);
	float3 lun_tc = mun_tc - make_float3(1.0/size_source.x, 0.0, 0.0);
	float3 llf_tc = mlf_tc - make_float3(1.0/size_source.x, 0.0, 0.0);
	float3 luf_tc = muf_tc - make_float3(1.0/size_source.x, 0.0, 0.0);
	float3 rln_tc = mln_tc + make_float3(1.0/size_source.x, 0.0, 0.0);
	float3 run_tc = mun_tc + make_float3(1.0/size_source.x, 0.0, 0.0);
	float3 rlf_tc = mlf_tc + make_float3(1.0/size_source.x, 0.0, 0.0);
	float3 ruf_tc = muf_tc + make_float3(1.0/size_source.x, 0.0, 0.0);

	//ab da sollte es wieder stimmen
	// fetch four linearly interpolated inputs
	float4 r_val;
	float4 m_val;
	float4 l_val;
	float4 f_val;
	float4 n_val;
	l_val.x = tex3D( volume_tex, lln_tc.x, lln_tc.y, lln_tc.z).w;
	l_val.y = tex3D( volume_tex, lun_tc.x, lun_tc.y, lun_tc.z).w;
	l_val.z = tex3D( volume_tex, llf_tc.x, llf_tc.y, llf_tc.z).w;
	l_val.w = tex3D( volume_tex, luf_tc.x, luf_tc.y, luf_tc.z).w;
	
	m_val.x = tex3D( volume_tex, mln_tc.x, mln_tc.y, mln_tc.z).w;
	m_val.y = tex3D( volume_tex, mun_tc.x, mun_tc.y, mun_tc.z).w;
	m_val.z = tex3D( volume_tex, mlf_tc.x, mlf_tc.y, mlf_tc.z).w;
	m_val.w = tex3D( volume_tex, muf_tc.x, muf_tc.y, muf_tc.z).w;
	
	r_val.x = tex3D( volume_tex, rln_tc.x, rln_tc.y, rln_tc.z).w;
	r_val.y = tex3D( volume_tex, run_tc.x, run_tc.y, run_tc.z).w;
	r_val.z = tex3D( volume_tex, rlf_tc.x, rlf_tc.y, rlf_tc.z).w;
	r_val.w = tex3D( volume_tex, ruf_tc.x, ruf_tc.y, ruf_tc.z).w;


	float2	tmp = lerp2(make_float2(l_val.x,l_val.y), make_float2(l_val.z, l_val.w), hg_z.z);	l_val.z = tmp.x;	l_val.w = tmp.y;    
			tmp = lerp2(make_float2(m_val.x,m_val.y), make_float2(m_val.z, m_val.w), hg_z.z);	m_val.z = tmp.x;	m_val.w = tmp.y;    
			tmp = lerp2(make_float2(r_val.x,r_val.y), make_float2(r_val.z, r_val.w), hg_z.z);	r_val.z = tmp.x;	r_val.w = tmp.y;    

	l_val.x = lerp2(l_val.z, l_val.w, hg_y.z);
	l_val.y = lerp2(m_val.z, m_val.w, hg_y.z);
	l_val.z = lerp2(r_val.z, r_val.w, hg_y.z);
	
	xx = dot(make_float3(l_val.x, l_val.y, l_val.z), make_float3(1.0,-2.0,1.0));


	mlf_tc = make_float3(0.0, 0.0, 0.0);
	muf_tc = make_float3(0.0, 0.0, 0.0);
	mln_tc = make_float3(0.0, 0.0, 0.0);
	mun_tc = make_float3(0.0, 0.0, 0.0);

	//__________________________
	//Hesse yy
	// left/right upper/lower near/far
	
	float3 lmf_tc = coord - make_float3(hg_x.x/size_source.x, 0.0, 0.0);
	float3 rmf_tc = coord + make_float3(hg_x.y/size_source.x, 0.0, 0.0);
              
	float3	lmn_tc = lmf_tc - make_float3(0.0, 0.0, hg_z.x/size_source.z);
	float3	rmn_tc = rmf_tc - make_float3(0.0, 0.0, hg_z.x/size_source.z);
			lmf_tc = lmf_tc + make_float3(0.0, 0.0, hg_z.y/size_source.z);
			rmf_tc = rmf_tc + make_float3(0.0, 0.0, hg_z.y/size_source.z);

	lln_tc = lmn_tc - make_float3(0.0, 1.0/size_source.y, 0.0);
	rln_tc = rmn_tc - make_float3(0.0, 1.0/size_source.y, 0.0);
	llf_tc = lmf_tc - make_float3(0.0, 1.0/size_source.y, 0.0);
	rlf_tc = rmf_tc - make_float3(0.0, 1.0/size_source.y, 0.0);
	lun_tc = lmn_tc + make_float3(0.0, 1.0/size_source.y, 0.0);
	run_tc = rmn_tc + make_float3(0.0, 1.0/size_source.y, 0.0);
	luf_tc = lmf_tc + make_float3(0.0, 1.0/size_source.y, 0.0);
	ruf_tc = rmf_tc + make_float3(0.0, 1.0/size_source.y, 0.0);

	l_val.x = tex3D( volume_tex, lln_tc.x, lln_tc.y, lln_tc.z).w;
	l_val.y = tex3D( volume_tex, rln_tc.x, rln_tc.y, rln_tc.z).w;
	l_val.z = tex3D( volume_tex, llf_tc.x, llf_tc.y, llf_tc.z).w;
	l_val.w = tex3D( volume_tex, rlf_tc.x, rlf_tc.y, rlf_tc.z).w;
	
	m_val.x = tex3D( volume_tex, lmn_tc.x, lmn_tc.y, lmn_tc.z).w;
	m_val.y = tex3D( volume_tex, rmn_tc.x, rmn_tc.y, rmn_tc.z).w;
	m_val.z = tex3D( volume_tex, lmf_tc.x, lmf_tc.y, lmf_tc.z).w;
	m_val.w = tex3D( volume_tex, rmf_tc.x, rmf_tc.y, rmf_tc.z).w;
	
	r_val.x = tex3D( volume_tex, lun_tc.x, lun_tc.y, lun_tc.z).w;
	r_val.y = tex3D( volume_tex, run_tc.x, run_tc.y, run_tc.z).w;
	r_val.z = tex3D( volume_tex, luf_tc.x, luf_tc.y, luf_tc.z).w;
	r_val.w = tex3D( volume_tex, ruf_tc.x, ruf_tc.y, ruf_tc.z).w;

	tmp = lerp2(make_float2(l_val.x,l_val.y), make_float2(l_val.z, l_val.w), hg_z.z);	l_val.z = tmp.x;	l_val.w = tmp.y;
	tmp = lerp2(make_float2(m_val.x,m_val.y), make_float2(m_val.z, m_val.w), hg_z.z);	m_val.z = tmp.x;	m_val.w = tmp.y;
	tmp = lerp2(make_float2(r_val.x,r_val.y), make_float2(r_val.z, r_val.w), hg_z.z);	r_val.z = tmp.x;	r_val.w = tmp.y;

	l_val.x = lerp2(l_val.z, l_val.w, hg_x.z);
	l_val.y = lerp2(m_val.z, m_val.w, hg_x.z);
	l_val.z = lerp2(r_val.z, r_val.w, hg_x.z);
	
	yy = dot(make_float3(l_val.x, l_val.y, l_val.z), make_float3(1.0,-2.0,1.0));

	//__________________________
	//Hesse zz
	// left/right upper/lower near/far
	float3 llm_tc = coord - make_float3(hg_x.x/size_source.x, 0.0, 0.0);
	float3 rlm_tc = coord + make_float3(hg_x.y/size_source.x, 0.0, 0.0);
              
			llm_tc = llm_tc - make_float3(0.0, hg_y.x/size_source.y, 0.0);
			rlm_tc = rlm_tc - make_float3(0.0, hg_y.x/size_source.y, 0.0);
	float3	lum_tc = llm_tc + make_float3(0.0, hg_y.y/size_source.y, 0.0);
	float3	rum_tc = rlm_tc + make_float3(0.0, hg_y.y/size_source.y, 0.0);

	lln_tc = llm_tc - make_float3(0.0, 0.0, 1.0/size_source.z);
	rln_tc = rlm_tc - make_float3(0.0, 0.0, 1.0/size_source.z);
	lun_tc = lum_tc - make_float3(0.0, 0.0, 1.0/size_source.z);
	run_tc = rum_tc - make_float3(0.0, 0.0, 1.0/size_source.z);
	llf_tc = llm_tc + make_float3(0.0, 0.0, 1.0/size_source.z);
	rlf_tc = rlm_tc + make_float3(0.0, 0.0, 1.0/size_source.z);
	luf_tc = lum_tc + make_float3(0.0, 0.0, 1.0/size_source.z);
	ruf_tc = rum_tc + make_float3(0.0, 0.0, 1.0/size_source.z);

	l_val.x = tex3D( volume_tex, lln_tc.x, lln_tc.y, lln_tc.z).w;
	l_val.y = tex3D( volume_tex, rln_tc.x, rln_tc.y, rln_tc.z).w;
	l_val.z = tex3D( volume_tex, lun_tc.x, lun_tc.y, lun_tc.z).w;
	l_val.w = tex3D( volume_tex, run_tc.x, run_tc.y, run_tc.z).w;
	
	m_val.x = tex3D( volume_tex, llm_tc.x, llm_tc.y, llm_tc.z).w;
	m_val.y = tex3D( volume_tex, rlm_tc.x, rlm_tc.y, rlm_tc.z).w;
	m_val.z = tex3D( volume_tex, lum_tc.x, lum_tc.y, lum_tc.z).w;
	m_val.w = tex3D( volume_tex, rum_tc.x, rum_tc.y, rum_tc.z).w;
	
	f_val.x = tex3D( volume_tex, llf_tc.x, llf_tc.y, llf_tc.z).w;
	f_val.y = tex3D( volume_tex, rlf_tc.x, rlf_tc.y, rlf_tc.z).w;
	f_val.z = tex3D( volume_tex, luf_tc.x, luf_tc.y, luf_tc.z).w;
	f_val.w = tex3D( volume_tex, ruf_tc.x, ruf_tc.y, ruf_tc.z).w;

	tmp = lerp2(make_float2(l_val.x,l_val.y), make_float2(l_val.z, l_val.w), hg_y.z);	l_val.z = tmp.x;	l_val.w = tmp.y;
	tmp = lerp2(make_float2(m_val.x,m_val.y), make_float2(m_val.z, m_val.w), hg_y.z);	m_val.z = tmp.x;	m_val.w = tmp.y;
	tmp = lerp2(make_float2(r_val.x,r_val.y), make_float2(r_val.z, m_val.w), hg_y.z);	f_val.z = tmp.x;	f_val.w = tmp.y;

	l_val.x = lerp2(l_val.z, l_val.w, hg_x.z);
	l_val.y = lerp2(m_val.z, m_val.w, hg_x.z);
	l_val.z = lerp2(r_val.z, r_val.w, hg_x.z);

	zz = dot(make_float3(l_val.x, l_val.y, l_val.z), make_float3(1.0,-2.0,1.0));



	float4 hg2_x = hg_x_dev;
	float4 hg2_y = hg_y_dev;
	float4 hg2_z = hg_z;
	//__________________________
	//Hesse xy
	// setup texture coordinate
	// left/right upper/lower near/far
	llf_tc = coord - make_float3(hg2_x.x/size_source.x, 0.0, 0.0);
	rlf_tc = coord + make_float3(hg2_x.y/size_source.x, 0.0, 0.0);
              
	luf_tc = llf_tc - make_float3(0.0, hg2_y.x/size_source.y, 0.0);
	ruf_tc = rlf_tc - make_float3(0.0, hg2_y.x/size_source.y, 0.0);
	llf_tc = llf_tc + make_float3(0.0, hg2_y.y/size_source.y, 0.0);
	rlf_tc = rlf_tc + make_float3(0.0, hg2_y.y/size_source.y, 0.0);
              
	lun_tc = luf_tc - make_float3(0.0, 0.0, hg2_z.x/size_source.z);
	run_tc = ruf_tc - make_float3(0.0, 0.0, hg2_z.x/size_source.z);
	lln_tc = llf_tc - make_float3(0.0, 0.0, hg2_z.x/size_source.z);
	rln_tc = rlf_tc - make_float3(0.0, 0.0, hg2_z.x/size_source.z);
	luf_tc = luf_tc + make_float3(0.0, 0.0, hg2_z.y/size_source.z);
	ruf_tc = ruf_tc + make_float3(0.0, 0.0, hg2_z.y/size_source.z);
	llf_tc = llf_tc + make_float3(0.0, 0.0, hg2_z.y/size_source.z);
	rlf_tc = rlf_tc + make_float3(0.0, 0.0, hg2_z.y/size_source.z);

	// read values (2 indirections)
	n_val.x = tex3D( volume_tex, lun_tc.x, lun_tc.y, lun_tc.z).w;
	n_val.y = tex3D( volume_tex, run_tc.x, run_tc.y, run_tc.z).w;
	n_val.z = tex3D( volume_tex, lln_tc.x, lln_tc.y, lln_tc.z).w;
	n_val.w = tex3D( volume_tex, rln_tc.x, rln_tc.y, rln_tc.z).w;
	f_val.x = tex3D( volume_tex, luf_tc.x, luf_tc.y, luf_tc.z).w;
	f_val.y = tex3D( volume_tex, ruf_tc.x, ruf_tc.y, ruf_tc.z).w;
	f_val.z = tex3D( volume_tex, llf_tc.x, llf_tc.y, llf_tc.z).w;
	f_val.w = tex3D( volume_tex, rlf_tc.x, rlf_tc.y, rlf_tc.z).w;

	// output derivative in x direction
	f_val = lerp2(n_val, f_val, hg_z.z);
	tmp = (make_float2(f_val.z,f_val.w) - make_float2(f_val.x,f_val.y)) * hg_y.z;
	xy = (tmp.y - tmp.x) * hg_x.z;


	hg2_x = hg_x_dev;
	hg2_y = hg_y;
	hg2_z = hg_z_dev;
	//__________________________
	//Hesse xz
	// setup texture coordinate
	// left/right upper/lower near/far
	llf_tc = coord - make_float3(hg2_x.x/size_source.x, 0.0, 0.0);
	rlf_tc = coord + make_float3(hg2_x.y/size_source.x, 0.0, 0.0);
              
	luf_tc = llf_tc - make_float3(0.0, hg2_y.x/size_source.y, 0.0);
	ruf_tc = rlf_tc - make_float3(0.0, hg2_y.x/size_source.y, 0.0);
	llf_tc = llf_tc + make_float3(0.0, hg2_y.y/size_source.y, 0.0);
	rlf_tc = rlf_tc + make_float3(0.0, hg2_y.y/size_source.y, 0.0);
              
	lun_tc = luf_tc - make_float3(0.0, 0.0, hg2_z.x/size_source.z);
	run_tc = ruf_tc - make_float3(0.0, 0.0, hg2_z.x/size_source.z);
	lln_tc = llf_tc - make_float3(0.0, 0.0, hg2_z.x/size_source.z);
	rln_tc = rlf_tc - make_float3(0.0, 0.0, hg2_z.x/size_source.z);
	luf_tc = luf_tc + make_float3(0.0, 0.0, hg2_z.y/size_source.z);
	ruf_tc = ruf_tc + make_float3(0.0, 0.0, hg2_z.y/size_source.z);
	llf_tc = llf_tc + make_float3(0.0, 0.0, hg2_z.y/size_source.z);
	rlf_tc = rlf_tc + make_float3(0.0, 0.0, hg2_z.y/size_source.z);

	// read values (2 indirections)
	n_val.x = tex3D(volume_tex, lun_tc.x, lun_tc.y, lun_tc.z).w;
	n_val.y = tex3D(volume_tex, run_tc.x, run_tc.y, run_tc.z).w;
	n_val.z = tex3D(volume_tex, lln_tc.x, lln_tc.y, lln_tc.z).w;
	n_val.w = tex3D(volume_tex, rln_tc.x, rln_tc.y, rln_tc.z).w;
	f_val.x = tex3D(volume_tex, luf_tc.x, luf_tc.y, luf_tc.z).w;
	f_val.y = tex3D(volume_tex, ruf_tc.x, ruf_tc.y, ruf_tc.z).w;
	f_val.z = tex3D(volume_tex, llf_tc.x, llf_tc.y, llf_tc.z).w;
	f_val.w = tex3D(volume_tex, rlf_tc.x, rlf_tc.y, rlf_tc.z).w;

	// output derivative in x direction
	f_val = (f_val - n_val) * hg_z.z;
	tmp = lerp2(make_float2(f_val.x,f_val.y), make_float2(f_val.z, f_val.w), hg_y.z);
	f_val.w = (tmp.y - tmp.x) * hg_x.z;

	
	hg2_x = hg_x;
	hg2_y = hg_y_dev;
	hg2_z = hg_z_dev;

	//__________________________
	//Hesse yz
	// setup texture coordinate
	// left/right upper/lower near/far
	llf_tc = coord - make_float3(hg2_x.x/size_source.x, 0.0, 0.0);
	rlf_tc = coord + make_float3(hg2_x.y/size_source.x, 0.0, 0.0);
              
	luf_tc = llf_tc - make_float3(0.0, hg2_y.x/size_source.y, 0.0);
	ruf_tc = rlf_tc - make_float3(0.0, hg2_y.x/size_source.y, 0.0);
	llf_tc = llf_tc + make_float3(0.0, hg2_y.y/size_source.y, 0.0);
	rlf_tc = rlf_tc + make_float3(0.0, hg2_y.y/size_source.y, 0.0);
              
	lun_tc = luf_tc - make_float3(0.0, 0.0, hg2_z.x/size_source.z);
	run_tc = ruf_tc - make_float3(0.0, 0.0, hg2_z.x/size_source.z);
	lln_tc = llf_tc - make_float3(0.0, 0.0, hg2_z.x/size_source.z);
	rln_tc = rlf_tc - make_float3(0.0, 0.0, hg2_z.x/size_source.z);
	luf_tc = luf_tc + make_float3(0.0, 0.0, hg2_z.y/size_source.z);
	ruf_tc = ruf_tc + make_float3(0.0, 0.0, hg2_z.y/size_source.z);
	llf_tc = llf_tc + make_float3(0.0, 0.0, hg2_z.y/size_source.z);
	rlf_tc = rlf_tc + make_float3(0.0, 0.0, hg2_z.y/size_source.z);

	// read values
	n_val.x = tex3D(volume_tex, lun_tc.x, lun_tc.y, lun_tc.z).w;
	n_val.y = tex3D(volume_tex, run_tc.x, run_tc.y, run_tc.z).w;
	n_val.z = tex3D(volume_tex, lln_tc.x, lln_tc.y, lln_tc.z).w;
	n_val.w = tex3D(volume_tex, rln_tc.x, rln_tc.y, rln_tc.z).w;
	f_val.x = tex3D(volume_tex, luf_tc.x, luf_tc.y, luf_tc.z).w;
	f_val.y = tex3D(volume_tex, ruf_tc.x, ruf_tc.y, ruf_tc.z).w;
	f_val.z = tex3D(volume_tex, llf_tc.x, llf_tc.y, llf_tc.z).w;
	f_val.w = tex3D(volume_tex, rlf_tc.x, rlf_tc.y, rlf_tc.z).w;

	// output derivative in y direction
	f_val = (f_val - n_val) * hg_z.z;
	tmp = (make_float2(f_val.z, f_val.w) - make_float2(f_val.x,f_val.y)) * hg_y.z;
	yz = lerp2(tmp.x, tmp.y, hg_x.z);

	float3x3 matrix;
	matrix.m[0].x = xx;
	matrix.m[0].y = xy;
	matrix.m[0].z = xz;
	matrix.m[1].x = xy;
	matrix.m[1].y = yy;
	matrix.m[1].z = yz;
	matrix.m[2].x = xz;
	matrix.m[2].y = yz;
	matrix.m[2].z = zz;
	return matrix;
}

__device__
float3 calculatePCD(float3 point)
{
	float3 g = -calculateGradient(point, make_float3(128.0, 128.0, 128.0));
	float lengthg = length(g);
	float3 n = -g/lengthg;	
	
	float3x3 pn;	pn.m[0].x = n.x;	pn.m[0].y = n.y;	pn.m[0].z = n.z;
	float3x3 P = minus(identity3x3(1.0), mul(pn,transpose(pn))); //sicher richtig (daher sollte der rest auch passen :)
	float3x3 H = calcHesse(point, make_float3(128.0));
	float3x3 G = mul( mul(mul(P,H),P), -lengthg);
	
	float T = G.m[0].x + G.m[1].y;// + G.m[2].z;	
	
	float F = 0.0;	
	//for(int i = 0; i < 3; i++)	for(int j = 0; j < 3; j++)		F += G.m[i][j]*G.m[i][j];	//http://en.wikipedia.org/wiki/Matrix_norm#Frobenius_norm
	float3x3 tmpf = mul(G, transpose(G));
	F = tmpf.m[0].x + tmpf.m[1].y;
	F = sqrt(F);

	float tmp = sqrt((2.0*F*F) - (T*T));
	float kappa1 = (T + tmp) / 2.0;
	float kappa2 = (T - tmp) / 2.0;

	float3x3 tmp2 = minus(G, identity3x3(kappa1, kappa2));
	//float3x3 tmp3 = mul(inverse(mul(transpose(tmp2),tmp2)),transpose(tmp2));
	float3x3 tmp3 = inverse(tmp2);

	//tmp3 = transpose(tmp3);
	float3 tmp4;
	if((kappa1) > (kappa2))	tmp4 = make_float3(tmp3.m[0].x, tmp3.m[0].y, tmp3.m[0].z);
	else					tmp4 = make_float3(tmp3.m[1].x, tmp3.m[1].y, tmp3.m[1].z);
	tmp4 = make_float3(tmp3.m[0].x, tmp3.m[1].x, tmp3.m[2].y);
	return tmp4;
	
	/*/
	
	float3 g = -calculateGradient(point, make_float3(128.0, 128.0, 128.0));
	float lengthg = length(g);
	float3x3 Ht = calcHesse(point, make_float3(128.0));
	
	float3 n = -g/lengthg;	
	
	Matrix3x3 pn = Matrix3x3(n.x, 0.0, 0.0, n.y, 0.0, 0.0, n.z, 0.0, 0.0);
	Matrix3x3 P = Matrix3x3(1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0) - ((pn)*Transposed(pn)); //sicher richtig (daher sollte der rest auch passen :)
	
	Matrix3x3 H = Matrix3x3(Ht.m[0].x, Ht.m[0].y, Ht.m[0].z, Ht.m[1].x, Ht.m[1].y, Ht.m[1].z, Ht.m[2].x, Ht.m[2].y, Ht.m[2].z);
	Matrix3x3 G = (P*H*P)*(-1.0/lengthg);

	Vector3 eigValues = Vector3(); //Vector3(kappa1, kappa2, 0.0);
	Matrix3x3 Gv = EigenvectorDecomposition(G, eigValues);
	//Matrix3x3 Gv = EigenvectorDecomposition(H, eigValues);

	return make_float3(Gv.m00, Gv.m10, Gv.m20);
	//*/
}

//REGEL 1: Gradient 
struct thrust_Gradient
{
    __device__ thrust_Gradient() {}

    template <typename Tuple>
	__device__ void operator()(Tuple t)
    {
        float4 vel = thrust::get<1>(t);
		float4 pos = thrust::get<2>(t);
		
		float4 grad = tex3D(volume_tex, pos.x, pos.y, pos.z);
		//float3 tmp2 = calculateGradient(make_float3(pos), make_float3(128.0,128.0,128.0));
		//grad.x = tmp2.x;
		//grad.y = tmp2.y;
		//grad.z = tmp2.z;

		float3 D = make_float3(vel.x, vel.y, vel.z);
		float3 new_D = D;

		//if(tex1D(transferfunction_tex, grad.w).w >= 0.01) 
		{
			//D_(n+1) = D_n - (G*D_n)G
			//D... Particle Direction
			//G... Gradient direction
			float3 G = normalize(make_float3(grad.x, grad.y, grad.z));
			
			new_D = D - (dot(G,D)*G);
				
			//Normalize D_(n+1)
			new_D = normalize(new_D);
		}

		thrust::get<0>(t) = make_float4(new_D.x, new_D.y, new_D.z, vel.w);
    }
};

//REGEL 2: Principial Curvature Directions
struct thrust_PCD
{
    __device__ thrust_PCD() {}

    template <typename Tuple>
	__device__ void operator()(Tuple t)
    {
        float4 old_vel = thrust::get<1>(t);
		float4 pos = thrust::get<2>(t);

		//D_(n+1) = ((1-k_c c_m)) D_n + (C(k_c c_m))
		//D... Particle Direction
		//C... Principal curvature direction
		//k_c... scale factor which specify the influence of PCD
		//c_m... magnitude of the PCD		

		float3 eigenvectors = calculatePCD(make_float3(pos));

		float c_m = length(eigenvectors);
		float w = particle_parameter.k_c*c_m;
		float3 new_vel = ((1.0-w)*make_float3(old_vel.x, old_vel.y, old_vel.z)) + (eigenvectors*w);		

        thrust::get<0>(t) = make_float4(new_vel, old_vel.w);
    }
};

//REGEL 2: Principial Curvature Directions
struct thrust_PCD2
{
    __device__ thrust_PCD2() {}

    template <typename Tuple>
	__device__ void operator()(Tuple t)
    {
        float4 old_vel = thrust::get<1>(t);
		float4 pos = thrust::get<2>(t);

		//D_(n+1) = ((1-k_c c_m)) D_n + (C(k_c c_m))
		//D... Particle Direction
		//C... Principal curvature direction
		//k_c... scale factor which specify the influence of PCD
		//c_m... magnitude of the PCD		

		float4 grad = tex3D(pcd_tex, pos.x, pos.y, pos.z);
		float3 eigenvectors = make_float3(grad.x, grad.y, grad.z);
		float c_m = length(grad);				

		float w = particle_parameter.k_c*c_m;
		float3 new_vel = ((1.0-w)*make_float3(old_vel.x, old_vel.y, old_vel.z)) + (eigenvectors*w);

        thrust::get<0>(t) = make_float4(new_vel, old_vel.w);
    }
};

__device__
float4 CalcualteFlockingVector(int3 gridPos, uint index, float4* oldPos, float4* oldVel, uint* cellStart, uint* cellEnd)
{
	uint gridindex = calcGridIndex(gridPos);

    // get start of bucket for this cell
    uint start = cellStart[gridindex];
	uint end = cellEnd[gridindex];

	float3 vi = make_float3(oldVel[index]);		//v...velocity
	float3 ri = make_float3(oldPos[index]);		//rho... denisty

	float num_neighboors = 0.0;
	float4 force = make_float4(0.0);
	
	if(start != 0xffffffff)
	{
		for(uint j=start; j<end && j < (start+10); j++) 
		{
			if (j != index)        
			{
				float3 rj = make_float3(oldPos[j]);
				float len = length(ri - rj);
				if(len < 1.0/64.f)
				{
					float4 vel2 = oldVel[j];
					force = force + vel2;
					num_neighboors += 1.0;
				}
			}
		}
    }

	return make_float4(force.x, force.y, force.z, num_neighboors);
}

//Regel 3a: Flocking
//Frs Grid hab ich mal die Variante vom CUDA SDK genommen. 
__global__ void CalculatingFlocking(float4* newVel, float4* oldPos, float4* oldVel, uint* cellStart, uint* cellEnd, int numParticles) 
{
	uint index = __mul24(blockIdx.x,blockDim.x) + threadIdx.x;
    if (index >= numParticles) return;    
    
    // read particle data from sorted arrays
	float3 old_Pos = make_float3(oldPos[index]);
    float4 old_Vel = oldVel[index];

    // get address in grid
    int3 gridPos = calcGridPos(old_Pos);

	float4 flocking_vec = make_float4(0.0, 0.0, 0.0, 0.0);
	for(int x = -1; x < 1; x++)
	{
		for(int y = -1; y < 1; y++)
		{
			for(int z = -1; z < 1; z++)
			{
				int3 neighbourPos = gridPos + make_int3(x, y, z);
				if(neighbourPos.x >= 0 && neighbourPos.y >= 0 && neighbourPos.z >= 0 && neighbourPos.x < 64 && neighbourPos.y < 64 && neighbourPos.z < 64)
				{
					flocking_vec += CalcualteFlockingVector(gridPos, index, oldPos, oldVel, cellStart, cellEnd);
				}
			}
		}
	}
	float4 nVel = make_float4(0.0);
	if(flocking_vec.w > 0.0)
	{
		float4 force = make_float4(flocking_vec.x, flocking_vec.y, flocking_vec.z, 0.0) / flocking_vec.w;
		
		//D_(n+1) = ((1-k_f)D_n) + (k_f F)
		//D... Particle Direction
		//F... Flocking Vector  (Flock direction for each particle is calculated as the average of the neighbooring particles weighted by distance).
		//k_f... constant, which specify the influence of flocking vector
		nVel = ((1.0-particle_parameter.k_f)*old_Vel) + (particle_parameter.k_f * force);
	}
	else
	{
		nVel = old_Vel;
	}

	newVel[index] = make_float4(nVel.x, nVel.y, nVel.z, old_Vel.w);
}

//Regel 3b: predefined direction
struct thrust_PredefinedDirection
{
    __device__ thrust_PredefinedDirection() {}

    template <typename Tuple>
	__device__ void operator()(Tuple t)
    {
        float4 old_vel = thrust::get<1>(t);

		//D_(n+1) = ((1-k_p)D_n) + (k_p P)
		//D... Particle Direction
		//P... Predefined Direction
		//k_p... constant, which specify the influence of predefined direction

        float3 new_vel = make_float3(old_vel.x, old_vel.y, old_vel.z);
		new_vel = (new_vel * (1.0-particle_parameter.k_p)) + (particle_parameter.predefinedDirection*particle_parameter.k_p);

        thrust::get<0>(t) = make_float4(new_vel, old_vel.w);
    }
};


__device__
float4 CalcualteDensityVector(int3 gridPos, uint index, float4* oldPos, uint* cellStart, uint* cellEnd)
{
	uint gridindex = calcGridIndex(gridPos);
    uint start = cellStart[gridindex];
	uint end = cellEnd[gridindex];

	float num_neighboors = 0.0;
	float3 force = make_float3(0.0);    
	float r = 1.0/64.f;

	float4 pos1 = oldPos[index];

	if(start != 0xffffffff)
	{
		for(uint j=start; j<end && j < (start+10); j++) 
		{
			if (j != index)
			{
				float4 pos2 = oldPos[j];				
				
				float3 f = make_float3(pos1 - pos2);
				float len = length(f);
				f = normalize(f);

				if(len < r)
				{
					len = r-len;
					force += f*len*2.0;
					num_neighboors += 1.0;
				}
			}
		}
    }

	return make_float4(force, num_neighboors);
}

//Regel 4a: Update Density
//Frs Grid hab ich mal die Variante vom CUDA SDK genommen. 
__global__ void UpdateDensity(float4* newPos, float4* oldPos, uint* cellStart, uint* cellEnd, int numParticles) 
{
	uint index = __mul24(blockIdx.x,blockDim.x) + threadIdx.x;
    if (index >= numParticles) return;    

	float4 old_Pos = oldPos[index];
	//GET NEIGHBORING PARTICLES	
	int3 gridPos = calcGridPos( make_float3(old_Pos));

	float4 force = make_float4(0.0f);
	for(int x = -1; x < 1; x++)
	{
		for(int y = -1; y < 1; y++)
		{
			for(int z = -1; z < 1; z++)
			{
				int3 neighbourPos = gridPos + make_int3(x, y, z);
				if(neighbourPos.x >= 0 && neighbourPos.y >= 0 && neighbourPos.z >= 0 && neighbourPos.x < 64 && neighbourPos.y < 64 && neighbourPos.z < 64)
				{
					force += CalcualteDensityVector(gridPos, index, oldPos, cellStart, cellEnd);
				}
			}
		}
	}

	float4 nPos = make_float4(0.0);

	if(force.w > 0.0)
	{		
		//D_(n+1) = ((1-k_f)D_n) + (k_f F)
		//D... Particle Direction
		//F... Flocking Vector  (Flock direction for each particle is calculated as the average of the neighbooring particles weighted by distance).
		//k_f... constant, which specify the influence of flocking vector
		force = force / force.w;
		nPos = old_Pos + (particle_parameter.k_d*force);
	}
	else	nPos = old_Pos;
	
	newPos[index] = make_float4(nPos.x, nPos.y, nPos.z, old_Pos.w);
}


//Integriere Partikel
struct thrust_Integrate
{
    __device__ thrust_Integrate() {}

    template <typename Tuple>
	__device__ void operator()(Tuple t)
    {
        float4 old_pos = thrust::get<1>(t);
        float4 vel = thrust::get<2>(t);
		unsigned int start = thrust::get<3>(t);

        float3 new_pos = make_float3(old_pos.x, old_pos.y, old_pos.z);
		float3 act_vel = normalize(make_float3(vel.x, vel.y, vel.z));

        float age = old_pos.w;	
        
        new_pos += act_vel * particle_parameter.elapsed_time;

		float3 dim = make_float3(128.0,128,128.0);
		float3 grid = make_float3(dim.x, dim.y, dim.z);
		int gridx = dim.x;
		int gridy = dim.y;

		int indX = int(start & (gridx-1));		
		int indY = int(start & ((gridx*gridy)-1));
			indY = indY / gridx;
		int indZ = start / (gridx*gridy); 

		float3 pos = make_float3(float(indX)/grid.x, float(indY)/grid.y, float(indZ)/grid.z);
		
		float r = 1.0/64.0;
		age += 0.001;
		if(age != age || age > 1.0 || new_pos.x != new_pos.x  || new_pos.y != new_pos.y || new_pos.z != new_pos.z || new_pos.x < r || new_pos.y < r || new_pos.z < r || new_pos.x > (1.0-r) || new_pos.y > (1.0-r) || new_pos.z > (1.0-r))
		{	
			new_pos = pos;	
			age = 0.0;
		}

		//if (age > lifetime) age = lifetime;

        thrust::get<0>(t) = make_float4(new_pos, age);
    }
};

//Berechne Tiefe mit Sortierungsvektor
struct thrust_CalcDepth
{
    __device__ thrust_CalcDepth() {}

    template <typename Tuple>
    __device__ void operator()(Tuple t)
    {
        float4 pos = thrust::get<0>(t);
		float3 m_pos = make_float3(pos.x, pos.y, pos.z);
		float depth = -dot(m_pos, particle_parameter.sortVector);

		thrust::get<1>(t) = depth;
    }
};

//Suche alle Mglichen Startpositionen fr das Partikel
__global__ void InitParticles(uint* block, cudaExtent res, int bx, float iso) 
{
	int x = threadIdx.x + (blockIdx.x % bx) * blockDim.x;
    int y = threadIdx.y + blockIdx.y * blockDim.y;
	int z = threadIdx.z + (blockIdx.x / bx) * blockDim.x;

	if(x > 1 && y > 1 && z > 1 && x < (res.width-1) && y < (res.height-1) && z < (res.depth-1))
	{
		float tx = (float)(x) / (float)(res.width);
		float ty = (float)(y) / (float)(res.height);
		float tz = (float)(z) / (float)(res.depth);

		float center	= tex3D( volume_tex, tx, ty, tz).w;
		float right		= tex3D( volume_tex, tx-1.0/(float)(res.width), ty, tz).w;
		float left		= tex3D( volume_tex, tx+1.0/(float)(res.width), ty, tz).w;
		float top		= tex3D( volume_tex, tx, ty+1.0/(float)(res.height), tz).w;
		float down		= tex3D( volume_tex, tx, ty-1.0/(float)(res.height), tz).w;
		float front		= tex3D( volume_tex, tx, ty, tz+1.0/(float)(res.depth)).w;
		float behind	= tex3D( volume_tex, tx, ty, tz-1.0/(float)(res.depth)).w;
		
		if(center > iso && (right > iso || left < iso || top < iso || down < iso || front < iso || behind < iso))
		{
			block[(z*res.width*res.height) + (y*res.width) + x] = (z*res.width*res.height) + (y*res.width) + x;
		}
		else
		{
			block[(z*res.width*res.height) + (y*res.width) + x] = 0xffffffff;
		}
	}
	else
	{
		block[z*res.width*res.height + y*res.width + x] = (res.width*res.height*res.depth)+1;
	}
}


//Berechne den Indices von der Startposition
struct convertFloatToIndices
{
	const float max_size;
	convertFloatToIndices (float _a) : max_size(_a) {}

	__host__ __device__
	unsigned int operator ()( const float& x) const 
	{
		float it = max_size * x;
		return (unsigned int)(it);
	}
};

//Berechne die Position und Velocity von der Startposition
struct thrust_CalcPosAndVel
{
	const float3 dim;
	thrust_CalcPosAndVel (float3 _dim) : dim(_dim) {}

	template <typename Tuple>
	__device__ void operator()(Tuple t)
	{
		uint index = thrust::get<0>(t);
		float3 grid = make_float3(dim.x, dim.y, dim.z);
		int gridx = dim.x;
		int gridy = dim.y;

		int indX = int(index & (gridx-1));		
		int indY = int(index & ((gridx*gridy)-1));
			indY = indY / gridx;
		int indZ = index / (gridx*gridy); 

		float4 v = tex3D(volume_tex, float(indX)/grid.x, float(indY)/grid.y, float(indZ)/grid.z);

		float4 pos = make_float4(float(indX)/grid.x, float(indY)/grid.y, float(indZ)/grid.z,0.0);
		//float4 vel = make_float4(v.x, v.y, v.z,0.0);
		float4 vel = make_float4(0.0, 0.0, 0.0, 0.0);

		thrust::get<1>(t) = pos;
		thrust::get<2>(t) = vel;
	}
};

extern "C" void InitParticlesCuda(uint* block, cudaExtent res, float isosurface)
{
    int nThreads = 8;
	dim3 threads;
	threads.x = nThreads;
	threads.y = nThreads;
	threads.z = nThreads;
	
	dim3 blocks;
	blocks.x = (res.width/nThreads + (!(res.width%nThreads)?0:1))
			 * (res.depth/nThreads + (!(res.depth%nThreads)?0:1));
	blocks.y = res.height/nThreads + (!(res.height%nThreads)?0:1);
	blocks.z = 1;
	
	int bx = ceil((float)res.width/(float)nThreads);

	InitParticles<<<blocks,threads>>>(block, res, bx, isosurface);
}

//Suche alle Mglichen Startpositionen fr das Partikel
extern "C" void GenerateQRSequenceCuda(uint* points, unsigned int count)
{
	float* X;
	cudaMalloc((void**) &X , count*sizeof(float));

	curandStatus_t curandResult;
    curandGenerator_t prng;
    curandResult = curandCreateGenerator(&prng, CURAND_RNG_PSEUDO_DEFAULT);
    if (curandResult != CURAND_STATUS_SUCCESS)
    {
        //PROBLEM
    }

    curandResult = curandSetPseudoRandomGeneratorSeed(prng, 1234);
    if (curandResult != CURAND_STATUS_SUCCESS)
    {
        //PROBLEM
    }

    curandResult = curandGenerateUniform(prng, (float *)X, count);
    
    if (curandResult != CURAND_STATUS_SUCCESS)
    {
        //PROBLEM
    }

    curandResult = curandDestroyGenerator(prng);
    if (curandResult != CURAND_STATUS_SUCCESS)
    {
        //PROBLEM
    }

	thrust::device_ptr<float> dev_ptr(X);
	thrust::device_ptr<uint> points_ptr(points);
    thrust::transform(dev_ptr, dev_ptr+count, points_ptr, convertFloatToIndices((float)(count)));

	cutilSafeCall(cudaFree(X));
}

extern "C" void SortStartPositionsAfterIndicesCuda(uint* positions, uint* indices, uint count)
{
	thrust::device_ptr<unsigned int> m_positions(positions);
	thrust::device_ptr<unsigned int> m_indices(indices);

	thrust::sort_by_key
	(
		thrust::device_ptr<uint>(m_indices),
        thrust::device_ptr<uint>(m_indices + count),
        thrust::device_ptr<uint>(m_positions)
	);
}

extern "C" void CaclulatePositionsAndVelocities(uint* indices, float4* pos, float4* vel, cudaExtent res, unsigned int count)
{
	thrust::device_ptr<uint> m_indices(indices);
    thrust::device_ptr<float4> m_pos(pos);
	thrust::device_ptr<float4> m_vel(vel);

	thrust::for_each
	(
        thrust::make_zip_iterator(thrust::make_tuple(m_indices, m_pos, m_vel)),
        thrust::make_zip_iterator(thrust::make_tuple(m_indices+count, m_pos+count, m_vel+count)),
        thrust_CalcPosAndVel(make_float3(res.width,res.height,res.depth))
	);
}

extern "C" void ParticlesGradientCuda(float4* pos, float4* vel, float4* newvel, unsigned int count)
{
    thrust::device_ptr<float4> m_newVel(newvel);
    thrust::device_ptr<float4> m_oldVel(vel);
	thrust::device_ptr<float4> m_pos(pos);

    thrust::for_each
	(
        thrust::make_zip_iterator(thrust::make_tuple(m_newVel, m_oldVel, m_pos)),
        thrust::make_zip_iterator(thrust::make_tuple(m_newVel+count, m_oldVel+count, m_pos+count)),
        thrust_Gradient()
	);
}

extern "C" void ParticlesPCDCuda(float4* pos, float4* vel, float4* newvel, unsigned int count)
{
	thrust::device_ptr<float4> m_newVel(newvel);
    thrust::device_ptr<float4> m_oldVel(vel);
	thrust::device_ptr<float4> m_pos(pos);

    thrust::for_each
	(
        thrust::make_zip_iterator(thrust::make_tuple(m_newVel, m_oldVel, m_pos)),
        thrust::make_zip_iterator(thrust::make_tuple(m_newVel+count, m_oldVel+count, m_pos+count)),
        thrust_PCD()
	);
}

extern "C" void ParticlesPCD2Cuda(float4* pos, float4* vel, float4* newvel, unsigned int count)
{
	thrust::device_ptr<float4> m_newVel(newvel);
    thrust::device_ptr<float4> m_oldVel(vel);
	thrust::device_ptr<float4> m_pos(pos);

    thrust::for_each
	(
        thrust::make_zip_iterator(thrust::make_tuple(m_newVel, m_oldVel, m_pos)),
        thrust::make_zip_iterator(thrust::make_tuple(m_newVel+count, m_oldVel+count, m_pos+count)),
        thrust_PCD2()
	);
}

extern "C" void ParticlesFlockingCuda(float4* pos, float4* vel, uint* cellStart, uint* cellEnd, float4* newvel, unsigned int count)
{
	uint numThreads = min(64, count);
    uint numBlocks = Floor(count, numThreads);

    CalculatingFlocking<<<numBlocks, numThreads>>>(newvel, pos, vel, cellStart, cellEnd, count);
}


extern "C" void ParticlesPredefinedDirectionCuda(float4* vel, float4* newvel, unsigned int count)
{
	thrust::device_ptr<float4> m_newVel(newvel);
    thrust::device_ptr<float4> m_oldVel(vel);

    thrust::for_each
	(
        thrust::make_zip_iterator(thrust::make_tuple(m_newVel, m_oldVel)),
        thrust::make_zip_iterator(thrust::make_tuple(m_newVel+count, m_oldVel+count)),
        thrust_PredefinedDirection()
	);
}

extern "C" void IntegrateParticlesCuda(float4* pos, float4* newpos, float4* vel, unsigned int* start, unsigned int count)
{
	thrust::device_ptr<float4> m_newPos(newpos);
    thrust::device_ptr<float4> m_oldPos(pos);
    thrust::device_ptr<float4> m_vel(vel);
	thrust::device_ptr<unsigned int> m_start(start);

    thrust::for_each
	(
        thrust::make_zip_iterator(thrust::make_tuple(m_newPos, m_oldPos, m_vel, m_start)),
        thrust::make_zip_iterator(thrust::make_tuple(m_newPos+count, m_oldPos+count, m_vel+count, m_start+count)),
        thrust_Integrate()
	);
}

extern "C" void ParticlesDensityCuda(float4* pos, uint* cellStart, uint* cellEnd, float4* newpos, unsigned int count)
{
	uint numThreads = min(64, count);
    uint numBlocks = Floor(count, numThreads);

    UpdateDensity<<<numBlocks, numThreads>>>(newpos, pos, cellStart, cellEnd, count);
}


extern "C" void CalcParticlesDepthCuda(	float4* pos, float* keys, unsigned int* indices, int count)
{
    thrust::device_ptr<float4> m_pos(pos);
    thrust::device_ptr<float> m_keys(keys);
    thrust::device_ptr<unsigned int> m_Index(indices);

    thrust::for_each
	(
        thrust::make_zip_iterator(thrust::make_tuple(m_pos, m_keys)),
        thrust::make_zip_iterator(thrust::make_tuple(m_pos+count, m_keys+count)), 
		thrust_CalcDepth()
	);

    thrust::sequence(m_Index, m_Index + count);
}

extern "C" void SortParticlesCuda(float* gridIndex, unsigned int *indices, unsigned int numParticles)
{
    thrust::sort_by_key
	(
		thrust::device_ptr<float>(gridIndex),
        thrust::device_ptr<float>(gridIndex + numParticles),
        thrust::device_ptr<unsigned int>(indices)
	);
}


extern "C" void SetParticle(Particle *hostparticle_parameter)
{
	cutilSafeCall( cudaMemcpyToSymbol(particle_parameter, hostparticle_parameter, sizeof(Particle)) );
}

struct thrust_CalculateGridPos
{
    __device__ thrust_CalculateGridPos() {}

    template <typename Tuple>
	__device__ void operator()(Tuple t)
    {
        float4 p = thrust::get<1>(t);

	    int3 gridPos = calcGridPos(make_float3(p.x, p.y, p.z));
		uint gridIndex = calcGridIndex(gridPos);

        thrust::get<0>(t) = gridIndex;
    }
};

__global__
void ReorderDataAndCellStartEnd(uint* cellStart, uint* cellEnd, float4* newPos, float4* newVel, float3* newForce, uint *particleIndex, uint *gridParticleIndex, float4* oldPos, float4* oldVel, float3* oldForce, uint numParticles)
{
	extern __shared__ uint sharedgridIndex[];    // blockSize + 1 elements
    uint index = __umul24(blockIdx.x,blockDim.x) + threadIdx.x;
	uint gridIndex;    
    if (index < numParticles) 
	{
        gridIndex = particleIndex[index];
	    sharedgridIndex[threadIdx.x+1] = gridIndex;

	    if (index > 0 && threadIdx.x == 0)	sharedgridIndex[0] = particleIndex[index-1];
	}

	__syncthreads();
	
	if (index < numParticles) 
	{
	    if (index == 0 || gridIndex != sharedgridIndex[threadIdx.x])
	    {
		    cellStart[gridIndex] = index;
            if(index > 0)	cellEnd[sharedgridIndex[threadIdx.x]] = index;
	    }

        if (index == numParticles - 1) cellEnd[gridIndex] = index + 1;

	    uint sortedIndex = gridParticleIndex[index];
        newPos[index] = oldPos[sortedIndex];
        newVel[index] = oldVel[sortedIndex];
		newForce[index] = oldForce[sortedIndex];
	}
}

__global__
void calcParticleGridIndex(uint* particleGridIndex, uint* gridParticleIndex, float4* pos, uint numParticles)
{
    uint index = __umul24(blockIdx.x, blockDim.x) + threadIdx.x;
    if (index >= numParticles) return;
    
    volatile float4 p = pos[index];
    int3 gridPos = calcGridPos(make_float3(p.x, p.y, p.z));
    uint gridIndex = calcGridIndex(gridPos);

    particleGridIndex[index] = gridIndex;
    gridParticleIndex[index] = index;
}

extern "C" void CalcBBGrid(uint* cellStart, uint*  cellEnd, float4* newPos, float4* newVel, float3* newForce, uint* indices, uint* gridIndex, float4* pos, float4* vel, float3* force, uint count, uint numCells)
{
	thrust::device_ptr<uint> m_gridIndex(gridIndex);
	thrust::device_ptr<float4> m_pos(pos);

	/*
	thrust::sequence(indices_ptr, indices_ptr + count);
	
	thrust::for_each
	(
        thrust::make_zip_iterator(thrust::make_tuple(m_gridIndex, m_pos)),
        thrust::make_zip_iterator(thrust::make_tuple(m_gridIndex+count, m_pos+count)), 
		thrust_CalculateGridPos()
	);
	*/

	uint numThreads = min(256, count);
    uint numBlocks = Floor(count, numThreads);
    calcParticleGridIndex<<< numBlocks, numThreads >>>(gridIndex, indices, pos, count);
	
	thrust::device_ptr<uint> indices_ptr(indices);
	thrust::sort_by_key(m_gridIndex, m_gridIndex+count, indices_ptr);

	uint sharedmemorysize = sizeof(uint)*(numThreads+1);
	ReorderDataAndCellStartEnd<<< numBlocks, numThreads, sharedmemorysize>>>(cellStart, cellEnd, newPos, newVel, newForce, gridIndex, indices, pos, vel,  force, count);
}

__device__
float EvaluateDensityPressure(int3 gridPos, uint index, float4* oldPos, uint* cellStart, uint* cellEnd)
{
	float densSum = 0.0;
	float h = 1.0/64.0;;
	
	uint gridindex = calcGridIndex(gridPos);
    uint start = cellStart[gridindex];

	float3 ri = make_float3(oldPos[index]);

	if(start != 0xffffffff)
	{
		uint end = cellEnd[gridindex];
		for(uint j=start; j<end && j < (start+40); j++) 
		{
			if (j != index) 
			{
				float3 rj = make_float3(oldPos[j]);
				
				float h2 = powf(h,2.0);
				float3 tmp = (ri - rj);
				float r2 = powf(tmp.x,2.0) + powf(tmp.y,2.0) + powf(tmp.z,2.0);

				if(r2 < h2) 
				{
					densSum += powf(h2-r2,3.0);
				}
            }
        }
    }

	return densSum;
}

__device__
float4 CalcualteForcePressureViscosity(int3 gridPos, uint index, float4* oldPos, float4* oldVel, float* oldPressure, uint* cellStart, uint* cellEnd)
{
	float h = 1.0/64.0;
	uint gridindex = calcGridIndex(gridPos);
    uint start = cellStart[gridindex];

	float pi = oldPressure[index];			//p...presure
	float3 vi = make_float3(oldVel[index]);	//v...velocity
	float3 ri = make_float3(oldPos[index]);	//rh_i... denisty

    float pres = 0.0;
	float3 vis = make_float3(0.0, 0.0, 0.0);

	float eps = 0.0000001;    
	if(start != 0xffffffff)
	{
		uint end = cellEnd[gridindex];
		for(uint j=start; j<end && (j < start+40); j++) 
		{
			if (j != index)        
			{
				float pj = oldPressure[j];				//p...presure
				float3 vj = make_float3(oldVel[j]);		//v... Velocity
				float3 rj = make_float3(oldPos[j]);		//r..pos
				float rhoj = oldPos[j].w;				//rho density

				float r = length(ri - rj);
				if(rhoj <= eps && rhoj >= -eps) rhoj = 1.0;

				if(r < h)
				{
					pres += (pi+pj)/(2.0*rhoj) * powf(h-r,3.0);
					vis += (vj-vi)/rhoj * ((-powf(r,3.0)/(2.0*powf(h,3.0))) + (powf(r,2.0)/powf(h,2.0)) + (h/(2.0*r)) - 1.0);
				}
			}
        }
    }

	return make_float4(vis.x, vis.y, vis.z, pres);
}


__global__
void EvaluateDensityAndPressure(float4*  pos, float*  pressure, float4* oldPos, uint*   cellStart, uint*   cellEnd, uint    numParticles)
{
    uint index = __mul24(blockIdx.x,blockDim.x) + threadIdx.x;
    if (index >= numParticles) return;    

    int3 gridPos = calcGridPos( make_float3(oldPos[index]));

	float density = 0.0;
    float3 force = make_float3(0.0f);
    for(int z=-1; z<=1; z++) 
	{
        for(int y=-1; y<=1; y++) 
		{
            for(int x=-1; x<=1; x++) 
			{
                int3 neighbourPos = gridPos + make_int3(x, y, z);
				if(neighbourPos.x >= 0 && neighbourPos.y >= 0 && neighbourPos.z >= 0 && neighbourPos.x < 64 && neighbourPos.y < 64 && neighbourPos.z < 64)
				{
					density += EvaluateDensityPressure(gridPos, index, oldPos, cellStart, cellEnd);
				}
            }
        }
    }

	float par_mass = particle_parameter.mass;
	float par_rho0 = particle_parameter.rho0;
	float par_sift = particle_parameter.siftness;
	//par_mass = 0.01;
	//par_rho0 = 0.01;
	//par_sift = 1.1;


	float h = 1.0/64.0;
	float PI = 3.14159265;
	float rhoi = density*315.0/(64.0*PI*powf(h,9.0))*par_mass;
	pos[index] = make_float4(oldPos[index].x, oldPos[index].y, oldPos[index].z, rhoi);
	pressure[index] = par_sift * (rhoi - par_rho0);
}

__global__
void CalcualteForcePressureAndViscosity(float* pressure, float3* viscosity, float4* oldPos, float4* oldVel, float* oldPressure, uint* cellStart, uint* cellEnd, uint numParticles)
{
    uint index = __mul24(blockIdx.x,blockDim.x) + threadIdx.x;
    if (index >= numParticles) return;    
    
	float3 pos = make_float3(oldPos[index]);
    float3 vel = make_float3(oldVel[index]);

    int3 gridPos = calcGridPos(make_float3(oldPos[index]));
	
    float3 force = make_float3(0.0f);
	float pres = 0.0;
	float3 vis = make_float3(0.0, 0.0, 0.0);
    for(int z=-1; z<=1; z++) 
	{
        for(int y=-1; y<=1; y++) 
		{
            for(int x=-1; x<=1; x++) 
			{
                int3 neighbourPos = gridPos + make_int3(x, y, z);
				if(neighbourPos.x >= 0 && neighbourPos.y >= 0 && neighbourPos.z >= 0 && neighbourPos.x < 64 && neighbourPos.y < 64 && neighbourPos.z < 64)
				{
					float4 res = make_float4(0.0, 0.0, 0.0, 0.0); 
					res = CalcualteForcePressureViscosity(gridPos, index, oldPos, oldVel, oldPressure, cellStart, cellEnd);

					vis.x += res.x;	vis.y += res.y;	vis.z += res.z;
					pres += res.w;
				}
            }
        }
    }

	float par_mass = particle_parameter.mass;
	float par_vis = particle_parameter.viscosity;
	//par_mass = 0.01;
	//par_vis = 0.1;

	//float3 mean = make_float3(0.0,0.0,0.0);
	float h = 1.0/64.0;
	float PI = 3.14159265;
	pressure[index] =  -pres * (15.0/(PI*powf(h,6.0)));	//Spiky Kernel
	viscosity[index] =  vis * par_vis * par_mass * (15.0/(2.0*PI*powf(h,3.0)));// * mean; //Viscosity Kernel
}


extern "C" void EvaluateDensityAndPressureCuda(float4*  pos, float*  pressure, float4* oldPos, uint*   cellStart, uint*   cellEnd, uint    numParticles, cudaExtent res)
{
	uint numThreads = min(64, numParticles);
    uint numBlocks = Floor(numParticles, numThreads);

	EvaluateDensityAndPressure<<<numBlocks, numThreads>>>(pos, pressure, oldPos, cellStart, cellEnd, numParticles);
}

extern "C" void CalcualteForcePressureAndViscosityCuda( float* pressure, float3* viscosity, float4* oldPos,	float4* oldVel,	float* oldPressure, uint*   cellStart, uint*   cellEnd, uint    numParticles, cudaExtent res)
{
	uint numThreads = min(64, numParticles);
    uint numBlocks = Floor(numParticles, numThreads);

	CalcualteForcePressureAndViscosity<<<numBlocks, numThreads>>>(pressure, viscosity, oldPos, oldVel, oldPressure, cellStart, cellEnd, numParticles);
}


//Integriere Partikel
struct thrust_IntegrateNavier
{
    __device__ thrust_IntegrateNavier() {}

    template <typename Tuple>
	__device__ void operator()(Tuple t)
    {
        float4 old_pos = thrust::get<2>(t);
        float4 old_vel = thrust::get<3>(t);
		float pressure = thrust::get<4>(t);
        float3 viscosity = thrust::get<5>(t);
		float3 acc_old = thrust::get<6>(t);
		unsigned int start = thrust::get<8>(t);
		float3 force = thrust::get<7>(t);

		float dt = particle_parameter.elapsed_time;
		float par_mass = particle_parameter.mass;
		float par_gra = particle_parameter.gravity;
		float par_damp = particle_parameter.DampingCoefficientCT;
		//par_mass = 0.01;
		//par_gra = 0.1;
		//par_damp = 0.2;
		dt = 0.005;

		float3 acc  = viscosity + force + pressure;
		//float3 acc  = force;
		acc.y -= par_gra*par_mass;

		//http://en.wikipedia.org/wiki/Leapfrog_integration
		//x_i+1 = x_i + v_i * dt + 0.5 * a_i * dt^2
		//v_i+1 = v_i + 0.5 * (a_i + a_i+1) * dt
		float3 new_pos = make_float3(old_pos.x, old_pos.y, old_pos.z) + (make_float3(old_vel)*dt) + (acc_old*0.5*powf(dt,2.0));
		float3 new_vel = make_float3(old_vel) + ((acc+acc_old)*0.5*dt);
		/*/
		float3 new_vel = make_float3(old_vel) + (acc*dt);
		float3 new_pos = make_float3(old_pos.x, old_pos.y, old_pos.z) + (new_vel*dt);
		//*/
		

		float r = 1.0/64.0;		
		if(new_pos.x < r)		{	new_pos.x = r;		new_vel.x *= par_damp;}
		if(new_pos.x > (1.0-r)) {	new_pos.x = 1.0-r;	new_vel.x *= par_damp;}
		if(new_pos.y < r)		{	new_pos.y = r;		new_vel.y *= par_damp;}
		if(new_pos.y > (1.0-r)) {	new_pos.y = 1.0-r;	new_vel.y *= par_damp;}
		if(new_pos.z < r)	    {	new_pos.z = r;		new_vel.z *= par_damp;}
		if(new_pos.z > (1.0-r))	{	new_pos.z = 1.0-r;	new_vel.z *= par_damp;}
		
		if(	new_pos.x != new_pos.x || new_pos.y != new_pos.y || new_pos.z != new_pos.z || 
			new_vel.x != new_vel.x || new_vel.y != new_vel.y || new_vel.z != new_vel.z ||
			old_pos.x != old_pos.x || old_pos.y != old_pos.y || old_pos.z != old_pos.z || old_pos.w != old_pos.w)  
		{   
			
			float3 dim = make_float3(128.0,128,128.0);
			float3 grid = make_float3(dim.x, dim.y, dim.z);
			int gridx = dim.x;
			int gridy = dim.y;

			int indX = int(start & (gridx-1));		
			int indY = int(start & ((gridx*gridy)-1));
				indY = indY / gridx;
			int indZ = start / (gridx*gridy); 

			new_pos = make_float3(float(indX)/grid.x, float(indY)/grid.y, float(indZ)/grid.z);
			new_vel = make_float3(0.0, 0.0, 0.0);

			/*/
			float3 r = make_float3((float)(start % 10000), (float)(start % 30000), (float)(start % 50000));
			r.x /= 10000.0;
			r.y /= 30000.0;
			r.z /= 50000.0;
			r -= make_float3(0.5, 0.5, 0.5);
			r /= 1.25;			

			r = make_float3(0.5, 0.6, 0.5)+r;

			new_pos = r;
			new_vel = make_float3(0.0,-0.01,0.0);
			*/
		}

        thrust::get<0>(t) = make_float4(new_pos, old_pos.w);
		thrust::get<1>(t) = make_float4(new_vel, 1.0);
		thrust::get<7>(t) = acc;
    }
};

extern "C" void IntegrateParticlesNavierStrokesCuda(float3* force, float3* actforce, float4* pos, float4* newpos, float4* vel,float4* newvel, float* pressure, float3* viscosity, unsigned int* startpos, unsigned int count)
{
	thrust::device_ptr<float4> m_newPos(newpos);
	thrust::device_ptr<float4> m_newVel(newvel);
	thrust::device_ptr<float3> m_actForce(actforce);

    thrust::device_ptr<float4> m_oldPos(pos);
	thrust::device_ptr<float4> m_vel(vel);
	thrust::device_ptr<float> m_oldPressure(pressure);
    thrust::device_ptr<float3> m_oldViscosity(viscosity);
	thrust::device_ptr<float3> m_oldForce(force);
	thrust::device_ptr<uint> m_startpos(startpos);

    thrust::for_each
	(
        thrust::make_zip_iterator(thrust::make_tuple(m_newPos, m_newVel, m_oldPos, m_vel, m_oldPressure, m_oldViscosity, m_oldForce, m_actForce, m_startpos)),
        thrust::make_zip_iterator(thrust::make_tuple(m_newPos+count, m_newVel+count, m_oldPos+count, m_vel+count, m_oldPressure+count, m_oldViscosity+count, m_oldForce+count,m_actForce+count, m_startpos+count)),
        thrust_IntegrateNavier()
	);
}

__device__
float3 CollisionDetection(int3 gridPos, uint index, float4* oldPos, float4* oldVel, uint* cellStart, uint* cellEnd, uint numParticles)
{
	uint gridindex = calcGridIndex(gridPos);

    // get start of bucket for this cell
    uint start = cellStart[gridindex];
	uint end = cellEnd[gridindex];
	float3 ri = make_float3(oldPos[index]);
	float3 vi = make_float3(oldVel[index]);
	float3 force = make_float3(0.0);

	float h = 1.0/64.0;
    
	if(start != 0xffffffff)
	{
		for(uint j=start; j < end && (j < start+40); j++) 
		{
			if (j != index) 
			{    
				float3 rj = make_float3(oldPos[j]);
				
				float3 relPos = rj - ri;
				float r = length(relPos);

				if(r < h) 
				{
					float3 norm = relPos / r;
					float3 relVel = make_float3(oldVel[j]) - vi;
					float3 relTanVel = relVel - (dot(relVel, norm) * norm);
				
					force -= particle_parameter.SpringCoefficient * (h - r) * norm;
					force += particle_parameter.DampingCoefficient * relVel;
					force += particle_parameter.ShearCoefficient * relTanVel;
					//force += 0.1*relPos;
				}
			}
        }
    }

	return force;
}

__global__
void CollisionDetectionCells(float3* newForce, float4* oldPos, float4* oldVel, uint*   cellStart, uint*   cellEnd, uint    numParticles)
{
    uint index = __mul24(blockIdx.x,blockDim.x) + threadIdx.x;
    if (index >= numParticles) return;    
    
    // read particle data from sorted arrays
	float3 pos = make_float3(oldPos[index]);
    float3 vel = make_float3(oldVel[index]);

    // get address in grid
    int3 gridPos = calcGridPos(make_float3(oldPos[index]));

    // examine neighbouring cells
    float3 force = make_float3(0.0);
    for(int z=-1; z<=1; z++) 
	{
        for(int y=-1; y<=1; y++) 
		{
            for(int x=-1; x<=1; x++) 
			{
                int3 neighbourPos = gridPos + make_int3(x, y, z);

				if(neighbourPos.x >= 0 && neighbourPos.y >= 0 && neighbourPos.z >= 0 && neighbourPos.x < 64 && neighbourPos.y < 64 && neighbourPos.z < 64)
				{
					force += CollisionDetection(neighbourPos, index, oldPos, oldVel, cellStart, cellEnd, numParticles);
				}
            }
        }
    }

	newForce[index] = make_float3(force.x, force.y, force.z);
}

extern "C" void CollisionDetectionCellsCuda( float3* force, float4* oldPos, float4* oldVel, uint*   cellStart, uint*   cellEnd, uint    numParticles, cudaExtent res)
{
	uint numThreads = min(64, numParticles);
    uint numBlocks = Floor(numParticles, numThreads);

	CollisionDetectionCells<<<numBlocks, numThreads>>>(force, oldPos, oldVel, cellStart, cellEnd, numParticles);
}

struct thrust_CollisionVolumne
{
    __device__ thrust_CollisionVolumne() {}

    template <typename Tuple>
	__device__ void operator()(Tuple t)
    {
        float4 old_pos = thrust::get<0>(t);
        float4 vel = thrust::get<1>(t);
		float3 force = thrust::get<2>(t);
		
		float3 ri = make_float3(old_pos.x, old_pos.y, old_pos.z); 
		float3 vi = make_float3(vel);

		float r = 1/64.0;
		float3 f = make_float3(0.0, 0.0, 0.0);
		if(ri.x >= 0.0 && ri.y >= 0.0 && ri.z >= 0.0 && ri.x <= 1.0 && ri.y <= 1.0 && ri.z <= 1.0)
		{
			float4 v = tex3D(volume_tex, ri.x, ri.y, ri.z);

			if(v.w > 0.2)
			{
				float3 vj = make_float3(v);
				//vj = normalize(vj)*0.1;
			
				float3 relPos = make_float3(v);
				float r = 1.0;

			
				float3 norm = relPos / r;
				float3 relVel = vj - vi;
				float3 relTanVel = relVel - (dot(relVel, norm) * norm);

				f = particle_parameter.SpringCoefficientCT * norm;
				f += particle_parameter.DampingCoefficientCT * relVel;
				f += particle_parameter.ShearCoefficientCT * relTanVel;
			}
		}
		thrust::get<2>(t) = force+f;
    }
};


extern "C" void CollisionDetectionVolumeCuda( float3* force, float4* pos, float4* vel, uint count)
{
    thrust::device_ptr<float4> m_oldPos(pos);
    thrust::device_ptr<float4> m_vel(vel);
	thrust::device_ptr<float3> m_newForce(force);

    thrust::for_each
	(
        thrust::make_zip_iterator(thrust::make_tuple(m_oldPos, m_vel, m_newForce)),
        thrust::make_zip_iterator(thrust::make_tuple(m_oldPos+count, m_vel+count, m_newForce+count)),
        thrust_CollisionVolumne()
	);
}

struct thrust_ResetVelocity
{
    __device__ thrust_ResetVelocity() {}

    template <typename Tuple>
	__device__ void operator()(Tuple t)
    {
        float4 old_vel = thrust::get<1>(t);
        float4 pos = thrust::get<2>(t);
		float4 vel = old_vel;

		if(old_vel.x != old_vel.x || old_vel.y != old_vel.y || old_vel.z != old_vel.z)	vel = tex3D(volume_tex, pos.x, pos.y, pos.z);

        thrust::get<0>(t) = make_float4(vel.x, vel.y, vel.z, old_vel.w);
    }
};

extern "C" void ResetVelocityOfErrorParticlesCuda(float4* new_vel, float4* old_vel, float4* pos, uint count)
{
	thrust::device_ptr<float4> m_new_vel(new_vel);
    thrust::device_ptr<float4> m_old_vel(old_vel);
	thrust::device_ptr<float4> m_pos(pos);

    thrust::for_each
	(
        thrust::make_zip_iterator(thrust::make_tuple(m_new_vel, m_old_vel, m_pos)),
        thrust::make_zip_iterator(thrust::make_tuple(m_new_vel+count, m_old_vel+count, m_pos+count)),
        thrust_ResetVelocity()
	);
}

extern "C" void initHGTexture(cudaArray* hgArray)
{
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float4>();

	bspline_hg_tex.normalized = true;             
    bspline_hg_tex.filterMode = cudaFilterModeLinear;      
	bspline_hg_tex.addressMode[0] = cudaAddressModeClamp ;  // clamp texture coordinates
	//bspline_hg_tex.addressMode[1] = cudaAddressModeWrap ;  // clamp texture coordinates
}

extern "C" void initHGDevTexture(cudaArray* hg_dev_Array)
{
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float4>();

	bspline_hg_dev_tex.normalized = true;                      
	bspline_hg_dev_tex.filterMode = cudaFilterModeLinear;      // linear interpolation
    bspline_hg_dev_tex.addressMode[0] = cudaAddressModeClamp ;  // clamp texture coordinates
	//bspline_hg_dev_tex.addressMode[1] = cudaAddressModeWrap ;  // clamp texture coordinates
}

extern "C" void BindHGCuda(cudaArray* hg_array)
{
	cudaChannelFormatDesc descFloat = cudaCreateChannelDesc<float4>();
	cutilSafeCall(cudaBindTextureToArray(bspline_hg_tex, hg_array, descFloat));
}

extern "C" void BindHGDevCuda(cudaArray* hg_dev_array)
{
	cudaChannelFormatDesc descFloat = cudaCreateChannelDesc<float4>();
	cutilSafeCall(cudaBindTextureToArray(bspline_hg_dev_tex, hg_dev_array, descFloat));
}

extern "C" void UnbindHGCuda()
{
	cutilSafeCall(cudaUnbindTexture(bspline_hg_tex));
}

extern "C" void UnbindHGDevCuda()
{
	cutilSafeCall(cudaUnbindTexture(bspline_hg_dev_tex));
}