#include "ParticleSystemCuda.h"

//#define DEBUG_PARTICLE_SYSTEM

ParticleSystemCuda::ParticleSystemCuda(int numParticles) : m_numParticles(numParticles), m_emitterIndex(0)
{
	res = make_cudaExtent(128,128,128);

	//Setzt die ganzen Arrays fr die Partikelberechnung
	std::cout << "Set Particle Parameter" << std::endl;
	m_particle_parameter = &Particle();
/*	m_particle_parameter->gravity = make_float3(0.f, 0.f, 1.0f);
    m_particle_parameter->globalDamping = 1.0f;
	m_particle_parameter->elapsed_time = 0.1f;
	m_particle_parameter->sortVector = make_float3(0.f,0.f,0.f);
	SetParticle(m_particle_parameter);*/

	std::cout << "Create Cuda Arrays" << std::endl;
	m_Pos = new CudaArray<float4>();
	m_Vel = new CudaArray<float4>();
	m_Index = new CudaArray<unsigned int>();
	m_Depth = new CudaArray<float>();
	m_Pressure = new CudaArray<float>();
	m_Viscosity = new CudaArray<float3>();
	m_Force = new CudaArray<float3>();

	m_CellStart = new CudaArray<uint>();
	m_CellEnd = new CudaArray<uint>();
	m_Indices2 = new CudaArray<uint>();
	m_Hash = new CudaArray<uint>();

	m_Pos->CreateBuffer(m_numParticles, true, true); 
	m_Vel->CreateBuffer(m_numParticles, true, true);
	m_Index->CreateBuffer(m_numParticles, true, false); 
	m_Depth->CreateBuffer(m_numParticles); 
	m_Pressure->CreateBuffer(m_numParticles, true, true);
	m_Viscosity->CreateBuffer(m_numParticles, true, false); 
	m_Force->CreateBuffer(m_numParticles, false, true);

	m_CellStart->CreateBuffer(64*64*64); 
	m_CellEnd->CreateBuffer(64*64*64); 
	m_Indices2->CreateBuffer(m_numParticles); 
	m_Hash->CreateBuffer(m_numParticles); 

	std::cout << "ParticleSystemCuda initialized" << std::endl;
}

ParticleSystemCuda::~ParticleSystemCuda(void)
{
	delete m_Pos;
	delete m_Vel;
	delete m_Depth;
	delete m_Index;
	delete m_Pressure;
	delete m_Viscosity;
	delete m_Force;

	delete m_CellStart;
	delete m_CellEnd;
	delete m_Indices2;
	delete m_Hash;
}

glm::vec3 random_points()
{
    glm::vec3 point = glm::vec3((rand() / (float) RAND_MAX)*2.0f-1.0f, (rand() / (float) RAND_MAX)*2.0f-1.0f, (rand() / (float) RAND_MAX)*2.0f-1.0f);
    while(glm::length(point) > 1.0f)
	{
        point =  glm::vec3((rand() / (float) RAND_MAX)*2.0f-1.0f, (rand() / (float) RAND_MAX)*2.0f-1.0f, (rand() / (float) RAND_MAX)*2.0f-1.0f);
	} 
    return point;
}

glm::vec3 ParticleSystemCuda::runEmitter(glm::vec3 pos)
{
	return runEmitter(pos, glm::vec3(0.f,0.0f,0.f));
}

glm::vec3 ParticleSystemCuda::runEmitter(glm::vec3 pos, glm::vec3 vel)
{
	glm::vec3 spread(0.05f, 0.0f, 0.05f);

    AddParticles(m_emitterIndex, pos, vel, spread, 0.1f, 5, 0.05f); 
    if (m_emitterIndex > m_numParticles-1)       m_emitterIndex = 0;

	return glm::vec3();
}

inline float frand()
{
    return rand() / (float) RAND_MAX;
}

void ParticleSystemCuda::InitCube(glm::vec3 origin, glm::vec3 size, glm::vec3 vel, float lifetime)
{
	/*
	for(unsigned int i=0; i < m_numParticles; i++) 
	{
		glm::vec3 pos = origin;
		pos += glm::vec3((rand() / (float) RAND_MAX)*2.0f-1.0f, (rand() / (float) RAND_MAX)*2.0f-1.0f, (rand() / (float) RAND_MAX)*2.0f-1.0f)*size;
		m_Pos->GetHost()[i] = make_float4(pos.x, pos.y, pos.z, 1.0f);
	    m_Vel->GetHost()[i] = make_float4(vel.x, vel.y, vel.z, lifetime);
	}*/

	srand(1973);
	float spacing = 1.f/32.f;
	float jitter = 0.01f/64.f;
    for(uint z=0; z<32; z++) 
	{
        for(uint y=0; y<32; y++) 
		{
            for(uint x=0; x<32; x++) 
			{
                uint i = (z*32*32) + (y*32) + x;
                
				if (i < m_numParticles) 
				{

					float x1 = (float)(x) * size.x;
					float y1 = (float)(y) * size.y;
					float z1 = (float)(z) * size.z;

                    m_Pos->GetHost()[i].x = origin.x + (spacing * x1) + (1.f/64.f) + (frand()*2.0f-1.0f)*jitter;
                    m_Pos->GetHost()[i].y = origin.y + (spacing * y1) + (1.f/64.f) + (frand()*2.0f-1.0f)*jitter;
                    m_Pos->GetHost()[i].z = origin.z + (spacing * z1) + (1.f/64.f) + (frand()*2.0f-1.0f)*jitter;
                    m_Pos->GetHost()[i].w = 0.0f;
					
					m_Vel->GetHost()[i].x = 0.0f;
				    m_Vel->GetHost()[i].y = -0.1f;
				    m_Vel->GetHost()[i].z = 0.0f;
				    m_Vel->GetHost()[i].w = 0.0f;
                }
            }
        }
    }

	m_Pos->Copy(false, 0, m_numParticles);
    m_Vel->Copy(false, 0, m_numParticles);
}

void ParticleSystemCuda::InitSphere(glm::vec3 origin, glm::vec3 vel, float size, float lifetime)
{
	for(unsigned int i=0; i < m_numParticles; i++) 
	{
		glm::vec3 pos = origin;
		pos = pos + (random_points()*size);
		vel = random_points()*glm::vec3(0.5);

		m_Pos->GetHost()[i] = make_float4(pos.x, pos.y, pos.z, 0.0f);
	    m_Vel->GetHost()[i] = make_float4(vel.x, vel.y, vel.z, lifetime);
	}

	m_Pos->Copy(false, 0, m_numParticles);
    m_Vel->Copy(false, 0, m_numParticles);
}
void ParticleSystemCuda::SetData(float* pos, float* vel, int numParticles)
{
	for(unsigned int i=0; i < numParticles; i++) 
	{
		m_Pos->GetHost()[i] = make_float4(pos[i*3], pos[i*3+1], pos[i*3+2], 0.0f);
	    m_Vel->GetHost()[i] = make_float4(vel[i*3], vel[i*3+1], vel[i*3+2], 1.0f);
	}

	m_Pos->Copy(false, 0, numParticles);
    m_Vel->Copy(false, 0, numParticles);
}

void ParticleSystemCuda::AddParticles(unsigned int &index, glm::vec3 _pos, glm::vec3 vel, glm::vec3 spread, float radius, int num_particles, float variance)
{
    unsigned int start = index; 
    unsigned int count = 0;

    for(unsigned int i=0; i<num_particles && index < m_numParticles; i++) 
	{
		glm::vec3 pos = _pos + random_points()*radius;
		glm::vec3 dir = random_points();
		vel += dir*spread;
		dir.y = abs(dir.y);

		m_Pos->GetHost()[index] = make_float4(pos.x, pos.y, pos.z, 0.f);
		m_Vel->GetHost()[index] = make_float4(vel.x, vel.y, vel.z, 1.f);

		index++;
        count++;
    }

    m_Pos->Copy(false, start, count);
	m_Vel->Copy(false, start, count);
}

void ParticleSystemCuda::Update(float duration)
{
	this->m_duration = duration;
	//int numberOFParticles = min(16*16*16, RenderStates::NumberOfParticles);
	int numberOFParticles = RenderStates::NumberOfParticles;


	if(!RenderStates::Play) return;

	if(RenderStates::ModusChanged == 2)
	{
		m_Pos->BindVBO();
		m_Vel->BindVBO();

		// set all cells to empty
		cutilSafeCall(cudaMemset(m_CellStart->GetDevRead(), 0xffffffff, 64*64*64*sizeof(uint)));	
		cutilSafeCall(cudaMemset(m_CellEnd->GetDevRead(),	0,			64*64*64*sizeof(uint)));	
		cutilSafeCall(cudaMemset(m_Force->GetDevWrite(), 0.f, numberOFParticles*sizeof(float3)));	

		//Berechnet das Grid
		CalcBBGrid(	m_CellStart->GetDevRead(), m_CellEnd->GetDevRead(), m_Pos->GetDevWrite(), m_Vel->GetDevWrite(), m_Force->GetDevWrite(), m_Indices2->GetDevRead(), m_Hash->GetDevRead(), m_Pos->GetDevRead(), m_Vel->GetDevRead(),  m_Force->GetDevRead(), numberOFParticles,	64*64*64);
		m_Pos->Swap();
		m_Vel->Swap();
		m_Force->Swap();

		//Berechnet die Regel 1 mit dem Gradienten
		if(RenderStates::ParticleGradientEnable) 
		{
			ParticlesGradientCuda(m_Pos->GetDevRead(), m_Vel->GetDevRead(), m_Vel->GetDevWrite(), numberOFParticles);
			m_Vel->Swap();
		}

		//Berechnet die Regel 2 mit der Principial Curvature Direction
		if(RenderStates::ParticleCurvatureEnable) 
		{
			ParticlesPCDCuda(m_Pos->GetDevRead(), m_Vel->GetDevRead(), m_Vel->GetDevWrite(), numberOFParticles);
			//ParticlesPCD2Cuda(m_Pos->GetDevRead(), m_Vel->GetDevRead(), m_Vel->GetDevWrite(), numberOFParticles);
			m_Vel->Swap();
		}
		//Berechnet die Regel 2 mit der Principial Curvature Direction
		if(RenderStates::ParticleCurvature2Enable) 
		{
			//ParticlesPCDCuda(m_Pos->GetDevRead(), m_Vel->GetDevRead(), m_Vel->GetDevWrite(), numberOFParticles);
			ParticlesPCD2Cuda(m_Pos->GetDevRead(), m_Vel->GetDevRead(), m_Vel->GetDevWrite(), numberOFParticles);
			m_Vel->Swap();
		}

		//Berechnet die Regel 3a mit Flocking
		if(RenderStates::ParticleFlockingEnable) 
		{	
			ParticlesFlockingCuda(m_Pos->GetDevRead(), m_Vel->GetDevRead(), m_CellStart->GetDevRead(), m_CellEnd->GetDevRead(), m_Vel->GetDevWrite(), numberOFParticles);
			m_Vel->Swap();
		}
	
		//Berechnet die Regel 3b mit Predefined Direction
		if(RenderStates::ParticlePredefinedEnable) 
		{	
			ParticlesPredefinedDirectionCuda(m_Vel->GetDevRead(), m_Vel->GetDevWrite(), numberOFParticles);
			m_Vel->Swap();
		}
		
		if(RenderStates::ParticleDensityEnable) 
		{
			ParticlesDensityCuda(m_Pos->GetDevRead(), m_CellStart->GetDevRead(), m_CellEnd->GetDevRead(), m_Pos->GetDevWrite(), numberOFParticles);
			m_Pos->Swap();
		}

		//Integriert die Position der Partikel
		IntegrateParticlesCuda(m_Pos->GetDevRead(), m_Pos->GetDevWrite(), m_Vel->GetDevRead(), m_initResPos->GetDevRead(), numberOFParticles);
		m_Pos->Swap();

		ResetVelocityOfErrorParticlesCuda(m_Vel->GetDevWrite(), m_Vel->GetDevRead(), m_Pos->GetDevRead(), numberOFParticles);
		m_Vel->Swap();

		//ReorderCuda(m_Vel->GetDevWrite(), m_Pos->GetDevWrite(), m_Vel->GetDevRead(), m_Pos->GetDevRead(), m_Indices2->GetDevRead(), numberOFParticles);
		//m_Pos->Swap();
		//m_Vel->Swap();

		m_Pos->UnbindVBO();
		m_Vel->UnbindVBO();

		//std::cout << "Debug" << std::endl;
		//m_Vel->Debug(0, numberOFParticles);
		//m_Pos->Debug(0, numberOFParticles);
		//std::cout << "Debug" << std::endl;
		//m_CellStart->Debug(0, 200);
	}
	else if(RenderStates::ModusChanged == 3)
	{
		m_Pos->BindVBO();
		m_Vel->BindVBO();
		m_Pressure->BindVBO();
		m_Viscosity->BindVBO();
	
		// set all cells to empty
		cutilSafeCall(cudaMemset(m_CellStart->GetDevRead(), 0xffffffff, 64*64*64*sizeof(uint)));	
		cutilSafeCall(cudaMemset(m_CellEnd->GetDevRead(), 0, 64*64*64*sizeof(uint)));	
		cutilSafeCall(cudaMemset(m_Viscosity->GetDevRead(), 0.f, numberOFParticles*sizeof(float3)));	
		cutilSafeCall(cudaMemset(m_Pressure->GetDevWrite(), 0.f, numberOFParticles*sizeof(float)));	m_Pressure->Swap();	
		cutilSafeCall(cudaMemset(m_Pressure->GetDevWrite(), 0.f, numberOFParticles*sizeof(float)));	m_Pressure->Swap();	
		
		//Berechnet das Grid
		CalcBBGrid(m_CellStart->GetDevRead(), m_CellEnd->GetDevRead(), m_Pos->GetDevWrite(), m_Vel->GetDevWrite(), m_Force->GetDevWrite(), m_Indices2->GetDevRead(), m_Hash->GetDevRead(), m_Pos->GetDevRead(), m_Vel->GetDevRead(), m_Force->GetDevRead(), numberOFParticles, 64*64*64);
		m_Pos->Swap();
		m_Vel->Swap();
		m_Force->Swap();
	
		EvaluateDensityAndPressureCuda(m_Pos->GetDevWrite(), m_Pressure->GetDevWrite(), m_Pos->GetDevRead(), m_CellStart->GetDevRead(), m_CellEnd->GetDevRead(), numberOFParticles, res);
		m_Pos->Swap();
		m_Pressure->Swap();	

		CalcualteForcePressureAndViscosityCuda( m_Pressure->GetDevWrite(), m_Viscosity->GetDevRead(), m_Pos->GetDevRead(), m_Vel->GetDevRead(), m_Pressure->GetDevRead(), m_CellStart->GetDevRead(), m_CellEnd->GetDevRead(), numberOFParticles, res);
		m_Pressure->Swap();
	
		if(RenderStates::FluidCollisionEnable)		CollisionDetectionCellsCuda( m_Force->GetDevWrite(), m_Pos->GetDevRead(), m_Vel->GetDevRead(), m_CellStart->GetDevRead(), m_CellEnd->GetDevRead(), numberOFParticles, res);
		else										cutilSafeCall(cudaMemset(m_Force->GetDevWrite(), 0.f, numberOFParticles*sizeof(float3)));
		if(RenderStates::FluidCollisionCTEnable)	CollisionDetectionVolumeCuda(m_Force->GetDevWrite(), m_Pos->GetDevRead(), m_Vel->GetDevRead(), m_numParticles);

		IntegrateParticlesNavierStrokesCuda(m_Force->GetDevRead(), m_Force->GetDevWrite(), m_Pos->GetDevRead(), m_Pos->GetDevWrite(), m_Vel->GetDevRead(), m_Vel->GetDevWrite(), m_Pressure->GetDevRead(), m_Viscosity->GetDevRead(), m_initResPos->GetDevRead(), numberOFParticles);
		m_Pos->Swap();
		m_Vel->Swap();
		m_Force->Swap();

		m_Pressure->UnbindVBO();
		m_Viscosity->UnbindVBO();	
		m_Pos->UnbindVBO();
		m_Vel->UnbindVBO();


		//DEBUG: AUSGABE AUF DIE CONSOLE
		//std::cout << "Debug" << std::endl;
		//m_Viscosity->Debug(0, numberOFParticles);
		//std::cout << "Debug" << std::endl;
		//Debug();
		//m_CellStart->Debug(0, 64*64*64);

	}

	m_time += duration;
} 


void ParticleSystemCuda::Debug()
{
	int count = 64*64*64;
	int sum = 0;
	int maxd = 0;

	m_CellStart->Copy(true, 0, count);
	m_CellEnd->Copy(true, 0, count);
		
	for(unsigned int i=0; i<count; i++) 
	{
		//if(abs(m_host[i].x+m_host[i].y+m_host[i].z) > 0.0)  printf("(%f) (%f) (%f) \n", m_host[i].x, m_host[i].y, m_host[i].z );
		//if(m_host[i].w >= 0.0)	printf("(%.4f) (%.4f) (%.4f) (%.4f) \n", m_host[i].x, m_host[i].y, m_host[i].z, m_host[i].w );
		unsigned int start = m_CellStart->GetHost()[i];
		unsigned int end = m_CellEnd->GetHost()[i];
		if(start < 0xffffffff)
		{
			unsigned int d = end-start;
			sum += d;
			if(d > maxd) maxd = d;
			//printf("%d %d - %d \n", m_CellStart->GetHost()[i], m_CellEnd->GetHost()[i], d);
		}
	}
	printf("Summe: %d - maxd: %d\n\r", sum, maxd);
}


void ParticleSystemCuda::DepthSort()
{
	//Berechnet die Tiefe und sortiert die Partikel anhand dessen.
	//Somit sind die Partikel sortiert und knnen transparent gerendert werden.
	/*
	m_Pos->BindVBO();
	m_Index->BindVBO();

    CalcParticlesDepthCuda(m_Pos->GetDevRead(), m_Depth->GetDevRead(), m_Index->GetDevRead(), m_numParticles);
	SortParticlesCuda(m_Depth->GetDevRead(), m_Index->GetDevRead(), m_numParticles);

	m_Pos->UnbindVBO();
	m_Index->UnbindVBO();
	*/
}

void ParticleSystemCuda::PrintParticles(unsigned int start, unsigned int count)
{
    m_Pos->Copy(true, 0, 0);
    float4 *pos = m_Pos->GetHost(); 

    //m_Vel->Copy(true, 0, 0);
    //float4 *vel = m_Vel->GetHost();

	m_Index->Copy(true, 0, 0);
	unsigned int* indices = m_Index->GetHost();

	m_Depth->Copy(true, 0, 0);
	float* key = m_Depth->GetHost();

    for(unsigned int i=start; i<start+count; i++) {
        printf("%d: ", i);
        printf("pos: (%f, %f, %f, %f)\n", pos[indices[i]].x, pos[indices[i]].y, pos[indices[i]].z, pos[indices[i]].w);
		printf("DOT: %f\n\r", key[i]);
        //printf("vel: (%.4f, %.4f, %.4f, %.4f)\n", vel[i].x, vel[i].y, vel[i].z, vel[i].w);
    }
	
}

void ParticleSystemCuda::InitStartPositions()
{
	/*
		Berechne alle mglichen Startpositionen im Volumen.

		Leider wird das im Paper


		Was ist eine mgliche Startposition?
		Voxels wo der Dichtewert bei ~0 ist und der Nachbar einen im Vgl. hohen Dichtewert hat.
		Bei InitParticlesCuda werden im Volumen genau solche Positionen extrahiert und in einem Array gespeichert.
		Falls es zutritt, wird die Koordinate im Array gespeichert, sonst wird eine Koordinate auerhalb gespeichert.
		Anschlieend werden alle falschen Koordinaten gelscht und in ein neuem Array gespeichert.
		Da die Koordinaten nun sortiert gespeichert sind, wird auch noch ein Zufallsarray gespeichert.
		Um eine gute, aber trotzdem zufllige Verteilung zu gewhrleisten, wird dabei ein Array mit Sobol' sequence Zahlen
		erstellt im Bereich [0,|mglicher Startpositionen|].

		Der Unterschied zwischen Random und Sobol' sequence Zahlen ist folgendes:
		Array mit: 0 1 2 3 4 5 6 7 8 9 10...
		Random: 0 1 2 3 4 5 6 7 8 9 10... mglich
		Pseudo Random: 0 3 49 29 10 1 39.. (schner verteilt als Random, aber nicht so schn wie das folgende...)
		Sobol' sequence: 0 80 10 19 2 ... (also schner um den Random-Datensatz verteilt)

		Mehr Infos: http://en.wikipedia.org/wiki/Sobol_sequence (Da sieht man schn den Unterschied zwischen Pseudo-Random und Sobel- Sequence)
	*/

	std::cout << "Init Startpositions" << std::endl;

	m_initPos = new CudaArray<uint>();
	m_initPos->CreateBuffer(res.width*res.height*res.depth);
	
	InitParticlesCuda(m_initPos->GetDevRead(), res, RenderStates::IsoSurface);
	m_initPos->Copy(true, 0, m_initPos->GetLength());
	std::cout << "InitParticlesCuda done" << std::endl;

	
	unsigned int* helper = new unsigned int[res.width*res.height*res.depth];
	unsigned int n = 0;
	for(unsigned int length = 0; length < res.width*res.height*res.depth; length++)
	{
		if(m_initPos->GetHost()[length] < res.width*res.height*res.depth)
		{
			helper[n] = m_initPos->GetHost()[length];
			n++;
		}		
	}
	delete m_initPos;
	
	std::cout << "Startpositions abgeschnitten" << std::endl;
	std::cout << "Number of Positions: " << n << std::endl;

	m_initResPos = new CudaArray<uint>();
	m_initResPos->CreateBuffer(n);
	
	std::cout << "CreateBuffer done" << std::endl;
	
	for(int i = 0; i < n; i++)	m_initResPos->GetHost()[i] = helper[i];
	delete[] helper;
	
	std::cout << "m_initResPos Host setted" << std::endl;
	m_initResPos->Copy(false, 0, n);

	std::cout << "m_initResPos" << std::endl;
	
	//Generate Quasi Random (QR) Sequence
	m_randomSequence = new CudaArray<uint>();
	m_randomSequence->CreateBuffer(n);
	GenerateQRSequenceCuda(m_randomSequence->GetDevRead(), n);

	std::cout << "GenerateQRSequenceCuda generated" << std::endl;
	
	//Sort Start Positions after QR Sequence
	SortStartPositionsAfterIndicesCuda(m_initResPos->GetDevRead(), m_randomSequence->GetDevRead(), n);
	delete m_randomSequence;

	std::cout << "Sort Start Positions after QR Sequence" << std::endl;
	
	
	//Berechne nun die Position und bestimme die Velocity
	m_Pos->BindVBO();
	m_Vel->BindVBO();
	CaclulatePositionsAndVelocities(m_initResPos->GetDevRead(), m_Pos->GetDevWrite(), m_Vel->GetDevWrite(), res,  m_numParticles); 
	m_Pos->Swap();
	m_Vel->Swap();
	m_Pos->UnbindVBO();
	m_Vel->UnbindVBO();
	/*/
	InitCube(glm::vec3(0.5,0.6,0.5), glm::vec3(0.2,0.2,0.2), glm::vec3(0.f,0.f,0.f), 1.f);
	m_Pos->Swap();
	m_Vel->Swap();
	InitCube(glm::vec3(0.5,0.6,0.5), glm::vec3(0.2,0.2,0.2), glm::vec3(0.f,0.f,0.f), 1.f);//*/

	/*InitCube(glm::vec3(0.0,0.0,0.0), glm::vec3(1.0,1.0,1.0), glm::vec3(0.f,0.f,0.f), 1.f);
	m_Pos->Swap();
	m_Vel->Swap();
	InitCube(glm::vec3(0.0,0.0,0.0), glm::vec3(1.0,1.0,1.0), glm::vec3(0.f,0.f,0.f), 1.f);//*/

	//m_Pos->Debug(0, 20);
	//FERTIG MIT DER BERECHNUNG DER STARTPOSITIONEN!!!

	std::cout << "StartPositions and Velocities calculated" << std::endl;
}

Particle* ParticleSystemCuda::Getparticle_parameter() 
{ 
	return m_particle_parameter; 
}

void ParticleSystemCuda::SetIsoValue(float v)
{
	InitStartPositions();
}