#pragma once
#include "Cuda.h"
#include "CudaArray.h"
#include "particleCuda.cuh"
#include "RenderStates.h"


extern "C" void SetParticle(Particle *hostparticle_parameter);
extern "C" void CalcBBGrid(uint*  cellStart, uint*  cellEnd, float4* sortedPos,float4* sortedVel, float3* sortedForce, uint* indices, uint* hash, float4* pos, float4* vel, float3* force, uint count, uint numCells);

//Initialize
extern "C" void InitParticlesCuda(uint* block, cudaExtent res, float isosurface);
extern "C" void SortStartPositionsAfterIndicesCuda(uint* block, uint* indices, unsigned int count);
extern "C" void GenerateQRSequenceCuda(uint* block, unsigned int count);

extern "C" void initTransferFunctionTexture(cudaArray* tfArray);
extern "C" void initVolumenTexture(cudaArray* volArray, cudaExtent volumeSize);
extern "C" void initPCDTexture(cudaArray* volArray, cudaExtent volumeSize);
extern "C" void BindVolumeCuda(cudaArray* volume_array);
extern "C" void UnbindVolumeCuda();
extern "C" void BindPCDCuda(cudaArray* volume_array);
extern "C" void UnbindPCDCuda();
extern "C" void BindTransferfunctionCuda(cudaArray* volume_array);
extern "C" void UnbindTransferfunctionCuda();

//Particles & Rules
extern "C" void ParticlesGradientCuda(float4* pos, float4* vel, float4* newvel, unsigned int count);
extern "C" void ParticlesPCDCuda(float4* pos, float4* vel, float4* newvel, unsigned int count);
extern "C" void ParticlesPCD2Cuda(float4* pos, float4* vel, float4* newvel, unsigned int count);
extern "C" void ParticlesFlockingCuda(float4* pos, float4* vel, uint* cellStart, uint* cellEnd, float4* newvel, unsigned int count);
extern "C" void ParticlesPredefinedDirectionCuda(float4* vel, float4* newvel, unsigned int count);
extern "C" void ParticlesDensityCuda(float4* pos, uint* cellStart, uint* cellEnd, float4* newpos, unsigned int count);

extern "C" void IntegrateParticlesCuda(float4* pos, float4* newpos, float4* vel, unsigned int* start, unsigned int count);
extern "C" void SortParticlesCuda(float *sortKeys, unsigned int *indices, unsigned int numParticles);
extern "C" void CalcParticlesDepthCuda( float4* pos, float* keys, unsigned int* indices, int numParticles);

extern "C" void CaclulatePositionsAndVelocities(uint* indices, float4* pos, float4* vel, cudaExtent res, unsigned int count); 

//Volume Block
extern "C" void CalcVolumeBlockCuda(float4* block, cudaExtent res, cudaExtent blocks_size, cudaExtent blocks_num);

//div
extern "C" void render_kernel(dim3 gridSize, dim3 blockSize, uint *d_output, uint imageW, uint imageH, float density, float brightness, float transferOffset, float transferScale);
extern "C" void copyInvViewMatrix(float *invViewMatrix, size_t sizeofMatrix);

extern "C" void IntegrateParticlesNavierStrokesCuda(float3* force, float3* actforce, float4* pos, float4* newpos, float4* vel, float4* newvel, float* pressure, float3* viscosity, unsigned int* startpos, unsigned int count);
extern "C" void CalcualteForcePressureAndViscosityCuda( float* pressure, float3* viscosity, float4* oldPos, float4* oldVel, float* oldPressure, uint*   cellStart, uint*   cellEnd, uint    numParticles, cudaExtent res);
extern "C" void EvaluateDensityAndPressureCuda(float4*  pos, float*  pressure, float4* oldPos, uint*   cellStart, uint*   cellEnd, uint    numParticles, cudaExtent res);
extern "C" void CollisionDetectionCellsCuda( float3* newVel, float4* oldPos, float4* oldVel, uint*   cellStart, uint*   cellEnd, uint    numParticles, cudaExtent res);

extern "C" void CollisionDetectionVolumeCuda(float3* force, float4* pos, float4* vel, uint    numParticles);

extern "C" void initHGTexture(cudaArray* hg_Array);
extern "C" void initHGDevTexture(cudaArray* hg_dev_Array);
extern "C" void BindHGCuda(cudaArray* hg_array);
extern "C" void BindHGDevCuda(cudaArray* hg_dev_array);
extern "C" void UnbindHGCuda();
extern "C" void UnbindHGDevCuda();
extern "C" void ResetVelocityOfErrorParticlesCuda(float4* new_vel, float4* oldVel, float4* oldPos, uint numParticles);

class ParticleSystemCuda
{
public:
	/**
	* Standardconstructor
	* @param numParticles number of Particles
	*/
	ParticleSystemCuda(int numParticles);
	/**
	* Standarddeconstructor
	*/
	~ParticleSystemCuda(void);

	/**
	 * Not used yet
	 */
	glm::vec3 runEmitter(glm::vec3 pos);
	/**
	 * Not used yet
	 */
	glm::vec3 runEmitter(glm::vec3 pos, glm::vec3 vel);
	/**
	 * Not used yet
	 */
	void ParticleSystemCuda::AddParticles(unsigned int &index, glm::vec3 pos, glm::vec3 vel, glm::vec3 spread, float r, int n, float lifetimeVariance);

	/**
	 * Not used yet
	 */
	void InitCube(glm::vec3 origin, glm::vec3 size, glm::vec3 vel, float lifetime);
	/**
	 * This method update the particle system
	 */
	void Update(float duration);
	/**
	 * Not used yet
	 */
	void DepthSort();
	/**
	 * Not used yet
	 */
	void InitSphere(glm::vec3 origin, glm::vec3 vel, float size, float lifetime);

	/**
	 * This method sets the init start positions
	 */
	void InitStartPositions();

	/**
	 * Not used yet
	 */
	void PrintParticles(unsigned int start, unsigned int count);

	/**
	 * Not used yet
	 */
	Particle* Getparticle_parameter();

	/**
	 * Iso Value changed
	 * @param v new iso value
	 **/
	void SetIsoValue(float v);

	Particle* m_particle_parameter;

	/**
	 * Initialize the data
	 * @param pos position
	 * @param vel velocity
	 * @param numParticles number of Particles
	 **/
	void SetData(float* pos, float* vel, int numParticles);

	/**
	 * Debug Mode
	 */
	void Debug();

	unsigned int m_numParticles;


	CudaArray<float4> *m_Pos;
	CudaArray<float4> *m_Vel;
	CudaArray<float> *m_Depth;
	CudaArray<unsigned int> *m_Index;

	CudaArray<float3> *m_Force;
	CudaArray<float> *m_Pressure;
	CudaArray<float3> *m_Viscosity;

	CudaArray<unsigned int>* m_CellStart;
	CudaArray<unsigned int>* m_CellEnd;
	CudaArray<unsigned int>* m_Indices2;
	CudaArray<unsigned int>* m_Hash;

	unsigned int m_emitterIndex;
	float3 m_sortVector;

	float m_time;
	float m_duration;

	cudaExtent res;


	CudaArray<uint>* m_initPos;
	CudaArray<uint>* m_initResPos;
	CudaArray<uint>* m_randomSequence;
};

