#include "RayCasting.h"
#include "SceneManager.h"
#include "Coordinaten.h"


RayCasting::RayCasting() : textureLoaded(false), transferFunction(false), screenHeight(0), screenWidth(0), light(false), mip(false), 
	average(false), phong(false), adaptive(false), xClipping(1.0), yClipping(1.0), zClipping(1.0), m_type(0)
{
	width = 512;
	height = 512;
	pbo = 0;     // OpenGL pixel buffer object
	tex = 0;     // OpenGL texture object
}

RayCasting::~RayCasting()
{
	delete texturShader;
}

void RayCasting::init()
{
	volumeWidth = 128;
	volumeHeight = 128;
	volumeDepth = 128;

	std::cout << "RayCasting::init" << std::endl;
	cudaGLSetGLDevice( cutGetMaxGflopsDeviceId() );
	//cutilCheckError( cutCreateTimer( &timer));


	std::cout << "Create Renderer" << std::endl;
	m_renderer = new RendererOpenGL();

	Light* m_light = new Light();

	cam = new Camera();
	cam->SetViewMatrix(glm::vec3(0.f,0.f,1.f), glm::vec3(0.f,0.f,0.f), glm::vec3(0.f,1.f, 0.f));
	cam->SetProjection(45.f, 1.f, 1.0f, 20.f);
	cam->SetViewport(800.f,800.f);

	Camera* cam_orth = new Camera();
	cam_orth->SetViewMatrix(glm::vec3(0.f,0.f,3.f), glm::vec3(0.f,0.f,0.f), glm::vec3(0.f,1.f, 0.f));
	cam_orth->SetProjection(-1.0f, 1.0f, -1.0f, 1.0f, 1.f, 10.f);
	cam_orth->SetViewport(800.f,800.f);

	MyShader* m_shader_direction = new MyShader("shader/directionShader");
	if (!m_shader_direction) 
	{
		cerr << "Could not compile direction shader program." << endl;
		return;
	}
	m_shader_direction->bind_frag_data_location("out_color");

	std::map<std::string, int> draw_depth_vbo;
	draw_depth_vbo.insert(std::pair<std::string, int>("vertex", VertexGeometry::AT_POSITIONS));
	draw_depth_vbo.insert(std::pair<std::string, int>("color", VertexGeometry::AT_COLORS));
	std::map<std::string, int> draw_depth_uniform;
	draw_depth_uniform.insert(std::pair<std::string, int>("model", View::UT_MODELMATRIX));	
	draw_depth_uniform.insert(std::pair<std::string, int>("view", View::UT_VIEWMATRIX));	
	draw_depth_uniform.insert(std::pair<std::string, int>("perspective", View::UT_PROJECTIONMATRIX));	
	 
	std::cout << "Create FBO" << std::endl;
	fbo_back = FBO::FBOPtr(new FBO(800,800, false, GL_RGBA8, GL_RGBA, GL_LINEAR, GL_LINEAR, Texture::TT_NORMAL));
	fbo_front = FBO::FBOPtr(new FBO(800,800, false, GL_RGBA8, GL_RGBA, GL_LINEAR, GL_LINEAR, Texture::TT_NORMAL));
	fbo_hdr = FBO::FBOPtr(new FBO(800,800, false, GL_RGBA32F, GL_RGBA, GL_LINEAR, GL_LINEAR, Texture::TT_MIPMAP));
	
	std::cout << "Create Box" << std::endl;
	Geometry* m_bb = Primitives::GetBox();
	m_bb->World = glm::mat4();
	m_bb->Local = glm::translate(0.f, 0.f, -2.f);
	m_bb->SetShader(0, m_shader_direction);
	m_bb->CreateDictShaderToVBO(0, draw_depth_vbo);
	m_bb->CreateDictShaderUniforms(0, draw_depth_uniform);
	
	m_bb_front = new View();
	m_bb_front->SetRoot(new INode(m_bb));
	//m_bb_front->SetCamera(SceneManager::get().getCamera());
	m_bb_front->SetCamera(cam);
	m_bb_front->SetLight(m_light);
	m_bb_front->SetPass(0);
	m_bb_front->AddFBO(fbo_front);

	m_bb_back = new View();
	m_bb_back->SetRoot(new INode(m_bb));
	//m_bb_back->SetCamera(SceneManager::get().getCamera());
	m_bb_back->SetCamera(cam);
	m_bb_back->SetLight(m_light);
	m_bb_back->SetPass(0);
	m_bb_back->AddFBO(fbo_back);

	float* pone = new float[1];
	float* pzero = new float[1];
	float* pstepsize = new float[1];
	pone[0] = 1.f;
	pzero[0] = 0.f;
	pstepsize[0] = 0.01f;


	std::cout << "Create rayCastingShader" << std::endl;
	MyShader* m_shader_raycasting = new MyShader("shader/rayCastingShader");
	if (!m_shader_raycasting) 
	{
		cerr << "Could not compile raycasting shader program." << endl;
		return;
	}
	m_shader_raycasting->bind_frag_data_location("out_color");

	std::map<std::string, int> raycasting_vbo;
	raycasting_vbo.insert(std::pair<std::string, int>("vertex", VertexGeometry::AT_POSITIONS));
	raycasting_vbo.insert(std::pair<std::string, int>("texCoordVolume", VertexGeometry::AT_TEXCOORD));
	std::map<std::string, int> raycasting_uniform;
	raycasting_uniform.insert(std::pair<std::string, int>("model", View::UT_MODELMATRIX));	
	raycasting_uniform.insert(std::pair<std::string, int>("view", View::UT_VIEWMATRIX));	
	raycasting_uniform.insert(std::pair<std::string, int>("perspective", View::UT_PROJECTIONMATRIX));
	raycasting_uniform.insert(std::pair<std::string, int>("entryTexture", View::UT_TEXTURE0));
	raycasting_uniform.insert(std::pair<std::string, int>("exitTexture", View::UT_TEXTURE1));
	raycasting_uniform.insert(std::pair<std::string, int>("colorTexture", View::UT_TEXTURE2));
	raycasting_uniform.insert(std::pair<std::string, int>("transferTexture", View::UT_TEXTURE3));
	raycasting_uniform.insert(std::pair<std::string, int>("tex_hg", View::UT_TEXTURE4));
	raycasting_uniform.insert(std::pair<std::string, int>("tex_hg_dev", View::UT_TEXTURE5));
	raycasting_uniform.insert(std::pair<std::string, int>("pcdTexture", View::UT_TEXTURE6));
	raycasting_uniform.insert(std::pair<std::string, int>("adaptive", View::UT_ADDAPTIVE));
	raycasting_uniform.insert(std::pair<std::string, int>("stepSize", View::UT_STEPSIZE));
	raycasting_uniform.insert(std::pair<std::string, int>("transfer", View::UT_TRANSFER));
	raycasting_uniform.insert(std::pair<std::string, int>("mip", View::UT_MIP));
	raycasting_uniform.insert(std::pair<std::string, int>("average", View::UT_AVERAGE));
	raycasting_uniform.insert(std::pair<std::string, int>("light", View::UT_LIGHT));
	raycasting_uniform.insert(std::pair<std::string, int>("phong", View::UT_PHONG));

	std::cout << "Create rayCastingShader" << std::endl;
	MyShader* m_shader_raycasting2D = new MyShader("shader/rayCastingShader2D");
	if (!m_shader_raycasting2D) 
	{
		cerr << "Could not compile raycasting shader 2D program." << endl;
		return;
	}
	m_shader_raycasting2D->bind_frag_data_location("out_color");

	std::map<std::string, int> raycasting2D_vbo;
	raycasting2D_vbo.insert(std::pair<std::string, int>("vertex", VertexGeometry::AT_POSITIONS));
	raycasting2D_vbo.insert(std::pair<std::string, int>("texCoordVolume", VertexGeometry::AT_TEXCOORD));
	std::map<std::string, int> raycasting2D_uniform;
	raycasting2D_uniform.insert(std::pair<std::string, int>("model", View::UT_MODELMATRIX));	
	raycasting2D_uniform.insert(std::pair<std::string, int>("view", View::UT_VIEWMATRIX));	
	raycasting2D_uniform.insert(std::pair<std::string, int>("perspective", View::UT_PROJECTIONMATRIX));
	raycasting2D_uniform.insert(std::pair<std::string, int>("entryTexture", View::UT_TEXTURE0));
	raycasting2D_uniform.insert(std::pair<std::string, int>("exitTexture", View::UT_TEXTURE1));
	raycasting2D_uniform.insert(std::pair<std::string, int>("colorTexture", View::UT_TEXTURE2));
	raycasting2D_uniform.insert(std::pair<std::string, int>("transferTexture", View::UT_TEXTURE3));
	raycasting2D_uniform.insert(std::pair<std::string, int>("tex_hg", View::UT_TEXTURE4));
	raycasting2D_uniform.insert(std::pair<std::string, int>("tex_hg_dev", View::UT_TEXTURE5));
	raycasting2D_uniform.insert(std::pair<std::string, int>("pcdTexture", View::UT_TEXTURE6));
	raycasting2D_uniform.insert(std::pair<std::string, int>("adaptive", View::UT_ADDAPTIVE));
	raycasting2D_uniform.insert(std::pair<std::string, int>("stepSize", View::UT_STEPSIZE));
	raycasting2D_uniform.insert(std::pair<std::string, int>("transfer", View::UT_TRANSFER));
	raycasting2D_uniform.insert(std::pair<std::string, int>("mip", View::UT_MIP));
	raycasting2D_uniform.insert(std::pair<std::string, int>("average", View::UT_AVERAGE));
	raycasting2D_uniform.insert(std::pair<std::string, int>("light", View::UT_LIGHT));
	raycasting2D_uniform.insert(std::pair<std::string, int>("phong", View::UT_PHONG));


	MyShader* m_shader_mip = new MyShader("shader/mip");
	if (!m_shader_mip) 
	{
		cerr << "Could not compile mip shader program." << endl;
		return;
	}
	m_shader_mip->bind_frag_data_location("out_color");

	std::map<std::string, int> mip_vbo;
	mip_vbo.insert(std::pair<std::string, int>("vertex", VertexGeometry::AT_POSITIONS));
	mip_vbo.insert(std::pair<std::string, int>("texCoordVolume", VertexGeometry::AT_TEXCOORD));
	std::map<std::string, int> mip_uniform;
	mip_uniform.insert(std::pair<std::string, int>("model", View::UT_MODELMATRIX));	
	mip_uniform.insert(std::pair<std::string, int>("view", View::UT_VIEWMATRIX));	
	mip_uniform.insert(std::pair<std::string, int>("perspective", View::UT_PROJECTIONMATRIX));
	mip_uniform.insert(std::pair<std::string, int>("entryTexture", View::UT_TEXTURE0));
	mip_uniform.insert(std::pair<std::string, int>("exitTexture", View::UT_TEXTURE1));
	mip_uniform.insert(std::pair<std::string, int>("colorTexture", View::UT_TEXTURE2));
	mip_uniform.insert(std::pair<std::string, int>("transferTexture", View::UT_TEXTURE3));
	mip_uniform.insert(std::pair<std::string, int>("tex_hg", View::UT_TEXTURE4));
	mip_uniform.insert(std::pair<std::string, int>("tex_hg_dev", View::UT_TEXTURE5));
	mip_uniform.insert(std::pair<std::string, int>("pcdTexture", View::UT_TEXTURE6));
	mip_uniform.insert(std::pair<std::string, int>("adaptive", View::UT_ADDAPTIVE));
	mip_uniform.insert(std::pair<std::string, int>("stepSize", View::UT_STEPSIZE));
	mip_uniform.insert(std::pair<std::string, int>("transfer", View::UT_TRANSFER));
	mip_uniform.insert(std::pair<std::string, int>("mip", View::UT_MIP));
	mip_uniform.insert(std::pair<std::string, int>("average", View::UT_AVERAGE));
	mip_uniform.insert(std::pair<std::string, int>("light", View::UT_LIGHT));
	mip_uniform.insert(std::pair<std::string, int>("phong", View::UT_PHONG));

	MyShader* m_shader_averaging = new MyShader("shader/averaging");
	if (!m_shader_averaging) 
	{
		cerr << "Could not compile averaging shader program." << endl;
		return;
	}
	m_shader_averaging->bind_frag_data_location("out_color");

	std::map<std::string, int> averaging_vbo;
	averaging_vbo.insert(std::pair<std::string, int>("vertex", VertexGeometry::AT_POSITIONS));
	averaging_vbo.insert(std::pair<std::string, int>("texCoordVolume", VertexGeometry::AT_TEXCOORD));
	std::map<std::string, int> averaging_uniform;
	averaging_uniform.insert(std::pair<std::string, int>("model", View::UT_MODELMATRIX));	
	averaging_uniform.insert(std::pair<std::string, int>("view", View::UT_VIEWMATRIX));	
	averaging_uniform.insert(std::pair<std::string, int>("perspective", View::UT_PROJECTIONMATRIX));
	averaging_uniform.insert(std::pair<std::string, int>("entryTexture", View::UT_TEXTURE0));
	averaging_uniform.insert(std::pair<std::string, int>("exitTexture", View::UT_TEXTURE1));
	averaging_uniform.insert(std::pair<std::string, int>("colorTexture", View::UT_TEXTURE2));
	averaging_uniform.insert(std::pair<std::string, int>("transferTexture", View::UT_TEXTURE3));
	averaging_uniform.insert(std::pair<std::string, int>("tex_hg", View::UT_TEXTURE4));
	averaging_uniform.insert(std::pair<std::string, int>("tex_hg_dev", View::UT_TEXTURE5));
	averaging_uniform.insert(std::pair<std::string, int>("pcdTexture", View::UT_TEXTURE6));
	averaging_uniform.insert(std::pair<std::string, int>("adaptive", View::UT_ADDAPTIVE));
	averaging_uniform.insert(std::pair<std::string, int>("stepSize", View::UT_STEPSIZE));
	averaging_uniform.insert(std::pair<std::string, int>("transfer", View::UT_TRANSFER));
	averaging_uniform.insert(std::pair<std::string, int>("mip", View::UT_MIP));
	averaging_uniform.insert(std::pair<std::string, int>("average", View::UT_AVERAGE));
	averaging_uniform.insert(std::pair<std::string, int>("light", View::UT_LIGHT));
	averaging_uniform.insert(std::pair<std::string, int>("phong", View::UT_PHONG));

	MyShader* m_shader_contour = new MyShader("shader/rayCastingShaderContour");
	if (!m_shader_contour) 
	{
		cerr << "Could not compile averaging shader program." << endl;
		return;
	}
	m_shader_contour->bind_frag_data_location("out_color");

	std::map<std::string, int> contour_vbo;
	contour_vbo.insert(std::pair<std::string, int>("vertex", VertexGeometry::AT_POSITIONS));
	contour_vbo.insert(std::pair<std::string, int>("texCoordVolume", VertexGeometry::AT_TEXCOORD));
	std::map<std::string, int> contour_uniform;
	contour_uniform.insert(std::pair<std::string, int>("model", View::UT_MODELMATRIX));	
	contour_uniform.insert(std::pair<std::string, int>("view", View::UT_VIEWMATRIX));	
	contour_uniform.insert(std::pair<std::string, int>("perspective", View::UT_PROJECTIONMATRIX));
	contour_uniform.insert(std::pair<std::string, int>("entryTexture", View::UT_TEXTURE0));
	contour_uniform.insert(std::pair<std::string, int>("exitTexture", View::UT_TEXTURE1));
	contour_uniform.insert(std::pair<std::string, int>("colorTexture", View::UT_TEXTURE2));
	contour_uniform.insert(std::pair<std::string, int>("transferTexture", View::UT_TEXTURE3));
	contour_uniform.insert(std::pair<std::string, int>("tex_hg", View::UT_TEXTURE4));
	contour_uniform.insert(std::pair<std::string, int>("tex_hg_dev", View::UT_TEXTURE5));
	contour_uniform.insert(std::pair<std::string, int>("pcdTexture", View::UT_TEXTURE6));
	contour_uniform.insert(std::pair<std::string, int>("adaptive", View::UT_ADDAPTIVE));
	contour_uniform.insert(std::pair<std::string, int>("stepSize", View::UT_STEPSIZE));
	contour_uniform.insert(std::pair<std::string, int>("transfer", View::UT_TRANSFER));
	contour_uniform.insert(std::pair<std::string, int>("mip", View::UT_MIP));
	contour_uniform.insert(std::pair<std::string, int>("average", View::UT_AVERAGE));
	contour_uniform.insert(std::pair<std::string, int>("light", View::UT_LIGHT));
	contour_uniform.insert(std::pair<std::string, int>("phong", View::UT_PHONG));

	MyShader* m_shader_exposure = new MyShader("shader/rayCastingShader_exposure");
	if (!m_shader_exposure) 
	{
		cerr << "Could not compile averaging shader program." << endl;
		return;
	}
	m_shader_exposure->bind_frag_data_location("out_color");

	std::map<std::string, int> exposure_vbo;
	exposure_vbo.insert(std::pair<std::string, int>("vertex", VertexGeometry::AT_POSITIONS));
	exposure_vbo.insert(std::pair<std::string, int>("texCoordVolume", VertexGeometry::AT_TEXCOORD));
	std::map<std::string, int> exposure_uniform;
	exposure_uniform.insert(std::pair<std::string, int>("model", View::UT_MODELMATRIX));	
	exposure_uniform.insert(std::pair<std::string, int>("view", View::UT_VIEWMATRIX));	
	exposure_uniform.insert(std::pair<std::string, int>("perspective", View::UT_PROJECTIONMATRIX));
	exposure_uniform.insert(std::pair<std::string, int>("entryTexture", View::UT_TEXTURE0));
	exposure_uniform.insert(std::pair<std::string, int>("exitTexture", View::UT_TEXTURE1));
	exposure_uniform.insert(std::pair<std::string, int>("colorTexture", View::UT_TEXTURE2));
	exposure_uniform.insert(std::pair<std::string, int>("transferTexture", View::UT_TEXTURE3));
	exposure_uniform.insert(std::pair<std::string, int>("tex_hg", View::UT_TEXTURE4));
	exposure_uniform.insert(std::pair<std::string, int>("tex_hg_dev", View::UT_TEXTURE5));
	exposure_uniform.insert(std::pair<std::string, int>("pcdTexture", View::UT_TEXTURE6));
	exposure_uniform.insert(std::pair<std::string, int>("adaptive", View::UT_ADDAPTIVE));
	exposure_uniform.insert(std::pair<std::string, int>("stepSize", View::UT_STEPSIZE));
	exposure_uniform.insert(std::pair<std::string, int>("transfer", View::UT_TRANSFER));
	exposure_uniform.insert(std::pair<std::string, int>("mip", View::UT_MIP));
	exposure_uniform.insert(std::pair<std::string, int>("average", View::UT_AVERAGE));
	exposure_uniform.insert(std::pair<std::string, int>("light", View::UT_LIGHT));
	exposure_uniform.insert(std::pair<std::string, int>("phong", View::UT_PHONG));
	
	std::cout << "Create Plane" << std::endl;
	Geometry* m_plane = Primitives::GetPlane();
	m_plane->World = glm::mat4();
	m_plane->Local = glm::translate(0.f, 0.f, -1.f);
	m_plane->SetShader(0, m_shader_raycasting);
	m_plane->CreateDictShaderToVBO(0, raycasting_vbo);
	m_plane->CreateDictShaderUniforms(0, raycasting_uniform);
	
	m_plane->SetShader(1, m_shader_raycasting2D);
	m_plane->CreateDictShaderToVBO(1, raycasting2D_vbo);
	m_plane->CreateDictShaderUniforms(1, raycasting2D_uniform);

	m_plane->SetShader(2, m_shader_mip);
	m_plane->CreateDictShaderToVBO(2, mip_vbo);
	m_plane->CreateDictShaderUniforms(2, mip_uniform);

	m_plane->SetShader(3, m_shader_averaging);
	m_plane->CreateDictShaderToVBO(3, averaging_vbo);
	m_plane->CreateDictShaderUniforms(3, averaging_uniform);

	m_plane->SetShader(4, m_shader_contour);
	m_plane->CreateDictShaderToVBO(4, contour_vbo);
	m_plane->CreateDictShaderUniforms(4, contour_uniform);

	m_plane->SetShader(5, m_shader_exposure);
	m_plane->CreateDictShaderToVBO(5, exposure_vbo);
	m_plane->CreateDictShaderUniforms(5, exposure_uniform);

	m_plane->m_TextureObjs.insert(std::pair<unsigned int, FBO::FBOPtr>(0, fbo_front));
	m_plane->m_TextureObjs.insert(std::pair<unsigned int, FBO::FBOPtr>(1, fbo_back));
	//m_plane->m_TextureObjs.insert(std::pair<unsigned int, FBO::FBOPtr>(2, fbo_back));
	//m_plane->m_TextureObjs.insert(std::pair<unsigned int, FBO::FBOPtr>(3, fbo_back));
	
	std::cout << "Create m_raycasting" << std::endl;
	m_raycasting = new View();
	m_raycasting->SetRoot(new INode(m_plane));
	m_raycasting->SetCamera(cam);
	m_raycasting->SetLight(m_light);
	m_raycasting->SetPass(0);
	m_raycasting->AddFBO(fbo_hdr);

		
	std::cout << "Create m_raycasting" << std::endl;
	m_raycasting2D = new View();
	m_raycasting2D->SetRoot(new INode(m_plane));
	m_raycasting2D->SetCamera(cam);
	m_raycasting2D->SetLight(m_light);
	m_raycasting2D->SetPass(1);
	m_raycasting2D->AddFBO(fbo_hdr);

	m_mip = new View();
	m_mip->SetRoot(new INode(m_plane));
	m_mip->SetCamera(cam);
	m_mip->SetLight(m_light);
	m_mip->SetPass(2);
	m_mip->AddFBO(fbo_hdr);

	m_averaging = new View();
	m_averaging->SetRoot(new INode(m_plane));
	m_averaging->SetCamera(cam);
	m_averaging->SetLight(m_light);
	m_averaging->SetPass(3);
	m_averaging->AddFBO(fbo_hdr);

	m_contour = new View();
	m_contour->SetRoot(new INode(m_plane));
	m_contour->SetCamera(cam);
	m_contour->SetLight(m_light);
	m_contour->SetPass(4);
	m_contour->AddFBO(fbo_hdr);

	m_exposure = new View();
	m_exposure->SetRoot(new INode(m_plane));
	m_exposure->SetCamera(cam);
	m_exposure->SetLight(m_light);
	m_exposure->SetPass(5);
	m_exposure->AddFBO(fbo_hdr);

	//_____ Volume Bricks
	std::cout << "Volume Bricks" << std::endl;
	MyShader* m_shader_volumeBricks = new MyShader("shader/volumeBricks");
	if (!m_shader_volumeBricks) 
	{
		cerr << "Could not compile volumeBricks shader program." << endl;
		return;
	}
	m_shader_volumeBricks->bind_frag_data_location("out_color");

	std::map<std::string, int> volumeBricks_vbo;
	volumeBricks_vbo.insert(std::pair<std::string, int>("in_bricksMinMax", VertexGeometry::AT_POSITIONS));
	std::map<std::string, int> volumeBricks_uniform;
	volumeBricks_uniform.insert(std::pair<std::string, int>("model", View::UT_MODELMATRIX));	
	volumeBricks_uniform.insert(std::pair<std::string, int>("view", View::UT_VIEWMATRIX));	
	volumeBricks_uniform.insert(std::pair<std::string, int>("perspective", View::UT_PROJECTIONMATRIX));
	volumeBricks_uniform.insert(std::pair<std::string, int>("transfertex", View::UT_TEXTURE0));
	volumeBricks_uniform.insert(std::pair<std::string, int>("volumeSize", View::UT_VOLUMESIZE));
	
	std::cout << "***************************" << std::endl;
	std::cout << "volumeWidth: " << volumeWidth << " << volumeHeight: " << volumeHeight << " volumeDepth: " << volumeDepth << std::endl;
	unsigned int m_numParticles = volumeWidth*volumeHeight*volumeDepth;
	
	std::cout << "volumeWidth: " << volumeWidth << " << volumeHeight: " << volumeHeight << " volumeDepth: " << volumeDepth << std::endl;

	/*glm::vec4* pos2 = new glm::vec4[m_numParticles];
	glm::vec3* norm2 = new glm::vec3[m_numParticles];
	glm::vec2* tex2 = new glm::vec2[m_numParticles];
	int* index2 = new int[m_numParticles];
	m_plane_tmp = new VertexGeometry(pos2, norm2, tex2, 0, 0, index2, (volumeWidth/16.f)*(volumeHeight/16.f)*(volumeDepth/16.f), (volumeWidth/16.f)*(volumeHeight/16.f)*(volumeDepth/16.f));
	/*/
	//Es werden sowieso die Cuda Array gerendert, daher mssen die hier nich mehr erstellt werden
	m_plane_tmp = new VertexGeometry((volumeWidth/16.f)*(volumeHeight/16.f)*(volumeDepth/16.f), (volumeWidth/16.f)*(volumeHeight/16.f)*(volumeDepth/16.f));//*/
	m_plane_tmp->m_rendertype = Geometry::RT_POINT;
	m_plane_tmp->World = glm::mat4();
	m_plane_tmp->Local = glm::translate(0.f, 0.f, -2.f)*glm::scale(2.f,2.f,2.f)*glm::translate(-0.5f, -0.5f, -0.5f);
	m_plane_tmp->SetShader(0, m_shader_volumeBricks);
	m_plane_tmp->CreateDictShaderToVBO(0, volumeBricks_vbo);
	m_plane_tmp->CreateDictShaderUniforms(0, volumeBricks_uniform);

	std::cout << "m_volumeBricks_front" << std::endl;
	m_volumeBricks_front = new View();
	m_volumeBricks_front->SetRoot(new INode(m_plane_tmp));
	m_volumeBricks_front->SetCamera(cam);
	m_volumeBricks_front->SetLight(m_light);
	m_volumeBricks_front->SetPass(0);
	m_volumeBricks_front->AddFBO(fbo_front);

	m_volumeBricks_back = new View();
	m_volumeBricks_back->SetRoot(new INode(m_plane_tmp));
	m_volumeBricks_back->SetCamera(cam);
	m_volumeBricks_back->SetLight(m_light);
	m_volumeBricks_back->SetPass(0);
	m_volumeBricks_back->AddFBO(fbo_back);

	std::cout << "create m_shader_hdr" << std::endl;
	MyShader* m_shader_hdr = new MyShader("shader/hdr");
	if (!m_shader_raycasting) 
	{
		cerr << "Could not compile raycasting shader program." << endl;
		return;
	}
	m_shader_raycasting->bind_frag_data_location("out_color");

	std::map<std::string, int> hdr_vbo;
	hdr_vbo.insert(std::pair<std::string, int>("vertex", VertexGeometry::AT_POSITIONS));
	hdr_vbo.insert(std::pair<std::string, int>("tex_coord", VertexGeometry::AT_TEXCOORD));
	std::map<std::string, int> hdr_uniform;
	hdr_uniform.insert(std::pair<std::string, int>("model", View::UT_MODELMATRIX));	
	hdr_uniform.insert(std::pair<std::string, int>("view", View::UT_VIEWMATRIX));	
	hdr_uniform.insert(std::pair<std::string, int>("perspective", View::UT_PROJECTIONMATRIX));
	hdr_uniform.insert(std::pair<std::string, int>("img", View::UT_TEXTURE0));
	hdr_uniform.insert(std::pair<std::string, int>("mode", View::UT_HDR_MODE));
	hdr_uniform.insert(std::pair<std::string, int>("enable", View::UT_HDR_ENABLE));
	
	Geometry* m_plane_hdr = Primitives::GetPlane();
	m_plane_hdr->World = glm::mat4();
	m_plane_hdr->Local = glm::translate(0.f, 0.f, -1.f);
	m_plane_hdr->SetShader(0, m_shader_hdr);
	m_plane_hdr->CreateDictShaderToVBO(0, hdr_vbo);
	m_plane_hdr->CreateDictShaderUniforms(0, hdr_uniform);

	float* pexposure = new float[1];
	pexposure[0] = 0.5f;

	std::cout << "create m_plane_hdr" << std::endl;
	m_plane_hdr->m_shader_uniforms.resize(1);
	m_plane_hdr->m_shader_uniforms[0].push_back(new ShaderObj("exposure", ShaderObj::PT_FLOAT, pexposure));
	m_plane_hdr->m_shader_uniforms[0].push_back(new ShaderObj("brightMax", ShaderObj::PT_FLOAT, pone));
	m_plane_hdr->CreateDictShaderSpecificUniforms(0);

	m_plane_hdr->m_TextureObjs.insert(std::pair<unsigned int, FBO::FBOPtr>(0, fbo_hdr));
	
	m_hdr = new View();
	m_hdr->SetRoot(new INode(m_plane_hdr));
	m_hdr->SetCamera(cam_orth);
	m_hdr->SetLight(m_light);
	m_hdr->SetPass(0);

	this->texturShader = new Shader();
	this->texturShader->loadCompileAndLinkShader("shader/colorgradientShader.vert", "shader/colorgradientShader.frag");
	this->texturShader->bind_frag_data_location("out_color");

	std::cout << "create particleShader" << std::endl;
	//______________ PARTICLE ______________ 
	MyShader* particleShader = new MyShader("shader/particle");
	if (!m_shader_direction) 
	{
		cerr << "Could not compile direction shader program." << endl;
		return;
	}
	m_shader_direction->bind_frag_data_location("out_color");

	std::map<std::string, int> particle_shader_vbo;
	particle_shader_vbo.insert(std::pair<std::string, int>("in_position", VertexGeometry::AT_POSITIONS));
	particle_shader_vbo.insert(std::pair<std::string, int>("in_velocity", VertexGeometry::AT_TEXCOORD));
	std::map<std::string, int> particle_shader_uniform;
	particle_shader_uniform.insert(std::pair<std::string, int>("model", View::UT_MODELMATRIX));	
	particle_shader_uniform.insert(std::pair<std::string, int>("view", View::UT_VIEWMATRIX));	
	particle_shader_uniform.insert(std::pair<std::string, int>("projection", View::UT_PROJECTIONMATRIX));	
	particle_shader_uniform.insert(std::pair<std::string, int>("color_tex", View::UT_TEXTURE0));
	particle_shader_uniform.insert(std::pair<std::string, int>("size", View::UT_PARTICLE_SIZE));
	particle_shader_uniform.insert(std::pair<std::string, int>("vol", View::UT_TEXTURE1));

	m_numParticles = 16*16*16;
	m_numParticles = RenderStates::NumberOfParticles;
	/*
	//Es werden sowieso die Cuda Array gerendert, daher mssen die hier nich mehr erstellt werden
	glm::vec4* pos = new glm::vec4[m_numParticles];
	glm::vec3* norm = new glm::vec3[m_numParticles];
	glm::vec2* tex = new glm::vec2[m_numParticles];
	int* index = new int[m_numParticles];*/
	
	m_particle_geo = new VertexGeometry(/*pos, norm, tex, 0, 0, index,*/ m_numParticles, m_numParticles);
	m_particle_geo->m_rendertype = Geometry::RT_POINT;
	m_particle_geo->World = glm::mat4();
	m_particle_geo->Local = glm::translate(0.f, 0.f, -2.f)*glm::scale(2.1f,2.1f,2.1f)*glm::translate(-0.5f, -0.5f, -0.5f);
	m_particle_geo->SetShader(0, particleShader);
	m_particle_geo->CreateDictShaderToVBO(0, particle_shader_vbo);
	m_particle_geo->CreateDictShaderUniforms(0, particle_shader_uniform);

	std::cout << "create ParticleSystemCuda" << std::endl;
	m_particle_cuda = new ParticleSystemCuda(m_numParticles);
	
	std::cout << "create m_particle" << std::endl;
	m_particle = new View();
	m_particle->SetRoot(new INode(m_particle_geo));
	m_particle->SetCamera(cam);
	m_particle->SetLight(m_light);
	m_particle->SetPass(0);
	m_particle->AddFBO(fbo_hdr);
	



	std::cout << "create clippingPlane" << std::endl;
	MyShader* m_shader_cp = new MyShader("shader/clippingPlane");
	if (!m_shader_cp) 
	{
		cerr << "Could not compile raycasting shader program." << endl;
		return;
	}
	m_shader_cp->bind_frag_data_location("out_color");

	std::map<std::string, int> cp_vbo;
	cp_vbo.insert(std::pair<std::string, int>("plane", VertexGeometry::AT_POSITIONS));
	std::map<std::string, int> cp_uniform;
	cp_uniform.insert(std::pair<std::string, int>("model", View::UT_MODELMATRIX));	
	cp_uniform.insert(std::pair<std::string, int>("view", View::UT_VIEWMATRIX));	
	cp_uniform.insert(std::pair<std::string, int>("perspective", View::UT_PROJECTIONMATRIX));
	cp_uniform.insert(std::pair<std::string, int>("screen", View::UT_TEXTURE0));
	
	Geometry* m_box_cp = Primitives::GetBox();
	m_box_cp->World = glm::mat4();
	m_box_cp->Local = glm::translate(0.f, 0.f, -2.f);
	m_box_cp->SetShader(0, m_shader_cp);
	m_box_cp->CreateDictShaderToVBO(0, cp_vbo);
	m_box_cp->CreateDictShaderUniforms(0, cp_uniform);
	m_box_cp->m_TextureObjs.insert(std::pair<unsigned int, FBO::FBOPtr>(0, fbo_back));

	Geometry* m_plane_cp = Primitives::GetPlane2();
	m_plane_cp->World = glm::mat4();
	m_plane_cp->Local = glm::translate(0.f, 0.f, -2.f);
	m_plane_cp->SetShader(0, m_shader_cp);
	m_plane_cp->CreateDictShaderToVBO(0, cp_vbo);
	m_plane_cp->CreateDictShaderUniforms(0, cp_uniform);
	m_plane_cp->m_TextureObjs.insert(std::pair<unsigned int, FBO::FBOPtr>(0, fbo_back));

	m_cp_back = new View();
	INode* cp_node = new INode(m_box_cp);
	cp_node->AttachChild(new INode(m_plane_cp));
	m_cp_back->SetRoot(cp_node);	
	m_cp_back->SetCamera(cam);
	m_cp_back->SetLight(m_light);
	m_cp_back->SetPass(0);
	m_cp_back->AddFBO(fbo_back);

	m_cp2_back = new View();
	m_cp2_back->SetRoot(new INode(m_plane_cp));
	m_cp2_back->SetCamera(cam);
	m_cp2_back->SetLight(m_light);
	m_cp2_back->SetPass(0);
	m_cp2_back->AddFBO(fbo_back);

	Geometry* m_box_cp2 = Primitives::GetBox();
	m_box_cp2->World = glm::mat4();
	m_box_cp2->Local = glm::translate(0.f, 0.f, -2.0f);
	m_box_cp2->SetShader(0, m_shader_cp);
	m_box_cp2->CreateDictShaderToVBO(0, cp_vbo);
	m_box_cp2->CreateDictShaderUniforms(0, cp_uniform);
	m_box_cp2->m_TextureObjs.insert(std::pair<unsigned int, FBO::FBOPtr>(0, fbo_front));
	
	m_cp_front = new View();
	m_cp_front->SetRoot(new INode(m_box_cp2));
	m_cp_front->SetCamera(cam);
	m_cp_front->SetLight(m_light);
	m_cp_front->SetPass(0);
	m_cp_front->AddFBO(fbo_front);


	Geometry* m_plane_cp2 = Primitives::GetPlane2();
	m_plane_cp2->World = glm::mat4();
	m_plane_cp2->Local = glm::translate(0.f, 0.f, -2.f);
	m_plane_cp2->SetShader(0, m_shader_cp);
	m_plane_cp2->CreateDictShaderToVBO(0, cp_vbo);
	m_plane_cp2->CreateDictShaderUniforms(0, cp_uniform);
	m_plane_cp2->m_TextureObjs.insert(std::pair<unsigned int, FBO::FBOPtr>(0, fbo_front));

	m_cp2_front = new View();
	m_cp2_front->SetRoot(new INode(m_plane_cp2));
	m_cp2_front->SetCamera(cam);
	m_cp2_front->SetLight(m_light);
	m_cp2_front->SetPass(0);
	m_cp2_front->AddFBO(fbo_front);

	std::cout << "Create B-Spline" << std::endl;

	int weights_size = 256;
	float* weights;
	weights = new float[weights_size*4];
	
	std::cout << "------------------------" << std::endl;
	std::cout << "Calculate B-Spline weight: " << std::endl;
	/*for(int index = 0; index < weights_size; index++)
	{
		float x = (float)(index)/(float)(weights_size);

		float x2 = x*x;
		float x3 = x2*x;

		float alpha = x - floorf(x);
		float w0 = 1.f/6.f*(-x3	+ (3.f*x2) - (3.f * alpha) + 1.f);
		float w1 = 1.f/6.f*(3.f*powf(alpha, 3.f)	- (6.f*pow(alpha,2.f)) + 4.f);
		float w2 = 1.f/6.f*((-3.f*powf(alpha, 3.f))	+ (3.f*pow(alpha,2.f)) + (3.f * alpha) + 1.f);
		float w3 = 1.f/6.f*powf(alpha, 3.f);

		float g0 = w0 + w1;
		float g1 = w2 + w3;
		float h0 = 1.f - (w1/(w0+w1))+x;
		float h1 = 1.f + (w3/(w2+w3))-x;

		weights[(index*4)+0] = h0;
		weights[(index*4)+1] = h1;
		weights[(index*4)+2] = g0; 
		weights[(index*4)+3] = g1;

		std::cout << "X: " << x << " ALPHA: " << alpha << std::endl;
		std::cout << "g0: " << g0 << " + g1 " << g1 << " = " << (g0 + g1) << std::endl;
	}*/

	float *data   = new float[4 * weights_size];
	float *ptr    = data;
	float  phi[4], x3, x2, x = 0.f;
	float  h      = 1.f / (float)(weights_size-1);
	float  c      = 1.f / 6.f;

	for(unsigned int i = 0; i < weights_size; ++i, x += h)
	{
		x3 = (x2 = x * x) * x;
		// cubic spline basis functions
		// non-negative partition of one; in [0, 2/3]
		phi[0] = c * (-x3 + 3*x2 - 3*x + 1.f);
		phi[1] = c * (3.f*x3 - 6.f*x2 + 4.f);
		phi[2] = c * (-3.f*x3 + 3.f*x2 + 3.f*x + 1.f);
		phi[3] = c * (x3);
		// offsets; first with changed sign; in [0, 1]
		*ptr++ = 1.0f + x - phi[1] / (phi[0] + phi[1]);
		*ptr++ = 1.0f - x + phi[3] / (phi[2] + phi[3]);
		// weight for left sample (lerp: right weight = 1-left weight)
		*ptr++ = phi[0] + phi[1];
		*ptr++ = phi[0] + phi[1];
	}

	std::cout << "------------------------" << std::endl;

	glGenTextures(1,&bspile_weights);
	glBindTexture(GL_TEXTURE_1D, bspile_weights);
	glTexParameteri(GL_TEXTURE_1D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_1D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_1D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexImage1D(GL_TEXTURE_1D,0,GL_RGBA32F,weights_size,0,GL_RGBA,GL_FLOAT,data);	
	glBindTexture(GL_TEXTURE_1D, 0.0);
	std::cout << "INIT HG & HG_DEV Texture" << std::endl;

	std::cout << "Create HG Texture: < " << weights_size << std::endl;
	m_hg = new CudaTexture<float4,2>(256,1,0);	
	
	std::cout << "INIT HG & HG_DEV Texture" << std::endl;
	initHGTexture(m_hg->GetCudaTexture());
	cutilSafeCall(cudaMemcpyToArray(m_hg->GetCudaTexture(), 0, 0, data, sizeof(float4)*weights_size, cudaMemcpyHostToDevice));
	BindHGCuda(m_hg->GetCudaTexture());

	delete[] data;
	
	data   = new float[4 * weights_size];
	ptr    = data;

	ptr = data; x = 0.f;
	for(int i = 0; i < weights_size; ++i, x += h)
	{
		x2 = x * x;
		// cubic spline basis functions derivatives
		// 0,1 negative; 2,3 positive; in [-2/3, 2/3]
		phi[0] = c * (-3.f*x2 + 6.f*x - 3.f);
		phi[1] = c * (9.f*x2 - 12.f*x);
		phi[2] = c * (-9.f*x2 + 6.f*x + 3.f);
		phi[3] = c * (3.f*x2);
		// offsets; first with changed sign; in [0, 1]
		*ptr++ = 1.f + x - phi[1] / (phi[0] + phi[1]);
		*ptr++ = 1.f - x + phi[3] / (phi[2] + phi[3]);
		// weight for right sample (central differences: left weight = -right weight)
		*ptr++ = phi[2] + phi[3];
		*ptr++ = phi[2] + phi[3];
	}

	glGenTextures(1,&bspile_dev_weights);
	glBindTexture(GL_TEXTURE_1D, bspile_dev_weights);
	glTexParameteri(GL_TEXTURE_1D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_1D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_1D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexImage1D(GL_TEXTURE_1D,0,GL_RGBA32F,weights_size,0,GL_RGBA,GL_FLOAT,data);
	glBindTexture(GL_TEXTURE_1D, 0.0);

	std::cout << "Create HG_DEV Texture" << std::endl;
	m_hg_dev = new CudaTexture<float4,2>(256,1,0);	

	initHGDevTexture(m_hg_dev->GetCudaTexture());
	cutilSafeCall(cudaMemcpyToArray(m_hg_dev->GetCudaTexture(), 0, 0, data, sizeof(float4)*weights_size, cudaMemcpyHostToDevice));
	BindHGDevCuda(m_hg_dev->GetCudaTexture());
	std::cout << "Initialized" << std::endl;
}

void RayCasting::AfterLoadingOfVolume(int width, int height, int depth)
{
	volumeWidth = width;
	volumeHeight = height;
	volumeDepth = depth;
	m_particle_cuda->res = make_cudaExtent(volumeWidth, volumeHeight, volumeDepth);
	std::cout << "Width: " << volumeWidth << " " << " << " << volumeHeight  << " " << volumeDepth << std::endl;
	m_particle_cuda->InitStartPositions();
}

void RayCasting::resize(int width, int height)
{
	this->screenWidth = width;
	this->screenHeight = height;
}

void RayCasting::render(float elapsedTime)
{
	m_bb_front->SetCamera(SceneManager::get().getCamera());
	m_bb_back->SetCamera(SceneManager::get().getCamera());
	m_volumeBricks_front->SetCamera(SceneManager::get().getCamera());
	m_volumeBricks_back->SetCamera(SceneManager::get().getCamera());
	m_raycasting->SetCamera(SceneManager::get().getCamera());
	m_particle->SetCamera(SceneManager::get().getCamera());
	m_cp_front->SetCamera(SceneManager::get().getCamera());
	m_cp_back->SetCamera(SceneManager::get().getCamera());
	m_cp2_front->SetCamera(SceneManager::get().getCamera());
	m_cp2_back->SetCamera(SceneManager::get().getCamera());

	/*
	m_contour->SetCamera(SceneManager::get().getCamera());
	m_exposure->SetCamera(SceneManager::get().getCamera());
	m_raycasting2D->SetCamera(SceneManager::get().getCamera());
	m_mip->SetCamera(SceneManager::get().getCamera());
	m_averaging->SetCamera(SceneManager::get().getCamera());
	*/
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if (this->textureLoaded) 
	{
		float duration = 0.f;
		if(RenderStates::ModusChanged == 2)	duration = 0.001;
		if(RenderStates::ModusChanged == 3)	duration = 0.1;
		//duration = 0.5;
		//std::cout << "Duration: " << duration<< std::endl;

		//Calculating Particle
		glm::vec3 new_pos = glm::vec3((float)(rand()%1000)/500.f-0.5f, (float)(rand()%1000)/500.f-0.5f, (float)(rand()%1000)/500.f-0.5f);
		glm::vec3 dir = SceneManager::get().getCamera()->GetCenterPositon()-SceneManager::get().getCamera()->GetPosition();
		dir = -dir;
		
		Particle* param = m_particle_cuda->Getparticle_parameter();
		param->sortVector = make_float3(dir.x, dir.y, dir.z);
		param->elapsed_time = duration;

		param->gravity = RenderStates::FluidGravity;
		param->globalDamping = 0.1f;
		param->predefinedDirection = make_float3(0.f,1.f,0.f);
		param->k_c = RenderStates::ParticleCurvatureValue;
		param->k_f = RenderStates::ParticleFlockingValue;
		param->k_p = RenderStates::ParticlePredefinedValue;
		param->k_d = RenderStates::ParticleDensityValue;

		/*std::cout << "k_c: " << RenderStates::ParticleCurvatureValue << std::endl;
		std::cout << "k_f: " << RenderStates::ParticleFlockingValue << std::endl;
		std::cout << "k_p: " << RenderStates::ParticlePredefinedValue << std::endl;
		std::cout << "k_d: " << RenderStates::ParticleDensityValue << std::endl;*/

		param->mass = RenderStates::FluidParticleMass;
		param->rho0 = RenderStates::FluidRestdensity;
		param->viscosity = RenderStates::FluidViscosity;
		param->siftness = RenderStates::FluidSiftness;

		param->SpringCoefficient = RenderStates::FluidSpringValue;
		param->DampingCoefficient = RenderStates::FluidDampingValue;
		param->ShearCoefficient = RenderStates::FluidShearValue;
		param->SpringCoefficientCT = RenderStates::FluidSpringCTValue;
		param->DampingCoefficientCT = RenderStates::FluidDampingCTValue;
		param->ShearCoefficientCT = RenderStates::FluidShearCTValue;
		SetParticle(param);

		//m_particle_cuda->runEmitter(new_pos);
		BindTransferfunctionCuda(m_texture->GetCudaTexture());
		BindVolumeCuda(m_volumen->GetCudaTexture());
		BindPCDCuda(m_volumen->GetCudaTexture());
			m_particle_cuda->Update(duration);
		UnbindPCDCuda();
		UnbindVolumeCuda();
		UnbindTransferfunctionCuda();

		m_particle_cuda->DepthSort();

		m_particle_geo->m_vbovert = m_particle_cuda->m_Pos->GetDevReadVBO();
		m_particle_geo->m_vbo_tex[0] = m_particle_cuda->m_Vel->GetDevReadVBO();
		m_particle_geo->m_vboind = m_particle_cuda->m_Index->GetDevReadVBO();
		//__________________
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);
		
		glDepthFunc(GL_LESS);
		glCullFace(GL_BACK);
		
		glActiveTexture(GL_TEXTURE0); 
		glBindTexture(GL_TEXTURE_1D, transferTextureHandle);
		
		//Render Frontboxes
		m_renderer->DrawView(m_volumeBricks_front);
		
		/*
		m_renderer->m_clear_depth_buffer = false;
		glDisable(GL_CULL_FACE);
		glEnable(GL_DEPTH_TEST);
		glDepthFunc(GL_LEQUAL);
		m_renderer->DrawView(m_cp_front);
		
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_CULL_FACE);
		m_renderer->DrawView(m_cp2_front);
		m_renderer->m_clear_depth_buffer = true;
		glEnable(GL_CULL_FACE);
		glEnable(GL_DEPTH_TEST);
		*/
		//Render Backboxes
		glClearDepth(0.f);
		glDepthFunc(GL_GREATER);
		glCullFace(GL_FRONT);
		
		m_renderer->DrawView(m_volumeBricks_back);
		/*
		m_renderer->m_clear_depth_buffer = false;
		glDisable(GL_CULL_FACE);
		glEnable(GL_DEPTH_TEST);
		glDepthFunc(GL_GEQUAL);
		m_renderer->DrawView(m_cp_back);
		
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_CULL_FACE);
		m_renderer->DrawView(m_cp2_back);
		m_renderer->m_clear_depth_buffer = true;
		glEnable(GL_CULL_FACE);
		glEnable(GL_DEPTH_TEST);*/

		glCullFace(GL_BACK);
		glClearDepth(1.f);
		glDepthFunc(GL_LESS);

		glBindTexture(GL_TEXTURE_1D, 0);

		/*/
		//Rendering Boxes
		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);
		m_renderer->DrawView(m_bb_front);
		glCullFace(GL_FRONT);
		m_renderer->DrawView(m_bb_back);
		m_renderer->GetErrors();//*/

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		//Raycasting
		glCullFace(GL_BACK);
		glDisable(GL_CULL_FACE);

		glEnable(GL_BLEND);
		//glDisable(GL_DEPTH_TEST);
		this->bindRayCastingTextures();
		if(m_type != 1)	
		{
			if(RenderStates::MIP == 1)				m_renderer->DrawView(m_mip);
			else if(RenderStates::Average == 1)		m_renderer->DrawView(m_averaging);
			else if(RenderStates::RenderMode == 1)	m_renderer->DrawView(m_contour);							
			else if(RenderStates::RenderMode == 2)	m_renderer->DrawView(m_exposure);
			else									m_renderer->DrawView(m_raycasting);
		}
		else	m_renderer->DrawView(m_raycasting2D);
		this->unbindRayCastingTextures();

		//Rendering Particle
		//glDisable(GL_DEPTH_TEST);
		glEnable(GL_DEPTH_TEST);
		glDisable(GL_CULL_FACE);
		m_renderer->m_clear_depth_buffer = false;
		//m_renderer->m_clear_depth_buffer = true;
	
		glActiveTexture(GL_TEXTURE0); 
		glBindTexture(GL_TEXTURE_1D, tfParticleHandle);
		glActiveTexture(GL_TEXTURE1); 
		glBindTexture(GL_TEXTURE_3D, volumeTextureHandle);
		
		m_renderer->DrawView(m_particle);
		glDisable(GL_BLEND);
		
		//glEnable(GL_CULL_FACE);
		glActiveTexture(GL_TEXTURE1); 
		glBindTexture(GL_TEXTURE_3D, 0);
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_1D, 0);
		m_renderer->m_clear_depth_buffer = true;
		//*/
		m_renderer->DrawView(m_hdr);	
	}

	m_renderer->GetErrors();
}

void RayCasting::setVolumeTextureHandle(GLuint tex)
{
	this->volumeTextureHandle = tex;
}

void RayCasting::setPCDTextureHandle(GLuint tex)
{
	pcdTextureHandle = tex;
}

void RayCasting::setTransferTextureHandle(GLuint tex)
{
	this->transferTextureHandle = tex;
}

void RayCasting::setTransferTexture2DHandle(GLuint tex)
{
	this->transferTextureHandle2D = tex;
}

void RayCasting::setTransferTextureParticleHandle(GLuint tex)
{
	this->tfParticleHandle = tex;
}

void RayCasting::setTransferFunction(bool active)
{
	this->transferFunction = active;
}

void RayCasting::setVolumeSize(float width, float height, float depth)
{
	this->volumeWidth = width;
	this->volumeHeight = height;
	this->volumeDepth = depth;
}

void RayCasting::setTextureLoaded(bool loaded)
{
	this->textureLoaded = loaded;
}

void RayCasting::setLight(bool active)
{
	this->light = active;
}

void RayCasting::setPhong(bool active)
{
	this->phong = active;
}

void RayCasting::setMip(bool active)
{
	this->mip = active;
}
void RayCasting::setAverage(bool active)
{
	this->average = active;
}
void RayCasting::setCompositing(bool active)
{
	this->average = active;
}
void RayCasting::setAdaptive(bool active)
{
	this->adaptive = active;
}
void RayCasting::setStepSize(double step)
{
	this->step = step;
}

void RayCasting::setXClipping(float x)
{
	this->xClipping = x;
	//this->calculateCubeVertices();
	//this->calculateColors();
}

void RayCasting::setYClipping(float y)
{
	this->yClipping = y;
	//this->calculateCubeVertices();
	//this->calculateColors();
}

void RayCasting::setZClipping(float z)
{
	this->zClipping = z;
	//this->calculateCubeVertices();
	//this->calculateColors();
}

void RayCasting::bindRayCastingTextures()
{
	glActiveTexture(GL_TEXTURE2); 
	glBindTexture(GL_TEXTURE_3D, this->volumeTextureHandle);

	// bind tranfer texture to texture unit 3 if exists
	if (transferFunction) 
	{
		glActiveTexture(GL_TEXTURE3); 
		if(m_type != 1)		glBindTexture(GL_TEXTURE_1D, this->transferTextureHandle);
		else				glBindTexture(GL_TEXTURE_2D, this->transferTextureHandle2D);
	} 

	glActiveTexture(GL_TEXTURE4); 
	glBindTexture(GL_TEXTURE_1D, this->bspile_weights);

	glActiveTexture(GL_TEXTURE5); 
	glBindTexture(GL_TEXTURE_1D, this->bspile_dev_weights);

	glActiveTexture(GL_TEXTURE6); 
	glBindTexture(GL_TEXTURE_3D, this->pcdTextureHandle);
}

void RayCasting::unbindRayCastingTextures()
{
	glActiveTexture(GL_TEXTURE2);
	glBindTexture(GL_TEXTURE_3D, 0);

	glActiveTexture(GL_TEXTURE3);
	if(m_type != 1)		glBindTexture(GL_TEXTURE_1D, 0);
	else				glBindTexture(GL_TEXTURE_2D, 0);

	glActiveTexture(GL_TEXTURE4);
	glBindTexture(GL_TEXTURE_1D, 0);
	
	glActiveTexture(GL_TEXTURE5);
	glBindTexture(GL_TEXTURE_1D, 0);

	glActiveTexture(GL_TEXTURE6);
	glBindTexture(GL_TEXTURE_3D, 0);

	glActiveTexture(GL_TEXTURE0);
}

void RayCasting::SetIsoValue(float val)
{
	BindTransferfunctionCuda(m_texture->GetCudaTexture());
	BindVolumeCuda(m_volumen->GetCudaTexture());
	BindPCDCuda(m_volumen->GetCudaTexture());
			m_particle_cuda->SetIsoValue(val);
	UnbindPCDCuda();
	UnbindVolumeCuda();
	UnbindTransferfunctionCuda();
}











//COPY VOM CUDA SDK UM ZU TESTEN, OB DIE VOLUMENDATEN BENZT WERDEN KNNEN!!!
void RayCasting::initPixelBuffer()
{
	std::cout << "initPixelBuffer start" << std::endl;
	m_renderer->GetErrors();

    if (pbo) {
        // unregister this buffer object from CUDA C
        cutilSafeCall(cudaGraphicsUnregisterResource(cuda_pbo_resource));

        // delete old buffer
        glDeleteBuffers(1, &pbo);
        glDeleteTextures(1, &tex);
    }

    // create pixel buffer object for display
    glGenBuffers(1, &pbo);
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER, pbo);
	glBufferData(GL_PIXEL_UNPACK_BUFFER, width*height*sizeof(GLubyte)*4, 0, GL_STREAM_DRAW);
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);

    // register this buffer object with CUDA
	cutilSafeCall(cudaGraphicsGLRegisterBuffer(&cuda_pbo_resource, pbo, cudaGraphicsMapFlagsWriteDiscard));	

    // create texture for display
    glGenTextures(1, &tex);
    glBindTexture(GL_TEXTURE_2D, tex);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glBindTexture(GL_TEXTURE_2D, 0);

	std::cout << "initPixelBuffer end" << std::endl;
	m_renderer->GetErrors();
}

int iDivUp(int a, int b){
    return (a % b != 0) ? (a / b + 1) : (a / b);
}

// render image using CUDA
void RayCasting::renderCuda()
{
	std::cout << "renderCuda start" << std::endl;
	m_renderer->GetErrors();

	dim3 blockSize(16, 16);
	dim3 gridSize;
	gridSize = dim3(iDivUp(width, blockSize.x), iDivUp(height, blockSize.y));

	copyInvViewMatrix(invViewMatrix, sizeof(float4)*3);

    // map PBO to get CUDA device pointer
    uint *d_output;
	// map PBO to get CUDA device pointer
	cutilSafeCall(cudaGraphicsMapResources(1, &cuda_pbo_resource, 0));
    size_t num_bytes; 
    cutilSafeCall(cudaGraphicsResourceGetMappedPointer((void **)&d_output, &num_bytes, cuda_pbo_resource));
    //printf("CUDA mapped PBO: May access %ld bytes\n", num_bytes);

    // clear image
    cutilSafeCall(cudaMemset(d_output, 0, width*height*4));

    // call CUDA kernel, writing results to PBO
    render_kernel(gridSize, blockSize, d_output, width, height, 0.05f, 1.f, 0.f, 1.f);

    cutilCheckMsg("kernel failed");

    cutilSafeCall(cudaGraphicsUnmapResources(1, &cuda_pbo_resource, 0));

	std::cout << "renderCuda end" << std::endl;
	m_renderer->GetErrors();
}

// display results using OpenGL (called by GLUT)
void RayCasting::display()
{
	std::cout << "display start" << std::endl;
	m_renderer->GetErrors();

	Camera* cam = SceneManager::get().getCamera();

	glm::mat4 tmp_m;
	tmp_m = tmp_m * glm::rotate(glm::mat4(), -0.f, glm::vec3(1.0f, 0.0f, 0.0f));
	tmp_m = tmp_m * glm::rotate(glm::mat4(), -0.f, glm::vec3(0.0f, 1.0f, 0.0f));
	tmp_m = tmp_m * glm::translate(glm::mat4(), glm::vec3(-0.f, -0.f, 4.f));

    invViewMatrix[0] = tmp_m[0][0]; invViewMatrix[1] = tmp_m[1][0]; invViewMatrix[2] =  tmp_m[2][0]; invViewMatrix[3] =  tmp_m[3][0];
    invViewMatrix[4] = tmp_m[0][1]; invViewMatrix[5] = tmp_m[1][1]; invViewMatrix[6] =  tmp_m[2][1]; invViewMatrix[7] =  tmp_m[3][1];
    invViewMatrix[8] = tmp_m[0][2]; invViewMatrix[9] = tmp_m[1][2]; invViewMatrix[10] = tmp_m[2][2]; invViewMatrix[11] = tmp_m[3][2];

	BindTransferfunctionCuda(m_texture->GetCudaTexture());
    renderCuda();
	//UnbindTransferfunctionCuda();
	
	// draw using texture
    // copy from pbo to texture
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, pbo);
    glBindTexture(GL_TEXTURE_2D, tex);
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, 0);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);

    // draw textured quad
	std::cout << "TEXTURE: " << tex << std::endl;

    glBindTexture(GL_TEXTURE_2D, 0);

	std::cout << "display end" << std::endl;
	m_renderer->GetErrors();
}