#pragma once
#include "MyRenderer.h"
#include "GL\glew.h"

class RendererOpenGL : public MyRenderer
{
public:
	/**
	 * Standardconstructor
	 **/
	RendererOpenGL(void);
	/**
	 * Standarddeconstructor
	 **/
	~RendererOpenGL(void);

	/**
	 * This method clears the accumulation buffer
	 **/
	inline virtual void ClearAccumBuffer();
	/**
	 * This method clears the color buffer
	 **/
	inline virtual void ClearColorBuffer();
	/**
	 * This method clears the depth buffer
	 **/
	inline virtual void ClearDepthBuffer();
	/**
	 * This method clears the stencil buffer
	 **/
	inline virtual void ClearStencilBuffer();
	/**
	 * This method sets the clear color
	 * @param color clear color
	 **/
	inline virtual void SetClearColor(glm::vec4 color);
	/**
	 * This method swaps the buffer
	 **/
	inline virtual void SwapBuffer();

	/**
	 * This method returns the error
	 **/
	inline virtual void GetErrors();

	/**
	 * This method sets the culling mode to FRONT
	 **/
	inline virtual void SetFrontCulling();
	/**
	 * This method sets the culling mode to BACK
	 **/
	inline virtual void SetBackCulling();
	/**
	 * This method enables Backfaceculling
	 **/
	inline virtual void EnableBackFaceCulling();
	/**
	 * This method disables Backfaceculling
	 **/
	inline virtual void DisableBackFaceCulling();

	/**
	 * This method enables a texture
	 * @param texture Texture to enable
	 * @param i position in texturestack
	 **/
	inline virtual void EnableTexture(Texture::TexturePtr texture, int i);
	/**
	 * This method disable a texture
	 * @param texture Texture to enable
	 * @param i position in texturestack
	 **/
	inline virtual void DisableTexture(Texture::TexturePtr texture, int i);
	/**
	 * This method loads a texture
	 * @param texture Texture to load
	 **/
	virtual void LoadTexture(Texture::TexturePtr texture);

protected:
	/**
	 * This Method draw the geometry object
	 */
	void DrawVertexGeometry();
	/**
	 * This Method draws the object
	 */
	void DrawGeometry();

private:
	void DrawGeometrySpecificUniforms();
};

