// local includes
#include "Shader.h"


Shader::Shader() : vertex_shader_handle(0), geometry_shader_handle(0), fragment_shader_handle(0)
{

}

Shader::~Shader()
{
	glDeleteProgram(shaderProgram);
}

void Shader::loadCompileAndLinkShader(char* vertexShaderLocation, char* geometryShaderLoacation, char* fragmentShaderLoacation)
{
	loadAndCompile(vertexShaderLocation, geometryShaderLoacation, fragmentShaderLoacation);
	link();
	glDeleteShader(vertex_shader_handle);
	glDeleteShader(fragment_shader_handle);
}

void Shader::loadCompileAndLinkShader(char* vertexShaderLocation, char* fragmentShaderLoacation)
{
	loadAndCompile(vertexShaderLocation, fragmentShaderLoacation);
	//glBindFragDataLocation(shaderProgram, 0, "out_color" );
	link();

	glDeleteShader(vertex_shader_handle);
	glDeleteShader(geometry_shader_handle);
	glDeleteShader(fragment_shader_handle);
}


GLint Shader::getShaderProgram() const
{
	return shaderProgram;
}

void Shader::loadAndCompile(char* vertexShaderLocation, char* fragmentShaderLoacation)
{
	// -- vertex shader --
	std::ifstream vs_file(vertexShaderLocation, std::ifstream::in);
	std::string vertex_shader_source = std::string(std::istreambuf_iterator<char>(vs_file), std::istreambuf_iterator<char>());

	// create handle
	vertex_shader_handle = glCreateShader(GL_VERTEX_SHADER);

	// convert to const c-string and attach to shader-handle
	const GLchar *vs_c_str = vertex_shader_source.c_str();
	glShaderSource(vertex_shader_handle, 1, &vs_c_str, NULL);

	// compile shader
	glCompileShader(vertex_shader_handle);

	// do error-checking
	GLint succeeded;
	glGetShaderiv(vertex_shader_handle, GL_COMPILE_STATUS, &succeeded);

	if (!succeeded || !glIsShader(vertex_shader_handle) )
	{
		// there was an error

		// get log-length
		int log_length = 0;
		glGetShaderiv(vertex_shader_handle, GL_INFO_LOG_LENGTH, &log_length);

		// get info-log
		std::string info_log(log_length, ' ');
		glGetShaderInfoLog(vertex_shader_handle, log_length, NULL, &info_log[0]);

		// print info-log
		std::cout << "- Vertex-shader compile error:\n\n" << info_log.c_str() << std::endl;
	} else
	{
		std::cout << "- Vertex-shader loaded and compiled successfully." << std::endl;
	}

	// -- fragment shader --
	std::ifstream fs_file(fragmentShaderLoacation, std::ifstream::in);
	std::string fragment_shader_source = std::string(std::istreambuf_iterator<char>(fs_file), std::istreambuf_iterator<char>());

	// create handle
	fragment_shader_handle = glCreateShader(GL_FRAGMENT_SHADER);

	// convert to const c-string and attach to shader-handle
	const GLchar *fs_c_str = fragment_shader_source.c_str();
	glShaderSource(fragment_shader_handle, 1, &fs_c_str, NULL);

	// compile shader
	glCompileShader(fragment_shader_handle);

	// do error-checking
	glGetShaderiv(fragment_shader_handle, GL_COMPILE_STATUS, &succeeded);

	if (!succeeded || !glIsShader(fragment_shader_handle) )
	{
		// there was an error

		// get log-length
		int log_length = 0;
		glGetShaderiv(fragment_shader_handle, GL_INFO_LOG_LENGTH, &log_length);

		// get info-log
		std::string info_log(log_length, ' ');
		glGetShaderInfoLog(fragment_shader_handle, log_length, NULL, &info_log[0]);

		// print info-log
		std::cout << "- Fragment-shader compile error:\n\n" << info_log.c_str() << std::endl;
	} else
	{
		std::cout << "- Fragment-shader loaded and compiled successfully." << std::endl;
	}
}

void Shader::loadAndCompile(char* vertexShaderLocation, char* geometryShaderLocation, char* fragmentShaderLoacation)
{
	// -- vertex shader --
	std::ifstream vs_file(vertexShaderLocation, std::ifstream::in);
	std::string vertex_shader_source = std::string(std::istreambuf_iterator<char>(vs_file), std::istreambuf_iterator<char>());

	// create handle
	vertex_shader_handle = glCreateShader(GL_VERTEX_SHADER);

	// convert to const c-string and attach to shader-handle
	const GLchar *vs_c_str = vertex_shader_source.c_str();
	glShaderSource(vertex_shader_handle, 1, &vs_c_str, NULL);

	// compile shader
	glCompileShader(vertex_shader_handle);

	// do error-checking
	GLint succeeded;
	glGetShaderiv(vertex_shader_handle, GL_COMPILE_STATUS, &succeeded);

	if (!succeeded || !glIsShader(vertex_shader_handle) )
	{
		// there was an error

		// get log-length
		int log_length = 0;
		glGetShaderiv(vertex_shader_handle, GL_INFO_LOG_LENGTH, &log_length);

		// get info-log
		std::string info_log(log_length, ' ');
		glGetShaderInfoLog(vertex_shader_handle, log_length, NULL, &info_log[0]);

		// print info-log
		std::cout << "- Vertex-shader compile error:\n\n" << info_log.c_str() << std::endl;
	} else
	{
		std::cout << "- Vertex-shader loaded and compiled successfully." << std::endl;
	}

	//______________________

	// -- geometry shader --
	std::ifstream gs_file(geometryShaderLocation, std::ifstream::in);
	std::string geometry_shader_source = std::string(std::istreambuf_iterator<char>(gs_file), std::istreambuf_iterator<char>());

	// create handle
	geometry_shader_handle = glCreateShader(GL_GEOMETRY_SHADER);

	// convert to const c-string and attach to shader-handle
	const GLchar *gs_c_str = geometry_shader_source.c_str();
	glShaderSource(geometry_shader_handle, 1, &gs_c_str, NULL);

	// compile shader
	glCompileShader(geometry_shader_handle);

	// do error-checking
	glGetShaderiv(geometry_shader_handle, GL_COMPILE_STATUS, &succeeded);

	if (!succeeded || !glIsShader(geometry_shader_handle) )
	{
		// there was an error

		// get log-length
		int log_length = 0;
		glGetShaderiv(geometry_shader_handle, GL_INFO_LOG_LENGTH, &log_length);

		// get info-log
		std::string info_log(log_length, ' ');
		glGetShaderInfoLog(geometry_shader_handle, log_length, NULL, &info_log[0]);

		// print info-log
		std::cout << "- Geometry-shader compile error:\n\n" << info_log.c_str() << std::endl;
	} else
	{
		std::cout << "- Geometry-shader loaded and compiled successfully." << std::endl;
	}

	//______________________

	// -- fragment shader --
	std::ifstream fs_file(fragmentShaderLoacation, std::ifstream::in);
	std::string fragment_shader_source = std::string(std::istreambuf_iterator<char>(fs_file), std::istreambuf_iterator<char>());

	// create handle
	fragment_shader_handle = glCreateShader(GL_FRAGMENT_SHADER);

	// convert to const c-string and attach to shader-handle
	const GLchar *fs_c_str = fragment_shader_source.c_str();
	glShaderSource(fragment_shader_handle, 1, &fs_c_str, NULL);

	// compile shader
	glCompileShader(fragment_shader_handle);

	// do error-checking
	glGetShaderiv(fragment_shader_handle, GL_COMPILE_STATUS, &succeeded);

	if (!succeeded || !glIsShader(fragment_shader_handle) )
	{
		// there was an error

		// get log-length
		int log_length = 0;
		glGetShaderiv(fragment_shader_handle, GL_INFO_LOG_LENGTH, &log_length);

		// get info-log
		std::string info_log(log_length, ' ');
		glGetShaderInfoLog(fragment_shader_handle, log_length, NULL, &info_log[0]);

		// print info-log
		std::cout << "- Fragment-shader compile error:\n\n" << info_log.c_str() << std::endl;
	} else
	{
		std::cout << "- Fragment-shader loaded and compiled successfully." << std::endl;
	}
}


void Shader::link()
{
	// -- link shader-program --
	shaderProgram = glCreateProgram();

    // attach shaders
	glAttachShader(shaderProgram, vertex_shader_handle);
	if(geometry_shader_handle != 0) glAttachShader(shaderProgram, geometry_shader_handle);
	glAttachShader(shaderProgram, fragment_shader_handle);

	// before linking, define output of fragment-shader
	glBindFragDataLocation(shaderProgram, 0, "fragColor");

    // link
    glLinkProgram(shaderProgram);

    // check for errors
	GLint succeeded;
    glGetProgramiv(shaderProgram, GL_LINK_STATUS, &succeeded);

    if (!succeeded || !glIsProgram(shaderProgram))
	{
		// there was an error

		// get log-length
		int log_length = 0;
		glGetProgramiv(shaderProgram, GL_INFO_LOG_LENGTH, &log_length);

		// get info-log
		std::string info_log(log_length, ' ');
		glGetProgramInfoLog(shaderProgram, log_length, NULL, &info_log[0]);

		// print info-log
		std::cout << "- Shader-program link error:\n\n" << info_log.c_str() << std::endl;
	} else
	{
		std::cout << "- Shader-program linked successfully.\n" << std::endl;
	}
}