#pragma once
// system includes
#include <iostream>
#include <fstream>

#include "GL/glew.h"
#include "GLM/glm.hpp"

class Shader {

public:
	/** The constructor. */
	Shader();
	/** The destructor. */
	~Shader();

	/** Loads, compiles and links the shader to the given parameters.
	 * @param  vertexShaderLocation The location of the vertex shader.
	 * @param  fragmentShaderLoacation The location of the fragment shader.
	 */
	void loadCompileAndLinkShader(char* vertexShaderLocation, char* geometryShaderLocation, char* fragmentShaderLoacation);

	/** Loads, compiles and links the shader to the given parameters.
	 * @param  vertexShaderLocation The location of the vertex shader.
	 * @param  fragmentShaderLoacation The location of the fragment shader.
	 */
	void loadCompileAndLinkShader(char* vertexShaderLocation, char* fragmentShaderLoacation);

	// Define the name of the variable inside the shader which represents the final color for each fragment.
	void bind_frag_data_location(const std::string &name)
	{
		if(shaderProgram > 0)
		{
			glBindFragDataLocation(shaderProgram, 0, name.c_str() );
			link();
		}
	}

	// Define the name of the variable inside the shader which represents the final color for each fragment.
	void bind_frag_data_location(const std::string &name, const std::string &name2)
	{
		if(shaderProgram > 0)
		{
			glBindFragDataLocation(shaderProgram, 0, name.c_str() );
			glBindFragDataLocation(shaderProgram, 1, name2.c_str() );
			link();
		}
	}

	/** Returns a handle for the shader program. */
	GLint getShaderProgram() const;

private:
	/** The handle for the shader program. */
	GLint shaderProgram;
	/** The handle for the vertex shader. */
	GLuint vertex_shader_handle;
	/** The handle for the frament shader. */
	GLuint fragment_shader_handle;
	/** The handle for the geometry shader. */
	GLuint geometry_shader_handle;

	/** Loads the shader to the given parameters.
	 * @param  vertexShaderLocation The location of the vertex shader.
	 * @param  fragmentShaderLoacation The location of the fragment shader.
	 */
	void loadAndCompile(char* vertexShaderLocation, char* fragmentShaderLoacation);

	/** Loads, compiles and links the shader to the given parameters.
	 * @param  vertexShaderLocation The location of the vertex shader.
	 * @param  fragmentShaderLoacation The location of the fragment shader.
	 */
	void loadAndCompile(char* vertexShaderLocation, char* geometryShaderLocation, char* fragmentShaderLoacation);


	/** Links the shader program.*/
	void link();
};