#include "Slicing.h"
#include "Coordinaten.h"
#include "SceneManager.h"


Slicing::Slicing() : textureLoaded(false), transferFunction(false)
{
	this->quadVerticesX = new glm::vec4[4];
	this->quadVerticesY = new glm::vec4[4];
	this->quadVerticesZ = new glm::vec4[4];

	this->quadTexCoordsX = new glm::vec3[4];
	this->quadTexCoordsY = new glm::vec3[4];
	this->quadTexCoordsZ = new glm::vec3[4];
}

Slicing::~Slicing()
{
	delete shader;

	delete[] quadVerticesX;
	delete[] quadVerticesY;
	delete[] quadVerticesZ;

	delete[] quadTexCoordsX;
	delete[] quadTexCoordsY;
	delete[] quadTexCoordsZ;
}

void Slicing::init()
{
	this->shader = new Shader();
	this->shader->loadCompileAndLinkShader("shader/slicingShader.vert", "shader/slicingShader.frag");

	this->setXSlice(0.5f);
	this->setYSlice(0.5f);
	this->setZSlice(0.5f);
}

void Slicing::resize(int width, int height)
{
	this->screenWidth = width;
	this->screenHeight = height;

	this->calculateTexturePosition();
}

void Slicing::render()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//if (this->textureLoaded) {

		// Bind shader
		glUseProgram(this->shader->getShaderProgram());
		glDisable(GL_CULL_FACE);

		glActiveTexture(GL_TEXTURE0); 
		glBindTexture(GL_TEXTURE_3D, this->slicingTextureHandle);
		GLuint texLocation = glGetUniformLocation(this->shader->getShaderProgram(), "colorTexture");
		glUniform1i(texLocation, 0);

		// transferFunction
		GLuint transferLocation;

		transferLocation = glGetUniformLocation(this->shader->getShaderProgram(), "transfer");
		if (transferFunction) 
		{
			glUniform1i(transferLocation, 1); //true

			glActiveTexture(GL_TEXTURE1); 
			glBindTexture(GL_TEXTURE_1D, this->transferTextureHandle);
			GLuint transferTextureLocation = glGetUniformLocation(this->shader->getShaderProgram(), "transferTexture");
			glUniform1i(transferTextureLocation, 1);
		} 
		else	glUniform1i(transferLocation, 0); //false
		

		// set matrix-uniforms
		GLint persp_uniform = glGetUniformLocation(this->shader->getShaderProgram(), "perspective");
		glUniformMatrix4fv(persp_uniform, 1, GL_FALSE, glm::gtx::type_ptr::value_ptr(glm::ortho(0.f, 1.f, 0.f, 1.f, -10.f, 10.f)));

		GLint view_uniform = glGetUniformLocation(this->shader->getShaderProgram(), "view");
		glUniformMatrix4fv(view_uniform, 1, GL_FALSE, glm::gtx::type_ptr::value_ptr(Camera().LookAt(glm::vec3(0.f,0.f,1.f), glm::vec3(0.f,0.f,0.f), glm::vec3(0.f, 1.f, 0.f))));

		// set attributes
		GLint vertex_attrib = glGetAttribLocation(this->shader->getShaderProgram(), "vertex");
		GLint texCoord_attrib  = glGetAttribLocation(this->shader->getShaderProgram(), "texCoord");

		glEnableVertexAttribArray(vertex_attrib);
		glEnableVertexAttribArray(texCoord_attrib);

		glVertexAttribPointer(vertex_attrib, 4, GL_FLOAT, GL_FALSE, 0, quadVerticesX);
		glVertexAttribPointer(texCoord_attrib,  3, GL_FLOAT, GL_FALSE, 0, quadTexCoordsX);
		glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, quadIndices);

		// --- draw y slice -- 
		glVertexAttribPointer(vertex_attrib, 4, GL_FLOAT, GL_FALSE, 0, quadVerticesY);
		glVertexAttribPointer(texCoord_attrib,  3, GL_FLOAT, GL_FALSE, 0, quadTexCoordsY);
		glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, quadIndices);

		// --- draw z slice -- 
		glVertexAttribPointer(vertex_attrib, 4, GL_FLOAT, GL_FALSE, 0, quadVerticesZ);
		glVertexAttribPointer(texCoord_attrib,  3, GL_FLOAT, GL_FALSE, 0, quadTexCoordsZ);
		glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, quadIndices);

		// disable attribute-pointers
		glDisableVertexAttribArray(vertex_attrib);
		glDisableVertexAttribArray(texCoord_attrib);

		// enable depth test
		glEnable(GL_DEPTH_TEST);

		// unbind shader-program
		glUseProgram(0);

		glActiveTexture(GL_TEXTURE0); 
		glBindTexture(GL_TEXTURE_3D, 0);
		glActiveTexture(GL_TEXTURE1); 
		glBindTexture(GL_TEXTURE_1D, 0);
		glEnable(GL_CULL_FACE);
//	}
}

void Slicing::setSlicingTextureHandle(GLuint tex)
{
	this->slicingTextureHandle = tex;
}

void Slicing::setTransferTextureHandle(GLuint tex)
{
	this->transferTextureHandle = tex;
}

void Slicing::setTransferFunction(bool active)
{
	this->transferFunction = active;
}

void Slicing::setVolumeSize(float width, float height,float depth)
{
	this->volumeWidth = width;
	this->volumeHeight = height;
	this->volumeDepth = depth;
}

void Slicing::setTextureLoaded(bool loaded)
{
	this->textureLoaded = loaded;
	this->calculateTexturePosition();
}

void Slicing::setXSlice(float x)
{
	this->quadTexCoordsX[0] = glm::vec3(x,0,1);
	this->quadTexCoordsX[1] = glm::vec3(x,1,1);
	this->quadTexCoordsX[2] = glm::vec3(x,1,0);
	this->quadTexCoordsX[3] = glm::vec3(x,0,0);
}

void Slicing::setYSlice(float y)
{
	this->quadTexCoordsY[0] = glm::vec3(0,y,1);
	this->quadTexCoordsY[1] = glm::vec3(1,y,1);
	this->quadTexCoordsY[2] = glm::vec3(1,y,0);
	this->quadTexCoordsY[3] = glm::vec3(0,y,0);
}

void Slicing::setZSlice(float z)
{
	this->quadTexCoordsZ[0] = glm::vec3(0,0,z);
	this->quadTexCoordsZ[1] = glm::vec3(1,0,z);
	this->quadTexCoordsZ[2] = glm::vec3(1,1,z);
	this->quadTexCoordsZ[3] = glm::vec3(0,1,z);
}

void Slicing::calculateTexturePosition()
{
	//calculate scalefactor
	float scaleHeight = (0.5f * this->screenHeight) / (this->volumeHeight);
	float scaleWidth = (0.5f * this->screenWidth) / (this->volumeWidth);
	float scale = min(scaleHeight, scaleWidth);

	float sizeX = this->volumeWidth * scale;
    float sizeY = this->volumeHeight * scale;
	float sizeZ = this->volumeDepth * scale;

	//calculate and set texture position
	float posX = 0.0;
	float posY = 0.5f;

	this->setQuadVerticesX(0.0f, 0.5f, 0.5f, 0.5f);
	this->setQuadVerticesY(0.5f, 0.5f, 0.5f, 0.5f);
	this->setQuadVerticesZ(0.0f, 0.0f, 0.5f, 0.5f);
}

void Slicing::setQuadVerticesX(float PosX, float PosY, float sizeX, float sizeY)
{
	std::cout << "PosX " << PosX << ", PosY " << PosY<< ", sizeX " << sizeX << ", sizeY" << sizeY << std::endl;
    this->quadVerticesX[0] = glm::vec4(PosX,           PosY + sizeY,   0.0f, 1.0f);
    this->quadVerticesX[1] = glm::vec4(PosX + sizeX,   PosY + sizeY,   0.0f, 1.0f);
    this->quadVerticesX[2] = glm::vec4(PosX + sizeX,   PosY,          0.0f, 1.0f);
    this->quadVerticesX[3] = glm::vec4(PosX,           PosY,           0.0f, 1.0f);
}

void Slicing::setQuadVerticesY(float PosX, float PosY, float sizeX, float sizeY)
{
    this->quadVerticesY[0] = glm::vec4(PosX,           PosY + sizeY,   0.0f, 1.0f);
    this->quadVerticesY[1] = glm::vec4(PosX + sizeX,   PosY + sizeY,   0.0f, 1.0f);
    this->quadVerticesY[2] = glm::vec4(PosX + sizeX,   PosY ,          0.0f, 1.0f);
    this->quadVerticesY[3] = glm::vec4(PosX,           PosY,           0.0f, 1.0f);
}

void Slicing::setQuadVerticesZ(float PosX, float PosY, float sizeX, float sizeY)
{
    this->quadVerticesZ[0] = glm::vec4(PosX,           PosY + sizeY,   0.0f, 1.0f);
    this->quadVerticesZ[1] = glm::vec4(PosX + sizeX,   PosY + sizeY,   0.0f, 1.0f);
    this->quadVerticesZ[2] = glm::vec4(PosX + sizeX,   PosY ,          0.0f, 1.0f);
    this->quadVerticesZ[3] = glm::vec4(PosX,           PosY,           0.0f, 1.0f);
}