#pragma once

#include "Shader.h"
#include "GLM/gtx.hpp"
#include "GLM/gtc.hpp"
#include "GLM/gtx/type_ptr.hpp"


class Slicing

{

public:
	/** The constructor. */
	Slicing();
	/** The destructor. */
	~Slicing();

	/** 
	 * Initialize the shaders an slicing parameter.
	 */
	void init();
	/** 
	 * Sets the new window size and calculates the vertices from the quad. 
	 * @param width The new width of the window, measured in pixel.
	 * @param height The new height of the window, measured in pixel.
	 */
	void resize(int width, int height);
	/** 
	 * Renders the scene. 
	 */
	void render();

	/**
	 * Sets the texturehandle for the slicing texture.
	 * @param tex The texture ID.
	 */
	void setSlicingTextureHandle(GLuint tex);
	/**
	 * Sets the texturehandle for the tranfer texture.
	 * @param tex The texture ID.
	 */
	void setTransferTextureHandle(GLuint tex);
	/** 
	 * Activates or deactivates the transfer function according to the given parameter.
	 * @param active True to activates the transfer function otherwise false.
	 */
	void setTransferFunction(bool active);
	/**
	 * Sets the width, height and depth for the current volume.
	 * @param width The widht of the volume.
	 * @param height The height of the volume.
	 * @param depth The depth of the volume.
	 */
	void setVolumeSize(float width, float height,float depth);
	/**
	* Sets the value of the textureLoaded property.
	* @param loaded True if the texture of the volume is loaded, false otherwise.
	*/
	void setTextureLoaded(bool loaded);

	/** 
	 * Sets the x-coordinate for the slice.
	 * The value must be in the intervall [0,1].
	 * @param x The x-coordinate of the slice.
	 */
	void setXSlice(float x);
	/** 
	 * Sets the y-coordinate for the slice.
	 * The value must be in the intervall [0,1].
	 * @param y The y-coordinate of the slice.
	 */
	void setYSlice(float y);
	/** 
	 * Sets the z-coordinate for the slice.
	 * The value must be in the intervall [0,1].
	 * @param z The z-coordinate of the slice.
	 */
	void setZSlice(float z);

protected:

private:
	/** The shader for slicing. */
	Shader* shader;
	/** The width of the screen. */
	int screenWidth;
	/** The height of the screen. */
	int screenHeight;
	/** The viewing matrix. */
	glm::mat4 view;

	/** True if the volume texture is loaded.*/
	bool textureLoaded;
	/** The width of the current volume. */
	float volumeWidth;
	/** The height of the current volume. */
	float volumeHeight;
	/** The depth of the current volume. */
	float volumeDepth;

	/** The vertices for the x-slices. */
	glm::vec4* quadVerticesX;
	/** The vertices for the y-slices. */
	glm::vec4* quadVerticesY;
	/** The vertices for the z-slices. */
	glm::vec4* quadVerticesZ;

	/** The texture coordinates for the x-slices. */
	glm::vec3* quadTexCoordsX;
	/** The texture coordinates for the y-slices. */
	glm::vec3* quadTexCoordsY;
	/** The texture coordinates for the z-slices. */
	glm::vec3* quadTexCoordsZ;

	/** The texturehandle for the volume texture. */
	GLuint slicingTextureHandle;
	/** The texturehandle for the transfer function texture. */
	GLuint transferTextureHandle;
	/** True if the transfer function is active.*/
	bool transferFunction;

	/** Calculates the texture position. */
	void calculateTexturePosition();
	/** 
	 * Calculates the vertices for the x-slices according to the given parameters.
	 * @param PosX The x-position of the left upper corner in screen coordinates.
	 * @param PosY The y-position of the left upper corner in screen coordinates.
	 * @param sizeX The width, measured in pixel.
	 * @param sizeY The height, measured in pixel.
	 */
	void setQuadVerticesX(float PosX, float PosY, float sizeX, float sizeY);
	/** 
	 * Calculates the vertices for the y-slices according to the given parameters.
	 * @param PosX The x-position of the left upper corner in screen coordinates.
	 * @param PosY The y-position of the left upper corner in screen coordinates.
	 * @param sizeX The width, measured in pixel.
	 * @param sizeY The height, measured in pixel.
	 */
	void setQuadVerticesY(float PosX, float PosY, float sizeX, float sizeY);
	/** 
	 * Calculates the vertices for the z-slices according to the given parameters.
	 * @param PosX The x-position of the left upper corner in screen coordinates.
	 * @param PosY The y-position of the left upper corner in screen coordinates.
	 * @param sizeX The width, measured in pixel.
	 * @param sizeY The height, measured in pixel.
	 */
	void setQuadVerticesZ(float PosX, float PosY, float sizeX, float sizeY);
};