#pragma once

#include <QPointF>
#include <QPoint>
#include <QColor>

/**
 *  TPoint class, inherit from QPoint F 
 *	
 *	defines a point with a color in the plane using floating point precision
 *	the coordinates can be accessed using the x() and y() functions from QPointF
 *	
 */


class TPoint : public QPointF
{

public:
	/** 
	 * Default-Constructor
	 */
	TPoint();
	/** 
	 * Copy-Construct a new Object TPoint with a given QPoint
	 */
	TPoint(const QPoint & point);
	/** 
	 * Copy-Construct a new Object TPoint with a given TPoint
	 */
	TPoint(const TPoint & tpoint);
	/** 
	 * Construct a new Object TPoint with the given coordinates (x, y) and a color.
	 */
	TPoint(float x, float y, const QColor& color);
	
	/** 
	 * Sets the color of a Point to QColor c
	 *
	 *@param c a QColor
	 */
	void setPointColor(const QColor &c) {t_color = c;}
	
	/** 
	 * Returns the color of a Point as QColor
	 *
	 *@return a QColor
	 */
	const QColor& pointColor() const { return t_color;}

private:
	/** 
	* A private variable.
	* QColor to store the Color of a Point
	*/
	QColor t_color;
};
