#pragma once
#include <glm\glm.hpp>
#include "Img.h"
#include <boost/shared_ptr.hpp>

class Texture
{
public:
	typedef boost::shared_ptr<Texture> TexturePtr;

	public:
		static std::vector<Texture*> textures;

		enum //TextureTyp
		{
			TT_NORMAL,
			TT_MIPMAP,
			TT_ANISOTRPIC
		};
		int TextureTyp;

		enum // CorrectionMode
		{
			CM_AFFINE,
			CM_PERSPECTIVE,
			CM_QUANTITY = 2
		};
		int Correction; // default: CM_PERSPECTIVE

		enum // FilterMode
		{
			FM_NEAREST			= 0x2600,
			FM_LINEAR			= 0x2601,
			FM_QUANTITY			= 2
		};
		int Filter; // default: FM_LINEAR

		enum // MipmapMode
		{
			MM_NEAREST			= 0x2600,
			MM_LINEAR			= 0x2601,
			MM_NEAREST_NEAREST	= 0x2700,
			MM_NEAREST_LINEAR	= 0x2702,
			MM_LINEAR_NEAREST	= 0x2701,
			MM_LINEAR_LINEAR	= 0x2703,
			MM_QUANTITY			= 6
		};
		int Mipmap; // default: MM_LINEAR_LINEAR

		enum // CoordinateMode
		{
			WM_CLAMP			= 0x2900,
			WM_REPEAT			= 0x2901,
			WM_CLAMP_BORDER		= 0x812D,
			WM_CLAMP_EDGE		= 0x812F,
			WM_MIRRORED_REPEAT	= 0x8370,
			WM_QUANTITY = 5
		};

		glm::vec4 BorderColor; // default: ColorRGBA(0,0,0,0)
		int CoordU; // default: WM_CLAMP_EDGE
		int CoordV; // default: WM_CLAMP_EDGE

		enum // TexGenMode
		{
			TG_NONE,
			TG_ENVIRONMENT_MAP,	/*IMPLEMENTIERT*/
			TG_PROJECTED_TEXTURE,
			TG_QUANTITY
		};
		int Texgen; // default: TG_NONE

		/********* MULTI TEXTURIERUNG *********/
		enum // ApplyMode
		{
			AM_REPLACE	= 0x1E01,
			AM_DECAL	= 0x2101,
			AM_MODULATE = 0x2100,
			AM_BLEND	= 0x0BE2,
			AM_ADD		= 0x0104,
			AM_COMBINE	= 0x8570,
			AM_QUANTITY	= 6
		};
		glm::vec4 BlendColor; // default: ColorRGBA(0,0,0,1)
		int Apply; // default: AM_REPLACE

		enum // ApplyCombineFunction
		{
			ACF_REPLACE		= 0x1E01,
			ACF_MODULATE	= 0x2100,
			ACF_ADD			= 0x0104,
			ACF_ADD_SIGNED	= 0x8574,
			ACF_SUBTRACT	= 0x84E7,
			ACF_INTERPOLATE	= 0x8575,
			ACF_DOT3_RGB	= 0x86AE,
			ACF_DOT3_RGBA	= 0x86AF,
			ACF_QUANTITY	= 8
		};
		enum // ApplyCombineSrc
		{
			ACS_TEXTURE			= 0x1702,
			ACS_PRIMARY_COLOR	= 0x8577,
			ACS_CONSTANT		= 0x8576,
			ACS_PREVIOUS		= 0x8578,
			ACS_QUANTITY		= 4
		};

		enum // ApplyCombineOperand
		{
			ACO_SRC_COLOR			= 0x0300,
			ACO_ONE_MINUS_SRC_COLOR = 0x0301,
			ACO_SRC_ALPHA			= 0x0302,
			ACO_ONE_MINUS_SRC_ALPHA	= 0x0303,
			ACO_QUANTITY			= 4
		};

		enum // ApplyCombineScale
		{
			ACSC_ONE,
			ACSC_TWO,
			ACSC_FOUR,
			ACSC_QUANTITY
		};

		int CombineFuncRGB; // default: ACF_REPLACE
		int CombineFuncAlpha; // default: ACF_REPLACE

		int CombineSrc0RGB; // default: ACS_TEXTURE
		int CombineSrc1RGB; // default: ACS_TEXTURE
		int CombineSrc2RGB; // default: ACS_TEXTURE
		int CombineSrc0Alpha; // default: ACS_TEXTURE
		int CombineSrc1Alpha; // default: ACS_TEXTURE
		int CombineSrc2Alpha; // default: ACS_TEXTURE
		int CombineOp0RGB; // default: ACO_SRC_COLOR
		int CombineOp1RGB; // default: ACO_SRC_COLOR
		int CombineOp2RGB; // default: ACO_SRC_COLOR
		int CombineOp0Alpha; // default: ACO_SRC_COLOR
		int CombineOp1Alpha; // default: ACO_SRC_COLOR
		int CombineOp2Alpha; // default: ACO_SRC_COLOR
		int CombineScaleRGB; // default: ACSC_ONE
		int CombineScaleAlpha; // default: ACSC_ONE

		/*********IMAGE*********/
		/**
		 * Standardconstructor
		 */
		Texture(): texname(0),  m_dimension(2), texname2(0), m_pingbong(false), current_texture(0) {};
		/**
		 * Constructor
		 * @param dimensions dimensions of texture
		 * @param pingpong true.. pingpong enables otherwise false
		 */
		Texture(unsigned int dimensions, bool pingbong = false):  texname(0), texname2(0), m_dimension(dimensions), m_pingbong(pingbong), current_texture(0) {};
		/**
		 * Constructor
		 * @param pkImage image
		 * @param pingpong true.. pingpong enables otherwise false
		 */
		Texture (Img* pkImage, bool pingbong = false) :  texname(0), texname2(0), m_dimension(2), m_pingbong(pingbong), current_texture(0), BlendColor(0,0,0,1), BorderColor(0,0,0,0), m_pImage(pkImage), TextureTyp(0), Texgen(0){};
		//virtual ~Texture ();

		/**
		 * This method sets the image
		 * @param pImage image to set
		 **/
		void SetImage (Img* pImage);
		/**
		 * This method returns the image
		 * @param retuns the image
		 */
		Img* GetImage () const;
		
		bool m_pingbong;
		unsigned int current_texture;
		
		unsigned int m_dimension;
		std::string texturename;

		unsigned int GetTexName();
		unsigned int GetTexName2();
		void SetTexName(unsigned int tex);

		unsigned int m_width;
		unsigned int m_height;
		unsigned int m_depth;

	protected:
		unsigned int texname;
		unsigned int texname2;

		Img* m_pImage;
	
	private:

};