#include "TransferFunctionWidget.h"
#include "TransferPoints.h"
#include "TPoint.h"
#include <QColor>
#include <QRect>
#include <QTextStream>
#include <QFile>
#include <iostream>
#include <fstream>

using std::cout;
using std::endl;
using std::cerr;


TransferFunctionWidget::TransferFunctionWidget(QWidget *parent) : QWidget(parent), tranferFunktionApplied(false)
{
	m_transferPoints = new TransferPoints(this, TransferPoints::CircleShape);

	transferFunctionTextureData = new float[ELEMENTS_PER_COLOR * COLORS_IN_TRANSFER_ARRAY];
	transferFunctionTextureData2D = new float[1024*1024*4];
	for(int i = 0; i < 1024*1024*4; i++) transferFunctionTextureData2D[i] = 0.f;

	connect(m_transferPoints, SIGNAL(pointsChanged(int)), this, SLOT(pointsUpdated(int)));
	connect(this,SIGNAL(changeColor(QColor)),m_transferPoints, SLOT(changePointColor(QColor)));
	connect(this, SIGNAL(histogramLoaded(double*, int)), m_transferPoints, SLOT(updateHistogram(double*, int)));
}
const QVector<TPoint>& TransferFunctionWidget::points() const
 {
     return m_transferPoints->points();
 }

TransferFunctionWidget::~TransferFunctionWidget()
{
	delete m_transferPoints;
	delete [] transferFunctionTextureData;
	delete [] transferFunctionTextureData2D;
}

void TransferFunctionWidget::paintEvent(QPaintEvent* paintevent)
{
	QPainter p(this);
	p.setPen(QColor(255,255,255));
	p.drawRect(0, 0, width() -1 , height() - 1);

}
inline static bool x_less_than(const TPoint &p1, const TPoint &p2)
{
	return p1.x() < p2.x();
}


void TransferFunctionWidget::pointsUpdated(int mode)
{
	std::cout << "Points Updated: mode"  << mode << std::endl;

	m_type = mode;
	if(mode != 1)
	{
		QVector<TPoint> newPoints;
	
		newPoints = m_transferPoints->points();
		qSort(newPoints.begin(), newPoints.end(), x_less_than);

		transferPoints()->setPoints(newPoints);
		transferPoints()->setPointLock(0, TransferPoints::LockToLeft);
		transferPoints()->setPointLock(newPoints.size() - 1, TransferPoints::LockToRight);
	}
	rebuildTransferFunctionTexture();
	update();
}

void TransferFunctionWidget::initTransferPoints(QString name, int mode)
{
	m_type = mode;
	QVector<TPoint> points;
		
	if (name == "default")
	{	
		points.push_back(TPoint(0.0f, this->height(), QColor(0, 191, 191, 0)));
		points.push_back(TPoint(94.0f, 120.0f, QColor(9, 10, 1, 0)));
		points.push_back(TPoint(167.0f, 29.0f, QColor(206, 109, 57, 193)));
		points.push_back(TPoint(367.0f, 29.0f, QColor(181, 206, 57, 189)));
		points.push_back(TPoint(this->width(), this->height(), QColor(255, 191, 191, 0)));
	}
	else if(name == "reset")
	{
		points.push_back(TPoint(0.0f, this->height(), QColor(0, 191, 191, 0)));
		points.push_back(TPoint(this->width(), 0.0f, QColor(255, 191, 191, 0)));
	} 
	else 
	{
		QString line;
		QStringList list;
		QFile file(name);
		file.open( QIODevice::ReadOnly | QIODevice::Text );
		while (!file.atEnd())
		{
			line = file.readLine();
			//cout << line.toStdString();
			list = line.split(" ", QString::SkipEmptyParts);
			if(list.size() == 6)
			{
				QColor color(list.at(2).toInt(),list.at(3).toInt(),list.at(4).toInt());
				color.setAlpha(list.at(5).toInt());
				//cout << "Alpha: " << list.at(5).toInt() << endl;
				points.push_back(TPoint(list.at(0).toFloat(),list.at(1).toFloat(),color));
			}
			else	cout << "Unable to read file, to less infos" << endl;
		}
	}
	
	m_transferPoints->updateMode(mode);
	m_transferPoints->setConnectionType(TransferPoints::LineConnection);
	m_transferPoints->setPoints(points);
	m_transferPoints->setPointLock(0, TransferPoints::LockToLeft);
	m_transferPoints->setPointLock(points.size()-1, TransferPoints::LockToRight);
	m_transferPoints->setSortType(TransferPoints::XSort);
	rebuildTransferFunctionTexture();
	update();
}

void TransferFunctionWidget::saveTransferPoints(QString fileName)
{
	QFile file(fileName);
	file.open(QIODevice::WriteOnly | QIODevice::Text);
	QTextStream line(&file);

	if (file.isOpen())
	{
		QVector<TPoint> newPoints;
		newPoints = m_transferPoints->points();
		
		for (int i = 0; i < newPoints.size(); ++i)
		{
			int x = newPoints.at(i).x();
			int y = newPoints.at(i).y();
			QColor color = newPoints.at(i).pointColor();
			line << x << " " << y << " " << color.red() << " " << color.green() << " " << color.blue() << " " << color.alpha() << "\n";
		}
		file.close();
	}
	else	cout << "Unable to save file";

}

void TransferFunctionWidget::setHistogram(double* histogram, int size)
{
	emit histogramLoaded(histogram, size);
}

void TransferFunctionWidget::DrawRectangle(float* tex, int i)
{
	for(int cy = 0; cy < (float)(m_transferPoints->tf_objects[i].size.y)/120.f*1024.f; cy++)
	{
		for(int cx = 0; cx < (m_transferPoints->tf_objects[i].size.x*2.f); cx++)
		{
			int x1 = 0;
			int y1 = 0;
					
			glm::vec4 res = glm::rotate(glm::mat4(), (float)(m_transferPoints->tf_objects[i].rotate), glm::vec3(0.f, 0.f, 1.f)) * glm::vec4((float)(cx), (float)(cy), 0.f, 1.f);

			x1 = glm::floor(res.x);
			y1 = glm::floor(res.y);
						
			x1 += m_transferPoints->tf_objects[i].pos.x*2.0;
			y1 += (float)(m_transferPoints->tf_objects[i].pos.y)/120.f*1024.f;

			if(x1 >= 0 && x1 < 1024 && y1 >= 0 && y1 < 1024)
			{
				int r1, g1, b1, a1;

				m_transferPoints->tf_objects[i].color.getRgb(&r1, &g1, &b1, &a1);
				r1 = std::min(255, std::max(0, r1));
				g1 = std::min(255, std::max(0, g1));
				b1 = std::min(255, std::max(0, b1));
				a1 = std::min(255, std::max(0, a1));
				//b1 = 255.0;
				//r1 = 0.0;
				//g1 = 0.0;
				a1 = 255.0/4.0;
				tex[(1024*4*y1) + (4*x1) + 0] += (float)(r1)/255.f;
				tex[(1024*4*y1) + (4*x1) + 1] += (float)(g1)/255.f;
				tex[(1024*4*y1) + (4*x1) + 2] += (float)(b1)/255.f;
				tex[(1024*4*y1) + (4*x1) + 3] += (float)(a1)/255.f;
			}
		}
	}
}

void TransferFunctionWidget::rebuildTransferFunctionTexture()
{
	std::cout << "rebuildTransferFunctionTexture 1" << std::endl;

	// Fill the array with the colors of the fantastic transfer function, oh yeah!
	// OIDA!! Zzzzzhhh

	double redStep = 0.0f;
	double greenStep = 0.0f;
	double blueStep = 0.0f;
	double alphaStep = 0.0f;

	double currentRed = 0.0f;
	double currentGreen = 0.0f;
	double currentBlue = 0.0f;
	double currentAlpha = 0.0f;

	int stepsBetweenPoints = 0;

	int p1Index = 0;
	int p2Index = 0;

	std::cout << "rebuildTransferFunctionTexture _ RType" << m_type << std::endl;

	if(m_type == 0 || m_type == 2)
	{
		const QVector<TPoint>* myPoints = &m_transferPoints->points();
		const TPoint* current = &myPoints->at(0);
		const TPoint* next;

		for (int i = 1; i < myPoints->size(); i++)
		{
			next = &myPoints->at(i);
			p1Index = (current->x() / this->width()) * (COLORS_IN_TRANSFER_ARRAY - 1);
			p2Index = (next->x() / this->width()) * (COLORS_IN_TRANSFER_ARRAY - 1);

			stepsBetweenPoints = p2Index - p1Index;

			redStep = (next->pointColor().redF() - current->pointColor().redF()) / stepsBetweenPoints;
			greenStep = (next->pointColor().greenF() - current->pointColor().greenF()) / stepsBetweenPoints;
			blueStep = (next->pointColor().blueF() - current->pointColor().blueF()) / stepsBetweenPoints;
			alphaStep = (next->pointColor().alphaF() - current->pointColor().alphaF()) / stepsBetweenPoints;

			currentRed = current->pointColor().redF();
			currentGreen = current->pointColor().greenF();
			currentBlue = current->pointColor().blueF();
			currentAlpha = current->pointColor().alphaF();

			for (int j = p1Index * ELEMENTS_PER_COLOR; j <= p2Index * ELEMENTS_PER_COLOR; j += 4)
			{
				transferFunctionTextureData[j + 0] = static_cast<float>(currentRed);
				transferFunctionTextureData[j + 1] = static_cast<float>(currentGreen);
				transferFunctionTextureData[j + 2] = static_cast<float>(currentBlue);
				transferFunctionTextureData[j + 3] = static_cast<float>(currentAlpha);
				
				currentRed += redStep;
				currentGreen += greenStep;
				currentBlue += blueStep;
				currentAlpha += alphaStep;
			}

			current = next;
		}

		emit ColorgradientChanged(transferFunctionTextureData, COLORS_IN_TRANSFER_ARRAY, m_type);

		if (tranferFunktionApplied) 
		{
			emit transferFunctionChanged(transferFunctionTextureData, COLORS_IN_TRANSFER_ARRAY, m_type);
		}
	}
	else
	{
		/**
		* Die 2D TF Functions
		**/
		float* tex = new float[1024*1024*4];
		for(int i = 0; i < 1024*1024*4; i++)
		{
			transferFunctionTextureData2D[i] = 0.f;
			tex[i] = 0.f;
		}
		std::cout << "rebuildTransferFunctionTexture TF" << std::endl;
		
		int numObj = m_transferPoints->tf_objects.size();
		std::cout << "SIZE: " << m_transferPoints->tf_objects.size() << std::endl;

		for(int i = 0; i < numObj; i++)
		{
			switch(m_transferPoints->tf_objects[i].type)
			{
			case TransferPoints::DrawRectangle:
				std::cout << "HEIGHT: " << m_transferPoints->tf_objects[i].size.y/120.f*1024.f << std::endl;
				std::cout << "WIDTH: "  << m_transferPoints->tf_objects[i].size.x*2.0 << std::endl;

				std::cout << "TFP->TF_OBJECXTS[i]" <<  m_transferPoints->tf_objects[i].pos.x*2.0 << " " << m_transferPoints->tf_objects[i].pos.y/120.f*1024.f << std::endl;

				for(int cy = 0; cy < (float)(m_transferPoints->tf_objects[i].size.y)/120.f*1024.f; cy++)
				{
					for(int cx = 0; cx < (m_transferPoints->tf_objects[i].size.x*2.f); cx++)
					{
						int x1 = 0;
						int y1 = 0;

						glm::vec4 res = glm::rotate(glm::mat4(), (float)(m_transferPoints->tf_objects[i].rotate), glm::vec3(0.f, 0.f, 1.f)) * glm::vec4((float)(cx), (float)(cy), 0.f, 1.f);

						x1 = glm::floor(res.x);
						y1 = glm::floor(res.y);
						
						x1 += m_transferPoints->tf_objects[i].pos.x*2.0;
						y1 += (float)(m_transferPoints->tf_objects[i].pos.y)/120.f*1024.f;

						if(x1 >= 0 && x1 < 1024 && y1 >= 0 && y1 < 1024)
						{
							int r1;
							int g1;
							int b1;
							int a1;
							m_transferPoints->tf_objects[i].color.getRgb(&r1, &g1, &b1, &a1);
							r1 = std::min(255, std::max(0, r1));
							g1 = std::min(255, std::max(0, g1));
							b1 = std::min(255, std::max(0, b1));
							a1 = std::min(255, std::max(0, a1));
							//b1 = 255.0;
							//r1 = 0.0;
							//g1 = 0.0;
							a1 = 255.0/4.0;
							tex[(1024*4*y1) + (4*x1) + 0] += (float)(r1)/255.f;
							tex[(1024*4*y1) + (4*x1) + 1] += (float)(g1)/255.f;
							tex[(1024*4*y1) + (4*x1) + 2] += (float)(b1)/255.f;
							tex[(1024*4*y1) + (4*x1) + 3] += (float)(a1)/255.f;
						}
					}
				}
				break;
			case TransferPoints::DrawEllipse:
				glm::vec2 m  = m_transferPoints->tf_objects[i].m;
				glm::mat2 c  = m_transferPoints->tf_objects[i].c;
				std::cout << "TransferPoints::DrawEllipse: m " << m.x << " " << m.y << std::endl;

				for(int cy = 0; cy < 1024; cy++)
				{
					for(int cx = 0; cx < 1024; cx++)
					{
						//glm::vec2 v = tf_objects[i].pos-m;
						glm::vec2 v = glm::vec2((float)(cx)/1024.f,(float)(cy)/1024.f)-m;
						glm::vec2 res = glm::mat2((float)(v.x), 0.f,(float)(v.y), 0.f)*glm::inverse(c) *v;
						//std::cout << "RES: " << res.x << " " << res.y << std::endl;
						float r = exp(-0.5f* res.x);
						
						//sum++;
						int r1;	int g1; int b1; int a1;
						m_transferPoints->tf_objects[i].color.getRgb(&r1, &g1, &b1, &a1);
							
						//std::cout << "R: " << r << std::endl;
						tex[(1024*(1023-cy)*4)+(cx*4)+0] += std::min(1.f, std::max(0.f, r*1.5f*float(r1)/255.f));
						tex[(1024*(1023-cy)*4)+(cx*4)+1] += std::min(1.f, std::max(0.f, r*1.5f*float(g1)/255.f));
						tex[(1024*(1023-cy)*4)+(cx*4)+2] += std::min(1.f, std::max(0.f, r*1.5f*float(b1)/255.f));
						tex[(1024*(1023-cy)*4)+(cx*4)+3] += std::min(1.f, std::max(0.f, r*1.5f*float(a1)/255.f));
					}
				}
				break;
			}
		}

		for(int cy = 0; cy < 1024; cy++)
		{
			for(int cx = 0; cx < 1024; cx++)
			{
				transferFunctionTextureData2D[(1024*cy*4)+(cx*4)+0] = glm::min(1.f, glm::max(0.f, tex[(1024*cy*4)+(cx*4)+0]));
				transferFunctionTextureData2D[(1024*cy*4)+(cx*4)+1] = glm::min(1.f, glm::max(0.f, tex[(1024*cy*4)+(cx*4)+1]));
				transferFunctionTextureData2D[(1024*cy*4)+(cx*4)+2] = glm::min(1.f, glm::max(0.f, tex[(1024*cy*4)+(cx*4)+2]));
				transferFunctionTextureData2D[(1024*cy*4)+(cx*4)+3] = glm::min(1.f, glm::max(0.f, tex[(1024*cy*4)+(cx*4)+3]));
			}
		}

		delete[] tex;

		std::cout << "Transferwidget TF created" << std::endl;

		if (tranferFunktionApplied) 
		{
			emit transferFunctionChanged(transferFunctionTextureData2D, 1024*1024, m_type);
		}
	}
}

void TransferFunctionWidget::setTranferFunktionApplied(bool apply)
{
	this->tranferFunktionApplied = apply;
}

float* TransferFunctionWidget::getTransferTextureData() const
{
	std::cout << "m_type: "<< m_type << std::endl;
	if(m_type == 1)	return transferFunctionTextureData2D;
	else			return transferFunctionTextureData;
}

void TransferFunctionWidget::SetMode(int mode)
{
	m_type = mode;
}