#pragma once

#include <QWidget>
#include <QPainter>
#include <QMouseEvent>
#include <QPointF>

#include <QVector>
#include "TPoint.h"
#include "TransferPoints.h"

#define ELEMENTS_PER_COLOR 4
#define COLORS_IN_TRANSFER_ARRAY 4096

class TransferFunctionWidget : public QWidget
{
	Q_OBJECT

public:
	/** The constructor for the TransferFunctionWidget. */
	TransferFunctionWidget(QWidget* parent = 0);
	/** The constructor for the TransferFunctionWidget. */
	~TransferFunctionWidget();


	/** paintEvent is called whenever something has changed so that the widget has to update
	*	the paintEvent is also called if update() is used.
	*/
	void paintEvent( QPaintEvent* e);
	/** Returns a List of the Points of the Transferfunction in a QVector */ 
	const QVector<TPoint>& points() const;
	/** Returns a Pointer to the transferPoints Object */
	TransferPoints* transferPoints() const {return m_transferPoints;}

	/** Function to save the Transferfunction in a File fileName.txt */
	void saveTransferPoints(QString fileName);
	/** This function is calles whenever the Transferfunction has changed
	*	and a new 1D Texture has to be calculated
	*/
	void rebuildTransferFunctionTexture();
	/** Sets tranferFunktionApplied to apply
	*	@param apply is true if the "Apply Transferfunktion" has been selected
	*/
	void setTranferFunktionApplied(bool apply);
	/** Returnes the Data of the TransferTexture */
	float* getTransferTextureData() const;
	/** Emits a Signal #histogramLoaded */
	void setHistogram(double* histogram, int size);
	/** Set Modus **/
	void SetMode(int mode);

public slots:
	/** Sortes the Points and calles the paintEvent */
	void pointsUpdated(int);
	/** Function to initialize the Points of the Transferfunction
	*	into the TransferFunctionWidget, default or from a file
	*	@param name is filename of a .txt file
	*/
	void initTransferPoints(QString name, int mode);

signals:
	/** This Signal is emitted when the selected color has changed */
	void changeColor(QColor color);
	/** This Signal is emitted when the Transferfunction - Texture hast changed*/
	void transferFunctionChanged(float* data, int size, int type);
	/** This Signal is emitted when the Transferfunction - Colorgradient has changed  */
	void ColorgradientChanged(float* data, int size, int type);
	/** This Signal is emitted when their is a new Histogram to be painted */
	void histogramLoaded(double* histogram, int size);

public:
	/** pointer to the Object transferPoints*/
	TransferPoints* m_transferPoints;

	void DrawRectangle(float* tex, int i);

private:	
	/** true if the Transferfunction has been applied to the Volume */
	bool tranferFunktionApplied;
	/** the interpolated colorvalues of the Transferfunktion */
	float* transferFunctionTextureData;
	float* transferFunctionTextureData2D;

	int m_type;
};