#include "TransferPoints.h"
#include <QEvent>
#include <QMouseEvent>
#include <QPainter>
#include <QtGlobal>
#include <iostream>
#include <QTime>

using std::cout;
using std::endl;
using std::cerr;

TransferPoints::TransferPoints(QWidget *widget, PointShape shape) : QObject(widget) , sizeX(0), m_mode(0), newAction(true)
{
     m_widget = widget;
     widget->installEventFilter(this);
 
     m_connectionType = LineConnection;
     m_sortType = NoSort;
     m_shape = shape;
	 m_pointPen = QPen(QColor(0,0,0,255),2);
     m_connectionPen = QPen(QColor(0,0,0,125),1);
     m_pointBrush = QBrush(QColor(191, 191, 191, 127));
     m_pointSize = QSize(7, 7);
     m_currentIndex = -1;
     m_editable = true;
     m_enabled = true;

	 m_points = new QVector<TPoint>[3];
	 for(int i = 0; i < 3; i++)	m_points[i] = QVector<TPoint>();
     //connect(this, SIGNAL(pointsChanged(int)), m_widget, SLOT(update()));
 }
 
 
void TransferPoints::setEnabled(bool enabled)
{
     if (m_enabled != enabled) 
	 {
         m_enabled = enabled;
         m_widget->update();
     }
}
 
 
 bool TransferPoints::eventFilter(QObject *object, QEvent *event)
 {
	int mode = m_mode;

	if(event->type() == QEvent::MouseButtonPress) 
	{
		newAction = true;
		if(m_mode == DrawPoint)	drawing_path = QPainterPath();
	}
	else if(event->type() == QEvent::MouseMove) 
	{
		if(newAction) 
		{
			start_rectangle = glm::vec2(((QMouseEvent *)event)->pos().x(), ((QMouseEvent *)event)->pos().y());
			if(m_mode == DrawPoint)		drawing_path.moveTo(((QMouseEvent *)event)->pos().x(), ((QMouseEvent *)event)->pos().y());
		}
		act_rectangle = glm::vec2(((QMouseEvent *)event)->pos().x(), ((QMouseEvent *)event)->pos().y());
		newAction = false;

		if(m_mode == DrawPoint)	drawing_path.lineTo(act_rectangle.x, act_rectangle.y);
	}
	else if(event->type() == QEvent::MouseButtonRelease)
	{
		tf_objects.push_back(TFObjects());
		tf_objects[tf_objects.size()-1].scale = glm::vec2(1.f, 1.f);
		tf_objects[tf_objects.size()-1].rotate = 0.f;

		//FERTIG
		switch(m_action_type)
		{
		case DrawPoint:
				std::cout << "DrawPoint" << std::endl;
				tf_objects[tf_objects.size()-1].path = drawing_path;
				tf_objects[tf_objects.size()-1].type = DrawPoint;
				tf_objects[tf_objects.size()-1].color = QColor(0, 200, 200, 100);
			break;
		case DrawEllipse:
		{
			tf_objects[tf_objects.size()-1].type = DrawEllipse;
			tf_objects[tf_objects.size()-1].color = QColor(200, 200, 0, 100);

			if(start_rectangle.x <= act_rectangle.x)
			{
				if(start_rectangle.y <= act_rectangle.y)	
				{
					tf_objects[tf_objects.size()-1].pos = glm::vec2(start_rectangle.x, start_rectangle.y);
					tf_objects[tf_objects.size()-1].size = glm::vec2(act_rectangle.x-start_rectangle.x, act_rectangle.y-start_rectangle.y);
				}
				else										
				{
					tf_objects[tf_objects.size()-1].pos = glm::vec2(start_rectangle.x, act_rectangle.y);
					tf_objects[tf_objects.size()-1].size = glm::vec2(act_rectangle.x-start_rectangle.x, start_rectangle.y-act_rectangle.y);
				}
			}
			else
			{
				if(start_rectangle.y <= act_rectangle.y)	
				{
					tf_objects[tf_objects.size()-1].pos = glm::vec2(act_rectangle.x, start_rectangle.y);
					tf_objects[tf_objects.size()-1].size = glm::vec2(start_rectangle.x-act_rectangle.x, act_rectangle.y-start_rectangle.y);
				}
				else
				{
					tf_objects[tf_objects.size()-1].pos = glm::vec2(act_rectangle.x, act_rectangle.y);
					tf_objects[tf_objects.size()-1].size = glm::vec2(start_rectangle.x-act_rectangle.x, start_rectangle.y-act_rectangle.y);
				}			
			}
			tf_objects[tf_objects.size()-1].m = (tf_objects[tf_objects.size()-1].pos+(tf_objects[tf_objects.size()-1].size/2.f)) / glm::vec2(512.f, 120.f);
			tf_objects[tf_objects.size()-1].m.y = 1.f-tf_objects[tf_objects.size()-1].m.y;
			tf_objects[tf_objects.size()-1].c = glm::mat2(tf_objects[tf_objects.size()-1].scale.x, 0.f, 0.f, tf_objects[tf_objects.size()-1].scale.y);

		break;
		}
	 case DrawRectangle:
		tf_objects[tf_objects.size()-1].type = DrawRectangle;
		tf_objects[tf_objects.size()-1].color = QColor(200, 0, 200, 100);

		if(start_rectangle.x <= act_rectangle.x)
		{
			if(start_rectangle.y <= act_rectangle.y)	
			{
				tf_objects[tf_objects.size()-1].pos = glm::vec2(start_rectangle.x, start_rectangle.y);
				tf_objects[tf_objects.size()-1].size = glm::vec2(act_rectangle.x-start_rectangle.x, act_rectangle.y-start_rectangle.y);
			}
			else
			{
				tf_objects[tf_objects.size()-1].pos = glm::vec2(start_rectangle.x, act_rectangle.y);
				tf_objects[tf_objects.size()-1].size = glm::vec2(act_rectangle.x-start_rectangle.x, start_rectangle.y-act_rectangle.y);
			}
		}
		else
		{
			if(start_rectangle.y <= act_rectangle.y)	
			{
				tf_objects[tf_objects.size()-1].pos = glm::vec2(act_rectangle.x, start_rectangle.y);
				tf_objects[tf_objects.size()-1].size = glm::vec2(start_rectangle.x-act_rectangle.x, act_rectangle.y-start_rectangle.y);
			}
			else
			{
				tf_objects[tf_objects.size()-1].pos = glm::vec2(act_rectangle.x, act_rectangle.y);
				tf_objects[tf_objects.size()-1].size = glm::vec2(start_rectangle.x-act_rectangle.x, start_rectangle.y-act_rectangle.y);
			}
		}
		break;
		}
	}

    if (object == m_widget && m_enabled) 
	{
         switch (event->type()) 
		 {
         case QEvent::MouseButtonPress:
         {
            QMouseEvent *me = (QMouseEvent *) event;
            TPoint clickPos = me->pos();

            int index = -1;
			for (int i=0; i<m_points[mode].size(); ++i) 
			{
				QPainterPath path;
                if (m_shape == CircleShape)	path.addEllipse(insertBoundingRect(i));
                else						path.addRect(insertBoundingRect(i));

                if (path.contains(clickPos)) 
				{
					index = i;
                    break;
				}
			}

            if (me->button() == Qt::LeftButton) 
			{
				if(m_mode == 0 || m_mode == 2)
				{
					//____________________________
					if (index == -1) 
					{	
						if (!m_editable)	return false;
                     
						int pos = 0;
						// Insert sort for x or y
						if (m_sortType == XSort) 
						{
							for (int i=0; i<m_points[mode].size(); ++i)
                            if (m_points[mode].at(i).x() > clickPos.x()) 
							{
								pos = i;
                                break;
                            }
						} 
						else if (m_sortType == YSort) 
						{
							for (int i=0; i<m_points[mode].size(); ++i)
							{
								if (m_points[mode].at(i).y() > clickPos.y()) 
								{
									pos = i;
									break;
								}
							}
						}
				
						bool validPos = true;
						if (m_points[mode].at(pos).x() - clickPos.x() < m_pointSize.width() / 2)	validPos = false;
						if(clickPos.x() - m_points[mode].at(pos - 1).x() < m_pointSize.width() / 2)	validPos = false;
					 
						if (validPos)
						{
							int y = clickPos.y();
							y = y > 120 ? 120 : y;
							y = y < 0 ? 0 : y;
							p_color.setAlpha(255 - y * 2.125);
							clickPos.setPointColor(p_color);
							m_points[mode].insert(pos, clickPos);
							m_locks.insert(pos, 0);
							m_currentIndex = pos;
							firePointChange();
						}
					} 
					else	
					{
						m_currentIndex = index;
						firePointChange();
					}
				
					return true;
				}
				else firePointChange();
			}
			else if (me->button() == Qt::RightButton) 
			{
				if(m_mode == 0 || m_mode == 2)
				{
					if (index >= 0 && m_editable) 
					{
						if (m_locks[index] == 0) 
						{
							m_locks.remove(index);
							m_points[mode].remove(index);
						}
						firePointChange();
						return true;
					}
				}
				else 
				{
					firePointChange();
					return true;
				}
			} 
			else if (me->button() == Qt::MiddleButton)
			{ 
				if(m_mode == 0 || m_mode == 2)
				{
					if (index >= 0 && m_editable) 
					{
						int y = clickPos.y();
						y = y > 120 ? 120 : y;
						y = y < 0 ? 0 : y;
						p_color.setAlpha(255 - y * 2.125);
						m_points[mode][index].setPointColor(p_color);
						firePointChange();
						return true;
					}
				}
				else	
				{
					firePointChange();
					return true;
				}
			}
		
			break;
		}
        case QEvent::MouseButtonRelease: m_currentIndex = -1;
         break; 
         case QEvent::MouseMove:
             if (m_currentIndex >= 0 && (m_mode == 0 || m_mode == 2))
			 {
				int x = ((QMouseEvent *)event)->pos().x();
				int y = ((QMouseEvent *)event)->pos().y();
				if (x > 0 && x < (m_points[mode][m_points[mode].size() - 1].x() - m_pointSize.width() / 2.0))
				{
					QColor c = m_points[mode][m_currentIndex].pointColor();

					y = y > 120 ? 120 : y;
					y = y < 0 ? 0 : y;
					c.setAlpha(255 - y * 2.125);
					TPoint point(x, y, c);
					movePoint(m_currentIndex, point);
				}
			 }
			 if(m_currentIndex >= 0 && m_mode == 1) movePoint(m_currentIndex, TPoint());
         break;
         case QEvent::Resize:
         {
			 if(m_mode == 0 || m_mode == 2)
			 {
	            QResizeEvent *e = (QResizeEvent *) event;
		        if (e->oldSize().width() == 0 || e->oldSize().height() == 0)	break;
				double stretch_x = e->size().width() / double(e->oldSize().width());
				double stretch_y = e->size().height() / double(e->oldSize().height());
				for (int i=0; i<m_points[mode].size(); ++i) 
				{
					TPoint p = m_points[mode][i];
					//movePoint(i, TPoint(p.x() * stretch_x, p.y() * stretch_y, p.pointColor()), false);
				}
			 }
			 firePointChange();
             break;
         }
 
         case QEvent::Paint:
         {
			QWidget *that_widget = m_widget;
			m_widget = 0;
			//QApplication::sendEvent(object, event);
			m_widget = that_widget;
			paintPoints();
			return true;
         }
         default:   break;
         }
     }
 
     return false;
 }
 
 void TransferPoints::paintPoints()
 {
	 int mode = m_mode;

	 QPainter p;
	 p.begin(m_widget);

	 switch(mode)
	 {
	 case 0:
		 //render histogram
		 if (this->histSize > 1) {
			 double maxIntensity = 1;
			 for (int i = 0; i < this->histSize; i++) 
			 {
				 if (maxIntensity < this->hist[i])	 maxIntensity = this->hist[i];
			 }

			 int maxHeight = m_widget->height();
			 double intensity;
			 int maxWidth = m_widget->width();
			 int width;
			 for (int j = 0; j < this->histSize; j++) 
			 {
				 p.setPen(QColor(80,80,80));
				 intensity = this->hist[j];
				 intensity = (maxHeight * intensity) / maxIntensity;

				width = (maxWidth * j) / this->histSize;
				p.drawLine(width, maxHeight, width, int(maxHeight-intensity));
			}
		}
		break;	
	 case 1:
		for(int y = 0; y < sizeX; y++)	
			p.drawPoint((int)(valuesX[y].x*(float)(m_widget->width())), (int)(((float)(m_widget->height()))-(valuesX[y].y*(float)(m_widget->height()))));
		break;
	 }

	//paint connection lines
	p.setRenderHint(QPainter::Antialiasing);

	if(m_mode == 0 || m_mode == 2)
	{
		if (m_connectionPen.style() != Qt::NoPen && m_connectionType != NoConnection) 
		{
			p.setPen(m_connectionPen);

			if (m_connectionType == CurveConnection) 
			{
				QPainterPath path;
				path.moveTo(m_points[mode].at(0));
				for (int i=1; i<m_points[mode].size(); ++i) 
				{
					TPoint p1 = m_points[mode].at(i-1);
					TPoint p2 = m_points[mode].at(i);
					double distance = p2.x() - p1.x();

					path.cubicTo(p1.x() + distance / 2, p1.y(), p1.x() + distance / 2, p2.y(), p2.x(), p2.y());
				}
				p.drawPath(path);
			} 
			else 
			{
				for (int i=1; i<m_points[mode].size(); ++i)	p.drawLine(m_points[mode][i - 1], m_points[mode][i]);
			}
		}

		//paint Points
		const TPoint* tPoint;
		for (int i=0; i<m_points[mode].size(); ++i) 
		{	
			tPoint = &m_points[mode].at(i);
			p.setBrush(QBrush(tPoint->pointColor()));
			QColor c = tPoint->pointColor();
			c.setAlpha(255);
			p.setPen(QPen(c,2));
				
	        QRectF bounds = pointBoundingRect(i);
		    if (m_shape == CircleShape)	p.drawEllipse(bounds);
			else	p.drawRect(m_points[mode].at(i).x(), m_points[mode].at(i).y(), m_pointSize.width(), m_pointSize.height());
		}
	}
	else
	{
		switch(m_action_type)
		{
		case TransferPoints::DrawPoint:	
		{
			p.setPen(QPen(tf_objects[tf_objects.size()-1].color,20));
			//p.drawPath(drawing_path);
			break;
		}
		case DrawEllipse:
			if(start_rectangle.x <= act_rectangle.x)
			{
				if(start_rectangle.y <= act_rectangle.y)	
				{
					p.translate(start_rectangle.x, start_rectangle.y);
//					p.drawEllipse(0.f, 0.f,	act_rectangle.x-start_rectangle.x, act_rectangle.y-start_rectangle.y);
					p.translate(-start_rectangle.x, -start_rectangle.y);
				}
				else										
				{
					p.translate(start_rectangle.x, act_rectangle.y);
//					p.drawEllipse(0.f, 0.f,		act_rectangle.x-start_rectangle.x, start_rectangle.y-act_rectangle.y);
					p.translate(-start_rectangle.x, -act_rectangle.y);
				}
			}
			else
			{
				if(start_rectangle.y <= act_rectangle.y)	
				{
					p.translate(act_rectangle.x, start_rectangle.y);
//					p.drawEllipse(0.f, 0.f,	start_rectangle.x-act_rectangle.x, act_rectangle.y-start_rectangle.y);
					p.translate(-act_rectangle.x,-start_rectangle.y);
				}
				else
				{
					p.translate(act_rectangle.x, act_rectangle.y);
//					p.drawEllipse(0.f, 0.f,	start_rectangle.x-act_rectangle.x, start_rectangle.y-act_rectangle.y);
					p.translate(-act_rectangle.x, -act_rectangle.y);
				}			
			}
			break;
		case DrawRectangle:
			if(start_rectangle.x <= act_rectangle.x)
			{
				if(start_rectangle.y <= act_rectangle.y)	
				{
					p.translate(start_rectangle.x, start_rectangle.y);
//					p.drawRect(0.f, 0.f,	act_rectangle.x-start_rectangle.x, act_rectangle.y-start_rectangle.y);
					p.translate(-start_rectangle.x, -start_rectangle.y);
				}	
				else
				{
					p.translate(start_rectangle.x, act_rectangle.y);
//					p.drawRect(0.f, 0.f,		act_rectangle.x-start_rectangle.x, start_rectangle.y-act_rectangle.y);
					p.translate(-start_rectangle.x, -act_rectangle.y);
				}
			}	
			else
			{
				if(start_rectangle.y <= act_rectangle.y)	
				{
					p.translate(act_rectangle.x, start_rectangle.y);
//					p.drawRect(0.f, 0.f,	start_rectangle.x-act_rectangle.x, act_rectangle.y-start_rectangle.y);
					p.translate(-act_rectangle.x, -start_rectangle.y);
				}	
				else
				{	
					p.translate(act_rectangle.x, act_rectangle.y);
//					p.drawRect(0.f, 0.f,	start_rectangle.x-act_rectangle.x, start_rectangle.y-act_rectangle.y);
					p.translate(-act_rectangle.x, -act_rectangle.y);
				}
			}
			break;
		}

		for(int i = 0; i < tf_objects.size(); i++)
		{
			p.setBrush(QBrush(tf_objects[i].color));
			p.translate(tf_objects[i].pos.x, tf_objects[i].pos.y);
			p.rotate(tf_objects[i].rotate);
			p.scale(tf_objects[i].scale.x, tf_objects[i].scale.y);

			switch(tf_objects[i].type)
			{
				case TransferPoints::DrawPoint:		
				{
					p.setPen(QPen(tf_objects[tf_objects.size()-1].color,10));
					//p.drawPath(tf_objects[i].path);
					break;
				}
				case TransferPoints::DrawEllipse:	
				{
					/*glm::vec2 m  = tf_objects[i].pos+(tf_objects[i].size/2.f);

					glm::mat2 c;
					if(tf_objects[i].size.x >= tf_objects[i].size.y)	c = Clustering::CalculateCovariance(tf_objects[i].size, 0.f);
					else												c = Clustering::CalculateCovariance(glm::vec2(tf_objects[i].size.y, tf_objects[i].size.x), 0.f);
					*/

					glm::vec2 m  = tf_objects[i].m;
					glm::mat2 c  = tf_objects[i].c;

					std::cout << "Mittelwert: " <<  m.x << " " <<  m.y << std::endl;

					int sum = 0;
					unsigned int* img = new unsigned int[512*120*4];
					for(int cy = 0.f; cy < 120; cy++)
					{
						for(int cx = 0; cx < 512; cx++)
						{
							//glm::vec2 v = tf_objects[i].pos-m;
							glm::vec2 v = glm::vec2((float)(cx)/512.f,(float)(cy)/120.f)-m;
							glm::vec2 res = glm::mat2((float)(v.x), 0.f,(float)(v.y), 0.f)*glm::inverse(c) *v;
							float r = exp(-0.5f* res.x);
						
							if(r > 0.01f)
							{
								int r1;	int g1; int b1; int a1;
								tf_objects[i].color.getRgb(&r1, &g1, &b1, &a1);

								img[(512*(119-cy)*4)+(cx*4)+0] = ((unsigned int)((float)(255.f)*r)*1.5)*(float)(r1);
								img[(512*(119-cy)*4)+(cx*4)+1] = ((unsigned int)((float)(255.f)*r)*1.5)*(float)(g1);
								img[(512*(119-cy)*4)+(cx*3)+2] = ((unsigned int)((float)(255.f)*r)*1.5)*(float)(b1);
								img[(512*(119-cy)*4)+(cx*4)+3] = ((unsigned int)((float)(255.f)*r)*1.5)*(float)(a1);

								QColor c = tf_objects[i].color;
								c.setAlpha((unsigned int)(r*255.f));
								p.setPen(QPen(c,2));	
								//p.drawPoint(cx,119-cy);
							}
							else
							{
								img[(512*(119-cy)*4)+(cx*4)+0] = 0;
								img[(512*(119-cy)*4)+(cx*4)+1] = 0;
								img[(512*(119-cy)*4)+(cx*4)+2] = 0;
								img[(512*(119-cy)*4)+(cx*4)+3] = 0;
							}
						}
					}

					//QPixmap pixmap = QPixmap::fromImage(QImage((unsigned char *) img, 512, 120, QImage::Format_ARGB32));
					//p.drawPixmap(0.f, 0.f, tf_objects[i].size.x, tf_objects[i].size.y, pixmap);
					//p.drawPixmap(0.f, 0.f, 512, 120, pixmap);


					QColor c2 = QColor(0,0,0);
					c2.setAlpha((unsigned int)(255.f));
					p.setPen(QPen(c2,2));	

						p.drawEllipse(0.f, 0.f, tf_objects[i].size.x, tf_objects[i].size.y);
					
					std::cout << "DrawEllipse POS: " << tf_objects[i].pos.x << " " << tf_objects[i].pos.y << std::endl;
					std::cout << "DrawEllipse SIZE: " << tf_objects[i].size.x << " " << tf_objects[i].size.y << std::endl;
					break;
				}
				case TransferPoints::DrawRectangle:	
				{
					p.drawRect(0.f, 0.f, tf_objects[i].size.x, tf_objects[i].size.y);
					break;
				}
			}

			p.scale(1.f/tf_objects[i].scale.x, 1.f/tf_objects[i].scale.y);
			p.rotate(-tf_objects[i].rotate);
			p.translate(-tf_objects[i].pos.x, -tf_objects[i].pos.y);
		}
	}
}
 
 static TPoint bound_point(const TPoint &point, const QRectF &bounds, int lock)
 {
     TPoint p = point;
 
     double left = bounds.left();
     double right = bounds.right();
     double top = bounds.top();
     double bottom = bounds.bottom();
 
     if (p.x() < left || (lock & TransferPoints::LockToLeft)) p.setX(left);
     else if (p.x() > right || (lock & TransferPoints::LockToRight)) p.setX(right);
 
     if (p.y() < top || (lock & TransferPoints::LockToTop)) p.setY(top);
     else if (p.y() > bottom || (lock & TransferPoints::LockToBottom)) p.setY(bottom);
 
     return p;
 }
 
 void TransferPoints::setPoints(const QVector<TPoint> &points)
 {
	 int mode = m_mode;
     m_points[mode].clear();
     for(int i=0; i<points.size(); ++i)	m_points[mode] << bound_point(points.at(i), boundingRect(), 0);
 
     m_locks.clear();
     if(m_points[mode].size() > 0) 
	 {
         m_locks.resize(m_points[mode].size());
         m_locks.fill(0);
     }
 }
 
 void TransferPoints::movePoint(int index, const TPoint &point, bool emitUpdate)
 {
	 int mode = m_mode;
	 bool validPos = true;

	 if (index < m_points[mode].size() - 1 && qAbs(m_points[mode][index + 1].x() - point.x()) < m_pointSize.width() * 0.5)	validPos = false;
	 if (index > 0 && qAbs(m_points[mode][index - 1].x() - point.x()) < m_pointSize.width() * 0.5)	validPos = false;
	 if (validPos)	m_points[mode][index] = bound_point(point, boundingRect(), m_locks.at(index));
     if (emitUpdate)	firePointChange();
 }
 
 
 inline static bool x_less_than(const TPoint &p1, const TPoint &p2)
 {
     return p1.x() < p2.x();
 }
 
 inline static bool y_less_than(const TPoint &p1, const TPoint &p2)
 {
     return p1.y() < p2.y();
 }
 
 void TransferPoints::firePointChange()
 {
	 int mode = m_mode;

     if (m_sortType != NoSort && (mode == 0 || mode == 2)) 
	 {
         TPoint oldCurrent;
         if (m_currentIndex != -1) oldCurrent = m_points[mode][m_currentIndex];
         
         if (m_sortType == XSort)		qSort(m_points[mode].begin(), m_points[mode].end(), x_less_than);
         else if (m_sortType == YSort)  qSort(m_points[mode].begin(), m_points[mode].end(), y_less_than);
 
         // Compensate for changed order...
         if (m_currentIndex != -1) 
		 {
             for (int i=0; i<m_points[mode].size(); ++i)
			 {
                 //if (m_points[i] == oldCurrent)
				 if (m_points[mode][i].x() == oldCurrent.x())
				 {
                     m_currentIndex = i;
                     break;
                 }
             }
         }
     }

	 emit pointsChanged(mode);
 }

const QVector<TPoint>& TransferPoints::points() const
{ 
	int mode = m_mode;
	return m_points[mode];
}

void TransferPoints::updateHistogram(double* histogram, int size)
{
	this->hist = histogram;
	this->histSize = size;

	m_widget->update();
}

void TransferPoints::updateValues(glm::vec2* values, int size)
{
	valuesX = values;
	sizeX = size;
}

void TransferPoints::updateMode(int mode)
{
	m_mode = mode;
	std::cout << "MODE: " << m_mode << std::endl;
	m_widget->update();	
}

void TransferPoints::updateActionObjectType(int type)
{
	m_action_type = type;
}

void TransferPoints::DoClustering(int K)
{
	glm::vec2* m = new glm::vec2[K];
	glm::mat2* c = new glm::mat2[K];

	std::cout << "Do Clustering 1" << std::endl;
	

	QTime* qtimer = new QTime();
	qtimer->start();
		
	Clustering::DoClustering(m, c, valuesX, sizeX, K);

	std::cout << "******************************" << std::endl;
	std::cout << "******************************" << std::endl;
	std::cout << "******************************" << std::endl;
	std::cout << "******************************" << std::endl;
	std::cout << "Elapsed Time: " << qtimer->elapsed() << std::endl;
	std::cout << "******************************" << std::endl;
	std::cout << "******************************" << std::endl;
	std::cout << "******************************" << std::endl;
	std::cout << "******************************" << std::endl;


	std::cout << "Do Clustering 2" << std::endl;

	std::cout << std::endl << std::endl << "*************************" << std::endl;
	for(int i = 0; i < K; i++) std::cout << "m: " << m[i].x << " " << m[i].y << std::endl;
	std::cout << "*************************" << std::endl;

	for(int i = 0; i < K; i++)
	{
		std::cout <<"***" << std::endl;
		glm::vec2 start_rectangle = m[i];
		glm::mat2 covariance = c[i];

		
		glm::vec2 rw;
		glm::mat2 ru;
		glm::mat2 rvt;
		Clustering::SVD(rw, ru, rvt, covariance);

		std::cout << "Acos: " << asin(ru[0][0])*180.f/3.14159265f << std::endl;
		std::cout << "Asin: " << acos(ru[0][1])*180.f/3.14159265f << std::endl;
		std::cout << "Asin: " << acos(ru[1][0])*180.f/3.14159265f << std::endl;
		std::cout << "Acos: " << asin(ru[1][1])*180.f/3.14159265f << std::endl;
		std::cout << "___" << std::endl;
		std::cout << "Acos: " << asin(rvt[0][0])*180.f/3.14159265f << std::endl;
		std::cout << "Asin: " << acos(rvt[0][1])*180.f/3.14159265f << std::endl;
		std::cout << "Asin: " << acos(rvt[1][0])*180.f/3.14159265f << std::endl;
		std::cout << "Acos: " << asin(rvt[1][1])*180.f/3.14159265f << std::endl;

		tf_objects.push_back(TFObjects());
		tf_objects[tf_objects.size()-1].type = DrawEllipse;
		tf_objects[tf_objects.size()-1].scale = glm::vec2(1.f, 1.f);
		tf_objects[tf_objects.size()-1].rotate = acos(rvt[1][1])*180.f/3.14159265f;

		
		if(i % 7 == 0)	tf_objects[tf_objects.size()-1].color = QColor(200, 0, 0, 70);
		if(i % 7 == 1)	tf_objects[tf_objects.size()-1].color = QColor(0, 200, 0, 20);
		if(i % 7 == 2)	tf_objects[tf_objects.size()-1].color = QColor(0, 0, 200, 70);
		if(i % 7 == 3)	tf_objects[tf_objects.size()-1].color = QColor(200, 200, 0, 70);
		if(i % 7 == 4)	tf_objects[tf_objects.size()-1].color = QColor(200, 0, 200, 70);
		if(i % 7 == 5)	tf_objects[tf_objects.size()-1].color = QColor(0, 200, 200, 70);
		if(i % 7 == 6)	tf_objects[tf_objects.size()-1].color = QColor(200, 200, 200, 70);

		tf_objects[tf_objects.size()-1].pos = glm::vec2(512.f*(m[i].x-(rw.x/2.f)), 120.f-(120.f*(m[i].y-(rw.y/2.f))));
		tf_objects[tf_objects.size()-1].size = glm::vec2(rw.x*512.f, rw.y*120.f);
		
		tf_objects[tf_objects.size()-1].m = m[i];
		tf_objects[tf_objects.size()-1].c = c[i];

		std::cout << "Mittelwert: " <<  m[i].x << " " <<  m[i].y << std::endl;
		std::cout << "RW: " <<  rw.x << " " <<  rw.y << std::endl;
		std::cout << "RU: " <<  ru[0][0] << " " <<  ru[0][1] << " " << ru[1][0] << " " << ru[1][1] << std::endl;
		std::cout << "RU: " <<  rvt[0][0] << " " <<  rvt[0][1] << " " << rvt[1][0] << " " << rvt[1][1] << std::endl;

		std::cout << "*********" << std::endl;
		/*/

		glm::vec2 D;
		glm::mat2 V;
		Clustering::Eigen(V, D, covariance);
		glm::mat2 iV = glm::inverse(covariance);
    
		glm::vec2 P1 = 2.f*sqrt(D.x) * glm::vec2(V[0][0], V[0][0]);
		glm::vec2 P2 = 2.f*sqrt(D.y) * glm::vec2(V[0][1], V[0][1]);
    
		float major = 0.f;
		if(abs(P1.x) >= abs(P2.y))	major = P1.x;
		else						major = P2.y;
  
		float x = 1.f*major;
        
		float a = iV[1][1];
		float b = x * (iV[0][1]+iV[1][0]);
		float c = powf(x,2.f) * iV[0][0] - 1.f;

		glm::vec2 rw = glm::vec2(abs(a), b);

	    tf_objects.push_back(TFObjects());
		tf_objects[tf_objects.size()-1].scale = glm::vec2(1.f, 1.f);
		//tf_objects[tf_objects.size()-1].rotate = asin(rvt[0][1])*180.f/3.14159265f;
		tf_objects[tf_objects.size()-1].rotate = 0.f;
		tf_objects[tf_objects.size()-1].type = DrawEllipse;
		tf_objects[tf_objects.size()-1].color = QColor(200, 200, 0, 100);
		tf_objects[tf_objects.size()-1].pos = glm::vec2((m[i].x-(rw.x/2.0)), 120.f-((m[i].y-(rw.y/2.0))));
		tf_objects[tf_objects.size()-1].size = glm::vec2(rw.x*1.f, rw.y*1.f);
		//*/
	}
	firePointChange();
}