#pragma once
#include "Geometry.h"

#include <GL\glew.h>
#include <glm\glm.hpp>

class VertexGeometry : public Geometry
{
public:
	/*	Dieses Enum dient zur korrekten Zuordnung der VBO zu den In Objekte des Shaders (vertex-attrib).
		Die Idee ist ein Dictionary (in Klasse View), welche jede in-Variable des Shaders dem VBO zuordnet.
		Die VBOs werden durch das nachfolgende Enum erkannt. 
		Also wenn im Dictionary ein AT-POSITIONS drinnen steht, wei die Engine, dass das Positions-VBO
		bentzt werden soll. 

		Um noch weitere VBOs zu untersttzen (Tangenten, Binormale, etc.), muss einerseits dieses Array
		erweitert werden, die VBOs erstellt werden und im DrawVertexGeometry erweitert werden (das switch).	*/
	enum ArrayType
	{
		AT_POSITIONS=0,
		AT_NORMALS=1,
		AT_TEXCOORD=2,
		AT_ANIMATION=3,
		AT_COLORS=4,
		AT_TEXCOORD1 = 5,
		AT_TEXCOORD2 = 6,
		AT_TEXCOORD3 = 7,
		AT_TEXCOORD4 = 8,
	};

public:
	/**
	 * Standard constructor
	 */
	VertexGeometry(void);
	/**
	 * Create empty VertexGeometry. Data are later setted with cuda
	 * @param vertsize number of vertices
	 * @param trigSize number of triangles
	 */
	VertexGeometry(int vertsize, int trigSize);
	/**
	 * Create empty VertexGeometry. Data are later setted with cuda
	 * @param positionsArray Positions array
	 * @param normalArray Normal array
	 * @param colors Color array
	 * @param texCoord Texcoord0 array
	 * @param texCoord1 Texcoord1 array
	 * @param texCoord2 Texcoord2 array
	 * @param indexArray Index array
	 * @param vertsize number of vertices
	 * @param trigSize number of triangles
	 */
	VertexGeometry(glm::vec4* positionsArray, glm::vec3* normalArray, glm::vec4* colors, glm::vec2* texCoord, glm::vec4* texCoord1, glm::vec4* texCoord2, int* indexArray, int vertsize, int trigSize);
	/**
	 * Create empty VertexGeometry. Data are later setted with cuda
	 * @param positionsArray Positions array
	 * @param normalArray Normal array
	 * @param texCoord Texcoord0 array
	 * @param texCoord1 Texcoord1 array
	 * @param texCoord2 Texcoord2 array
	 * @param indexArray Index array
	 * @param vertsize number of vertices
	 * @param trigSize number of triangles
	 */
	VertexGeometry(glm::vec4* positionsArray, glm::vec3* normalArray, glm::vec2* texCoord, glm::vec4* texCoord1, glm::vec4* texCoord2, int* indexArray, int vertsize, int trigsize);
	/**
	 * Create empty VertexGeometry. Data are later setted with cuda
	 * @param positionsArray Positions array
	 * @param normalArray Normal array
	 * @param colors Color array
	 * @param indexArray Index array
	 * @param vertsize number of vertices
	 * @param trigSize number of triangles
	 */
	VertexGeometry(glm::vec4* positionsArray, glm::vec3* normalArray, glm::vec2* texCoord, glm::vec4* colors, int* indexArray, int vertsize, int trigsize);
	/**
	 *Not used
	 */
	VertexGeometry(glm::vec4* positionsArray, glm::vec3* normalArray, glm::vec2* texCoord, glm::vec4* bones_weights, glm::vec4* bones_indices, glm::mat4* bones_matrizen, int* indexArray, int vertsize, int trigsize, int bonesize);
	/**
	 *Not used
	 */
	VertexGeometry(glm::vec4* positionsArray, glm::vec3* normalArray, glm::vec2* texCoord, glm::vec4** animation, int* indexArray, int vertsize, int trigsize, int animationsize);
	/**
	 * Standard deconstructor
	 */
	~VertexGeometry(void);

	/**
	 * Sort operator
	 */
	const bool operator<(VertexGeometry i) const;
	/**
	 * Sort operator
	 */
	const bool operator<(VertexGeometry* i) const;

	/**
	 * Update VBO
	 */
	void Update(float* positionsArray);
	
	/**
	 * Build VBO
	 */
	void BuildVBO();

	/**
	 * Update VertexGeometry
	 */
	void Update(float time, float duration);

	glm::vec4* m_PositionArray;
	unsigned int* m_IndexArray;
	glm::vec3* m_NormalArray;
	glm::vec4* m_ColorArray;

	glm::mat4* m_OffsetMatrix;

	std::map<unsigned int, glm::vec4*> m_TextureCoords;
	std::map<unsigned int, glm::vec4*> m_AnimationVertex;

	unsigned int m_normal_count;
	unsigned int m_color_count;
	unsigned int m_vertex_count;
	unsigned int m_index_count;
	unsigned int m_animation_count;

	unsigned int m_vbovert;
	unsigned int m_vbonorm;
	unsigned int m_vbocolor;
	std::map<unsigned int, unsigned int> m_vbo_tex;
	std::map<unsigned int, unsigned int> m_vbo_animation;

	unsigned int m_vboind;	
	unsigned int m_vao_id;
	unsigned int m_geometrytype;

	float distToCam;

	bool update_bones;

};

