#pragma once
#include "INode.h"
#include "Camera.h"
#include "Light.h"
#include "FBO.h"
#include "MyShader.hpp"

class View
{
public:
	enum UniformType
	{
		UT_IDENTITY,
		UT_MODELMATRIX,
		UT_OLDMODELMATRIX,
		UT_VIEWMATRIX,
		UT_OLDVIEWMATRIX,
		UT_PROJECTIONMATRIX,
		UT_OLDPROJECTIONMATRIX,
		UT_VIEWPROJECTIONMATRIX,
		UT_MODELVIEWPROJECTIONMATRIX,
		UT_OLDMODELVIEWPROJECTIONMATRIX,
		UT_MODELVIEWPROJECTIONLIGHTMATRIX,
		UT_WORLDINVERSETRANSPOSE3x3,
		UT_VIEWPORT_WIDTH,
		UT_VIEWPORT_HEIGHT,
		UT_ACT_FBO_DEPTH,
		UT_FBO_SIZE_3D,
		UT_VIEWPORT,
		UT_TEXTURE0,
		UT_TEXTURE1,
		UT_TEXTURE2,
		UT_TEXTURE3,
		UT_TEXTURE4,
		UT_TEXTURE5,
		UT_TEXTURE6,
		UT_TEXTURE7,
		UT_LIGHT_POSITION,
		UT_LIGHT_AMBIENT,
		UT_LIGHT_DIFFUSE,
		UT_LIGHT_SPECULAR,
		UT_LIGHT_CONSTANTATTENUATION,
		UT_LIGHT_LINEARATTENUATION,
		UT_LIGHT_QUADRATICATTENUATION,
		UT_LIGHT_SHINENESS,
		UT_MATERIAL_EMISSIVE,
		UT_MATERIAL_AMBIENT,
		UT_MATERIAL_DIFFUSE,
		UT_MATERIAL_SPECULAR,
		UT_MATERIAL_SHININESS,
		UT_TIME_DURATION,
		UT_BONES,
		UT_OLDBONES,
		UT_EYE_POSITION,
		UT_VOLUMESIZE,
		UT_RENDERMODE,
		UT_ADDAPTIVE,
		UT_STEPSIZE,
		UT_COMPOSITING_MODE,
		UT_MIP,
		UT_TRANSFER,
		UT_AVERAGE,
		UT_LIGHT,
		UT_PHONG,
		UT_PARTICLE_SIZE,
		UT_HDR_MODE,
		UT_HDR_ENABLE

	};

public:
	/**
	 * Standardconstructor
	 */
	View(void);
	/**
	 * Standarddeconstructor
	 */
	~View(void);

	/**
	 * Sets Light
	 * @param light Light object
	 */
	void SetLight(Light* light);
	/**
	 * Sets Camera
	 * @param camera Camera object
	 */
	void SetCamera(Camera* camera);
	/**
	 * Sets Rootnode
	 * @param node rootnode
	 */
	void SetRoot(INode* inode);

	/**
	 * Returns Camera
	 * @param camera object
	 */
	Camera* GetCamera();
	/**
	 * Returns root node object
	 * @param root-node object
	 */
	INode* GetRoot();
	/**
	 * Returns light object
	 * @param light Light Object
	 */
	Light* GetLight();

	/**
	 * Sets Pass object
	 * @param pass pass variable
	 */
	void SetPass(unsigned int pass);
	/**
	 * Returns Pass variable
	 * @return pass pass variable
	 */
	unsigned int GetPass();

	/**
	 * Add FBO
	 * @param FBO FBO object
	 */
	void AddFBO(FBO::FBOPtr fbo);
	/**
	 * Sets FBO at id
	 * @param FBO FBO object
	 * @param id id
	 */
	void SetFBO(unsigned int id, FBO::FBOPtr fbo);
	/**
	 * Returns FBO
	 * @return FBO object
	 */
	FBO::FBOPtr GetFBO();
	/**
	 * Returns FBO at id
	 * @param id id
	 * @return FBO object
	 */
	FBO::FBOPtr GetFBO(int at);

	/**
	 * Returns Grid index
	 * @return size size
	 */
	glm::vec3 GetGrid();
	/**
	 * Sets Grid INdex
	 * @return grid grid
	 */
	void SetGrid(glm::vec3 grid);

private:
	unsigned int m_pass;

	INode* m_root;
	Camera* m_camera;
	Light* m_light;
	std::vector<FBO::FBOPtr> m_fbo;

	glm::vec3 m_grid;
};